/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2004 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   As a special exception, Pasquale Tricarico gives permission to
   link this program with Qt commercial edition, and distribute the
   resulting executable, without including the source code for the Qt
   commercial edition in the source distribution.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "orsa_file.h"
#include "orsa_error.h"
#include "orsa_secure_math.h"

#include <stdio.h>

#include "sdncal.h"
#include "jpleph.h"
#include "jpl_int.h"

#include <iostream>

using namespace std;

namespace orsa {
  

  // UPDATE (Apr 2004): the results are the same, no matter what default object is selected
  // const JPL_planets JPLFile::default_ephem_center = SOLAR_SYSTEM_BARYCENTER;
  // const JPL_planets JPLFile::default_ephem_center = SUN;
  //
  // CORRECT: keep the EARTH as default center
  // const JPL_planets JPLFile::default_ephem_center = EARTH;
  //
  const JPL_planets JPLFile::default_ephem_center = SOLAR_SYSTEM_BARYCENTER;
  
  // JPL
  JPLFile::JPLFile(string name) : calc_velocity(true) {
    
    int N=0; // number of elements in nam[][6] and val[]
    const int max_N = 256;
    char nam[max_N][6];
    double val[max_N];
    
    // jpl_database = (jpl_eph_data *) jpl_init_ephemeris(name.c_str(),nam,val);
    jpl_database = (void *) jpl_init_ephemeris(name.c_str(),NULL,NULL);
    
    if (jpl_database != 0) {
      N = ((jpl_eph_data*)jpl_database)->ncon;
      if (N > max_N) {
	char err_msg[1024];
	sprintf(err_msg,"ERROR: assumed max_N=%i is smaller than N=%i. Please recompile with a bigger max_N.",max_N,N);
	ORSA_ERROR_MACRO(err_msg);
       	exit(0);
      }
      jpl_close_ephemeris((jpl_eph_data *)jpl_database);
      
      // jpl_database = (void *) jpl_init_ephemeris(name.c_str(),nam,val);
      jpl_database = (void *) jpl_init_ephemeris(name.c_str(),nam,val);
    }
    
    if (jpl_database==0) {
      char err_msg[1024];
      sprintf(err_msg,"Can't open JPL ephemeris file %s",name.c_str());
      ORSA_ERROR_MACRO(err_msg);
      // exit is too much, should handle better this problem
      exit(0);
    }
    
    // jpl_eph_data *jpldb = (jpl_eph_data *) jpl_database;
    // ephem_start.SetTime(FromUnits(jpldb->ephem_start,DAY));
    // ephem_end.SetTime(  FromUnits(jpldb->ephem_end,  DAY));
    
    bool_ephem_start_computed = bool_ephem_end_computed = false;  
    
    map_tag = new map<std::string,double>;
    
    // fill map_tag, DON'T REMOVE
    {
      string tag;
      char ctag[7];
      ctag[6] = 0;
      // cerr << "N: " << N << endl;
      for (int l = 0; l < N; l ++) {
	memcpy(ctag, nam[l], 6);
	tag = ctag;
	remove_leading_trailing_spaces(tag);
	(*map_tag)[tag] = val[l];
#if 0
	printf(" [l=%03i][%s] = %20.12e\n",l,tag.c_str(),val[l]);
	printf(" map_tag[%s] = %20.12e\n",tag.c_str(),(*map_tag)[tag]);
	printf(" map_tag[%s] = %20.12e\n",tag.c_str(),GetTag(tag));
#endif
      } 
    } 
  }
  
  JPLFile::~JPLFile() {
    if (jpl_database != 0) jpl_close_ephemeris((jpl_eph_data *)jpl_database);
    if (map_tag) delete map_tag;
  }
  
  UniverseTypeAwareTime JPLFile::EphemStart() {
    if (!bool_ephem_start_computed) ComputeEphemStart();
    return ephem_start;
  }
  
  UniverseTypeAwareTime JPLFile::EphemEnd() {
    if (!bool_ephem_end_computed) ComputeEphemEnd();
    return ephem_end;
  }
  
  void JPLFile::ComputeEphemStart() {
    jpl_eph_data *jpldb = (jpl_eph_data *) jpl_database;
    ephem_start.SetTime(FromUnits(jpldb->ephem_start,DAY));
    bool_ephem_start_computed = true;
  }
  
  void JPLFile::ComputeEphemEnd() {
    jpl_eph_data *jpldb = (jpl_eph_data *) jpl_database;
    ephem_end.SetTime(FromUnits(jpldb->ephem_end,DAY));
    bool_ephem_end_computed = true;
  }
  
  double JPLFile::GetAU_MKS() {
    return (GetTag("AU")*1.0e3);
  }
  
  double JPLFile::GetMSun_MKS() {
    return (GetTag("GMS")*secure_pow(GetAU_MKS(),3.0)*secure_pow(24*3600.0,-2.0)/GetG_MKS());
  }
  
  double JPLFile::GetMJupiter_MKS() {
    return (GetTag("GM5")*secure_pow(GetAU_MKS(),3.0)*secure_pow(24*3600.0,-2.0)/GetG_MKS());
  }
  
  double JPLFile::GetMEarth_MKS() {
    // const double EMRAT = val[7]; // Earth/Moon mass ratio
    const double EMRAT = GetTag("EMRAT");
    // return ((val[10]*EMRAT/(1+EMRAT))*pow(AU_MKS,3)*pow(24*3600.0,-2)/G_MKS);
    // return ((GetTag("GMB")*EMRAT/(1+EMRAT))*pow(AU_MKS,3.0)*pow(24*3600.0,-2.0)/G_MKS);
    return ((GetTag("GMB")*EMRAT/(1+EMRAT))*secure_pow(GetAU_MKS(),3.0)*secure_pow(24*3600.0,-2.0)/GetG_MKS());
  }
  
  double JPLFile::GetMMoon_MKS() {
    // const double EMRAT = val[7]; // Earth/Moon mass ratio
    const double EMRAT = GetTag("EMRAT");
    // return ((val[10]/(1+EMRAT))*pow(AU_MKS,3)*pow(24*3600.0,-2)/G_MKS);
    // return ((GetTag("GMB")/(1+EMRAT))*pow(AU_MKS,3.0)*pow(24*3600.0,-2.0)/G_MKS);
    return ((GetTag("GMB")/(1+EMRAT))*secure_pow(GetAU_MKS(),3.0)*secure_pow(24*3600.0,-2.0)/GetG_MKS());
  }
  
  double JPLFile::GetC_MKS() {
    return (GetTag("CLIGHT")*1.0e3);
  }
  
  double JPLFile::GetREarth_MKS() {
    return (GetTag("RE")*1.0e3);
  }
  
  double JPLFile::GetRMoon_MKS() {
    return (GetTag("AM")*1.0e3);
  }
  
  double JPLFile::GetMass(const JPL_planets planet) {
    
    // const double file_G = FromUnits( FromUnits( FromUnits(val[17],AU,3),MSUN,-1),DAY,-2);
    // cerr << "JPLFile::GetMass G=" << file_G << endl;
    
    // const double EMRAT = val[7]; // Earth/Moon mass ratio
    const double EMRAT = GetTag("EMRAT");
    
    double GM=0;
    
    switch(planet) {
    case MERCURY:
      GM = GetTag("GM1");
      break;
    case VENUS:
      GM = GetTag("GM2");
      break;
    case MARS:
      GM = GetTag("GM4");
      break;
    case JUPITER:
      GM = GetTag("GM5");
      break;
    case SATURN:
      GM = GetTag("GM6");
      break;
    case URANUS:
      GM = GetTag("GM7");
      break;
    case NEPTUNE: 
      GM = GetTag("GM8");
      break;
    case PLUTO:   
      GM = GetTag("GM9");
      break;
    case EARTH:
      GM = GetTag("GMB")*EMRAT/(1+EMRAT);
      break;
    case MOON:
      GM = GetTag("GMB")/(1+EMRAT);
      break;
    case EARTH_MOON_BARYCENTER:
      GM = GetTag("GMB");
      break;
    case SUN:
      GM = GetTag("GMS");
      break;
    default:
      GM = 0;
      break;
    }
    //
    // take the right units
    GM = FromUnits(FromUnits(GM,AU,3),DAY,-2);
    
    // return (GM/file_G);
    return (GM/GetG());
  }
  
  double JPLFile::GetTag(string tag) {
    remove_leading_trailing_spaces(tag);
    return (*map_tag)[tag];
  }
  
  //
  
  void JPLFile::GetEph(const UniverseTypeAwareTime &date, JPL_planets target, JPL_planets center, Vector &position, Vector &velocity) {
    double  xv[6];   
    // Date ET_date=date.GetDate();
    // ET_date.ConvertToTimeScale(ET);
    
    /*
      const double julian_time = ET_date.GetJulian();
    */
    
    jpl_eph_data *jpldb = (jpl_eph_data *) jpl_database;
    
    /* 
       cerr << "ET_date.GetTime()....: " << ET_date.GetTime()     << endl;
       cerr << "ephem_start.GetTime(): " << ephem_start.GetTime() << endl;  
       cerr << "ephem_end.GetTime()..: " << ephem_end.GetTime()   << endl;
    */
    
    if ( (date.GetTime() < EphemStart().GetTime()) ||
	 (date.GetTime() > EphemEnd().GetTime()) ) {
      
      // cerr << "Warning: requested time out of the jpl database range.\n";
      // debug->debug("Warning: requested time out of the jpl database range.");
      ORSA_ERROR_MACRO("Warning: requested time out of the jpl database range.");
      // int y,m,d;
      // date.GetGregor(y,m,d);
      // cerr << "date: " << d << " " << m << " " << y << endl;
      // debug->debug("date: %i %i %i",d,m,y);
      // Date date(ET);  
      // date.SetJulian(jpldb->ephem_start);
      // date = EphemStart().GetDate();
      // date.ConvertToTimeScale(ET);
      // date.GetGregor(y,m,d);
      // cerr << "JPL ephem start date: " << d << " " << m << " " << y << endl;
      // debug->debug("JPL ephem start date: %i %i %i",d,m,y);
      // date.SetJulian(jpldb->ephem_end);
      // date = EphemEnd().GetDate();
      // date.ConvertToTimeScale(ET);
      // date.GetGregor(y,m,d);
      // cerr << "JPL ephem end date: " << d << " " << m << " " << y << endl;
      // debug->debug("JPL ephem end date: %i %i %i",d,m,y);
      return;
    }
    
    // jpl_pleph(jpldb,julian_time,target,center,xv,calc_velocity);  
    jpl_pleph(jpldb,date.GetDate().GetJulian(ET),target,center,xv,calc_velocity ? 1 : 0);
    // jpl_pleph(jpldb,date.GetDate().GetJulian(),target,center,xv,calc_velocity);  
    
    if ((target==NUTATIONS) || 
	(target==LIBRATIONS)) {
      // no units correction needed, are radians
      position.Set(xv[0],xv[1],xv[2]);
      velocity.Set(xv[3],xv[4],xv[5]);
      return;
    }
    
    // Earth, Earth_J2000
    /* 
       Vector re,ve,rej,vej;
       { 
       double xve[6], xvej[6];
       Date dj; dj.SetJ2000();
       jpl_pleph(jpldb,ET_date.GetJulian(),EARTH,center,xve, calc_velocity ? 1 : 0);  
       jpl_pleph(jpldb,dj.GetJulian(),     EARTH,center,xvej,calc_velocity ? 1 : 0);  
       
       // position (from AU)
       xve[0] = FromUnits(xve[0],AU);
       xve[1] = FromUnits(xve[1],AU);
       xve[2] = FromUnits(xve[2],AU);
       //
       re.Set(xve[0],xve[1],xve[2]);
       
       xvej[0] = FromUnits(xvej[0],AU);
       xvej[1] = FromUnits(xvej[1],AU);
       xvej[2] = FromUnits(xvej[2],AU);
       //
       rej.Set(xvej[0],xvej[1],xvej[2]);
       
       if (calc_velocity) {
       // velocity (from AU/DAY)
       xve[3] = FromUnits(xve[3],AU);
       xve[4] = FromUnits(xve[4],AU);
       xve[5] = FromUnits(xve[5],AU);
       //
       xve[3] = FromUnits(xve[3],DAY,-1);
       xve[4] = FromUnits(xve[4],DAY,-1);
       xve[5] = FromUnits(xve[5],DAY,-1);
       //
       ve.Set(xve[3],xve[4],xve[5]);
       
       // velocity (from AU/DAY)
       xvej[3] = FromUnits(xvej[3],AU);
       xvej[4] = FromUnits(xvej[4],AU);
       xvej[5] = FromUnits(xvej[5],AU);
       //
       xvej[3] = FromUnits(xvej[3],DAY,-1);
       xvej[4] = FromUnits(xvej[4],DAY,-1);
       xvej[5] = FromUnits(xvej[5],DAY,-1);
       //
       vej.Set(xvej[3],xvej[4],xvej[5]);
       }
       }
    */
    
    // position (from AU)
    xv[0] = FromUnits(xv[0],AU);
    xv[1] = FromUnits(xv[1],AU);
    xv[2] = FromUnits(xv[2],AU);
    //
    position.Set(xv[0],xv[1],xv[2]);
    
    if (calc_velocity) {
      // velocity (from AU/DAY)
      xv[3] = FromUnits(xv[3],AU);
      xv[4] = FromUnits(xv[4],AU);
      xv[5] = FromUnits(xv[5],AU);
      //
      xv[3] = FromUnits(xv[3],DAY,-1);
      xv[4] = FromUnits(xv[4],DAY,-1);
      xv[5] = FromUnits(xv[5],DAY,-1);
      //
      velocity.Set(xv[3],xv[4],xv[5]);
    }
    
    /* 
       if (universe->GetReferenceSystem() == ECLIPTIC) {
       Angle obl = obleq(ET_date);
       position.rotate(0.0,-obl.GetRad(),0.0);
       velocity.rotate(0.0,-obl.GetRad(),0.0);
       }
    */
    //
    // THIS IS THE ONLY CORRECT ROTATION:
    // from mean equatorial J2000 (ICRF) to mean ecliptic J2000
    if (universe->GetReferenceSystem() == ECLIPTIC) {
      Angle obl = obleq_J2000();
      position.rotate(0.0,-obl.GetRad(),0.0);
      velocity.rotate(0.0,-obl.GetRad(),0.0);
    }
    //
    /* 
       if (universe->GetReferenceSystem() == ECLIPTIC) {
       // Vector dr = re-rej;
       // Vector dv = ve-vej;
       Vector dr = rej;
       Vector dv = vej;
       //
       position -= dr;
       velocity -= dv;
       //
       Angle obl = obleq_J2000();
       // dr.rotate(0.0,-obl.GetRad(),0.0);
       // dv.rotate(0.0,-obl.GetRad(),0.0);
       position.rotate(0.0,-obl.GetRad(),0.0);
       velocity.rotate(0.0,-obl.GetRad(),0.0);
       //
       position += dr;
       velocity += dv;
       }
    */
    /* 
       if (universe->GetReferenceSystem() == ECLIPTIC) {
       EquatorialToEcliptic(position,ET_date);
       EquatorialToEcliptic(velocity,ET_date);
       }
    */
    /* 
       if (universe->GetReferenceSystem() == ECLIPTIC) {
       EquatorialToEcliptic_J2000(position);
       EquatorialToEcliptic_J2000(velocity);
       }
    */
  }
  
  // small utility
  void SetMRV(double &mass, Vector &r, Vector &v, JPLFile &jf, Date d, JPL_planets p) {
    mass = jf.GetMass(p);  
    jf.GetEph(d,p,r,v);
  }
  
  //! The Sun is automatically included
  void SetupSolarSystem(Frame &frame, const list<JPL_planets> &l, const UniverseTypeAwareTime &t) {
    
    const Date epoch(t.GetDate());
    
    frame.clear();
    frame.SetDate(epoch);
    
    Config conf;
    OrsaConfigFile ocf(&conf);
    ocf.Read();
    ocf.Close();
    const string jpl_path = conf.paths[JPL_EPHEM_FILE]->GetValue().c_str();
    JPLFile jf(jpl_path);
    
    string name;
    double mass;
    Vector r,v;
    
    // date checks
    
    if (t.GetTime() < jf.EphemStart().GetTime()) { 
      char err_msg[1024];
      sprintf(err_msg,"*** Warning: epoch requested is before ephem file start time! (%g < %g)",t.GetTime(),jf.EphemStart().GetTime());
      ORSA_ERROR_MACRO(err_msg);
      return; 
    }
    //
    if (t.GetTime() > jf.EphemEnd().GetTime()) {
      char err_msg[1024];
      sprintf(err_msg,"*** Warning: epoch requested is after ephem file end time! (%g > %g)",t.GetTime(),jf.EphemStart().GetTime());
      ORSA_ERROR_MACRO(err_msg);
      return;
    }
    
    // auto include the Sun
    name="Sun"; SetMRV(mass,r,v,jf,epoch,SUN); frame.push_back(Body(name,mass,r,v));
    
    JPL_planets pl;
    list<JPL_planets>::const_iterator it = l.begin();
    while (it != l.end()) {
      pl = *it;
      if (pl==SUN) {
	++it;
	continue; // sun already included
      } else if (pl==EARTH_AND_MOON) {
	pl = EARTH; name = JPL_planet_name(pl); SetMRV(mass,r,v,jf,epoch,pl);  frame.push_back(Body(name,mass,r,v));
  	pl = MOON;  name = JPL_planet_name(pl); SetMRV(mass,r,v,jf,epoch,pl);  frame.push_back(Body(name,mass,r,v));
      } else {
	name = JPL_planet_name(pl); SetMRV(mass,r,v,jf,epoch,pl); frame.push_back(Body(name,mass,r,v));
      }
      ++it;
    }    
  }
  
  ////
  
  /* 
     bool is_genuine_planet(const JPL_planets p) {
     bool is_genuine;
     switch(p) {
     case SOLAR_SYSTEM_BARYCENTER:
     case EARTH_MOON_BARYCENTER:
     case NUTATIONS:
     case LIBRATIONS:
     case EARTH_AND_MOON:
     is_genuine = false;
     break;
     default:
     is_genuine = true;
     break;
     }
     return (is_genuine);
     }
  */
  
  string JPL_planet_name(const JPL_planets p) {
    
    string name;
    
    switch(p) {
    case SUN:      name = "Sun";      break;
    case MERCURY:  name = "Mercury";  break;
    case VENUS:    name = "Venus";    break;
    case EARTH:    name = "Earth";    break;
    case MARS:     name = "Mars";     break;
    case JUPITER:  name = "Jupiter";  break;
    case SATURN:   name = "Saturn";   break;
    case URANUS:   name = "Uranus";   break;
    case NEPTUNE:  name = "Neptune";  break;
    case PLUTO:    name = "Pluto";    break;
      //
    case MOON:                   name = "Moon";                   break;
    case EARTH_MOON_BARYCENTER:  name = "Earth-Moon barycenter";  break;
      //
    default: break;
    }
    
    return (name);
  }
  
  // data from http://ssd.jpl.nasa.gov/phys_props_planets.html (Dec 6 2003)
  // Sun data: http://www.solarviews.com/eng/sun.htm
  double radius(const JPL_planets p) { 
    double r=0;
    switch(p) {
    case SUN:     r = FromUnits(695000.  ,KM); break;
    case MERCURY: r = FromUnits(  2440.  ,KM); break;
    case VENUS:   r = FromUnits(  6051.84,KM); break;
    case EARTH:   r = FromUnits(  6371.01,KM); break;
    case MARS:    r = FromUnits(  3389.92,KM); break;
    case JUPITER: r = FromUnits( 69911.  ,KM); break;
    case SATURN:  r = FromUnits( 58232.  ,KM); break;
    case URANUS:  r = FromUnits( 25362.  ,KM); break;
    case NEPTUNE: r = FromUnits( 24624.  ,KM); break;
    case PLUTO:   r = FromUnits(  1151.  ,KM); break;
      //
    case MOON:    r = FromUnits(  1738.  ,KM); break;
      //
    default: break;
    }
    return (r);
  }
  
} // namespace orsa
