/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2004 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   As a special exception, Pasquale Tricarico gives permission to
   link this program with Qt commercial edition, and distribute the
   resulting executable, without including the source code for the Qt
   commercial edition in the source distribution.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

// define QT_CLEAN_NAMESPACE to avoid glx.h inclusion probems
#define QT_CLEAN_NAMESPACE

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif // HAVE_CONFIG_H

#include "xorsa_opengl.h"
#include "xorsa_objects_combo.h"
#include "xorsa_units_combo.h"
#include "xorsa_plot_area.h"
#include "xorsa_world.h"
#include "xorsa_wrapper.h"
#include "xorsa_location_selector.h"

#include <orsa_orbit.h>
#include <orsa_file.h>
#include <orsa_error.h>
#include <orsa_secure_math.h>

#ifdef HAVE_GSL
#include <orsa_orbit_gsl.h>
#endif

#include <sdncal.h>

#include <qlayout.h> 
#include <qcheckbox.h> 
#include <qpushbutton.h> 
#include <qstring.h>
#include <qframe.h>
#include <qlabel.h>
#include <qmenubar.h>
#include <qapplication.h>
#include <qdockwindow.h>
#include <qvbox.h> 
#include <qstatusbar.h> 
#include <qbuttongroup.h>
#include <qradiobutton.h>
#include <qfiledialog.h> 
#include <qpngio.h> 
#include <qtoolbutton.h> 
#include <qtooltip.h> 

#include <iostream>
#include "support.h"

// icons
// #include "tmp.xpm"
#include "orbits.xpm"
#include "png.xpm"
#include "film.xpm"
#include "ps.xpm"
#include "pdf.xpm"
#include "bright_positive_z.xpm"
#include "moid.xpm"

// the latest header!
#if !defined(_WIN32) && !defined(Q_OS_MACX)
#include <GL/glx.h>
#endif

using namespace std;
using namespace orsa;

double coefficient_scale_label(const double x, const unsigned int round_digits=1) {
  double l = secure_log10(fabs(x));
  double t = secure_pow(10.0,floor(l)-round_digits+1);
  if (x < 0) t *= -1;
  const double c = fabs(x/t);
  
  if (c <= 1.0) return (1.0 * t);
  if (c <= 2.0) return (2.0 * t);
  if (c <= 5.0) return (5.0 * t);
  if (c <= 10.) return (10. * t);
  if (c >  10.) return (20. * t);
  
  // should never happen
  return x;
}

double coefficient_scale_60_label(const double x) {
  if (x <=  1.0) return ( 1.0);
  if (x <=  2.0) return ( 2.0);
  if (x <=  5.0) return ( 5.0);
  if (x <= 10.0) return (10.0);
  if (x <= 15.0) return (15.0);
  if (x <= 20.0) return (20.0);
  if (x <= 30.0) return (30.0);
  if (x >  30.0) return (60.0);
  
  // should never happen
  return x;
}

double coefficient_angle_label(const double x) {
  // x in degrees!
  if (x >= 1.0) {
    return coefficient_scale_60_label(x);
  } else if (x >= 1.0/60.0) {
    return (1.0/60.0*coefficient_scale_60_label(x*60.0));
  } else {
    return (1.0/3600.0*coefficient_scale_60_label(x*3600.0));
  }
}

static void identity(GLdouble m[16]) {
  m[0+4*0] = 1; m[0+4*1] = 0; m[0+4*2] = 0; m[0+4*3] = 0;
  m[1+4*0] = 0; m[1+4*1] = 1; m[1+4*2] = 0; m[1+4*3] = 0;
  m[2+4*0] = 0; m[2+4*1] = 0; m[2+4*2] = 1; m[2+4*3] = 0;
  m[3+4*0] = 0; m[3+4*1] = 0; m[3+4*2] = 0; m[3+4*3] = 1;
}

GLboolean invert(GLdouble src[16], GLdouble inverse[16]) {
  
  double t;
  int i, j, k, swap;
  GLdouble tmp[4][4];
  
  identity(inverse);
  
  for (i = 0; i < 4; i++) {
    for (j = 0; j < 4; j++) {
      tmp[i][j] = src[i*4+j];
    }
  }
  
  for (i = 0; i < 4; i++) {
    /* look for largest element in column. */
    swap = i;
    for (j = i + 1; j < 4; j++) {
      if (fabs(tmp[j][i]) > fabs(tmp[i][i])) {
	swap = j;
      }
    }
    
    if (swap != i) {
      /* swap rows. */
      for (k = 0; k < 4; k++) {
	t = tmp[i][k];
	tmp[i][k] = tmp[swap][k];
	tmp[swap][k] = t;
        
	t = inverse[i*4+k];
	inverse[i*4+k] = inverse[swap*4+k];
	inverse[swap*4+k] = t;
      }
    }
    
    if (tmp[i][i] == 0) {
      /* no non-zero pivot.  the matrix is singular, which
	 shouldn't happen.  This means the user gave us a bad
	 matrix. */
      return GL_FALSE;
    }
    
    t = tmp[i][i];
    for (k = 0; k < 4; k++) {
      tmp[i][k] /= t;
      inverse[i*4+k] /= t;
    }
    for (j = 0; j < 4; j++) {
      if (j != i) {
	t = tmp[j][i];
	for (k = 0; k < 4; k++) {
	  tmp[j][k] -= tmp[i][k]*t;
	  inverse[j*4+k] -= inverse[i*4+k]*t;
	}
      }
    }
  }
  return GL_TRUE;
}


// XOrsaLabelsModeCombo

XOrsaLabelsModeCombo::XOrsaLabelsModeCombo(QWidget *parent) : QComboBox(false,parent) {
  
  insertItem("massless");
  insertItem("massive");
  insertItem("all");
  
  connect(this,SIGNAL(activated(int)),this,SLOT(SetLabelsMode(int)));
  
  // sync!
  setCurrentItem(1);
  activated(1);
}

void XOrsaLabelsModeCombo::SetLabelsMode(int i) {
  // look at the combobox for the right order
  switch (i) {
  case 0: m = MASSLESS; break;
  case 1: m = MASSIVE; break;
  case 2: m = ALL;     break;
  }  
  emit ModeChanged(m);
}

void XOrsaLabelsModeCombo::SetLabelsMode(OpenGL_Labels_Mode lm) {
  // look at the combobox for the right order
  switch (lm) {
  case MASSLESS: setCurrentItem(0); SetLabelsMode(0);  break;
  case MASSIVE:  setCurrentItem(1); SetLabelsMode(1);  break;
  case ALL:      setCurrentItem(2); SetLabelsMode(2);  break;
  }
}

OpenGL_Labels_Mode XOrsaLabelsModeCombo::GetLabelsMode() {
  return m;
}

//

// test program, return the index of the most perturbing object,
void AutoOrbitIndex(const Frame &f, vector<int> &index) {
  unsigned int j,k;
  index.resize(f.size());
  for (j=0;j<index.size();++j) index[j]=0;
  double tmp,min;
  // double tmp_a,tmp_e; // debug only
  const double min_init=-1;
  Orbit orbit;
  for (j=0;j<f.size();++j) {
    min=min_init;
    
    for (k=0;k<f.size();++k) {
      if (f[k].mass() == 0) continue;
      if (j==k) continue;
      orbit.Compute(f[j],f[k]);
      tmp = orbit.a*(1.0+orbit.e);
      if (((tmp < min) || (min==min_init)) && (orbit.e < 1.0) && (f[k].mass() > f[j].mass())) {
	index[j] = k;
	min = tmp;
	// tmp_a = orbit.a;
	// tmp_e = orbit.e;
      }
    }
    
    /* 
       for (k=0;k<f.size();++k) {
       if (f[k].mass() == 0) continue;
       if (j==k) continue;
       // orbit.Compute(f[j],f[k]);
       // tmp = orbit.a*(1.0+orbit.e);
       tmp = f[k].mass()/(f[j].position()-f[k].position()).LengthSquared();
       cerr << "auto index: " << f[j].name() << " <---> " << f[k].name() << "   tmp: " << tmp << endl;
       if (tmp > min) {
       index[j] = k;
       min = tmp;
       // tmp_a = orbit.a;
       // tmp_e = orbit.e;
       }
       }
    */
    
    // debug
    // cerr << f[j].name() << " ---> " << f[index[j]].name() << "  a: " << tmp_a << "  e: " << tmp_e << endl;
  }
}












































// XOrsaOpenGL

XOrsaOpenGL::XOrsaOpenGL(QWidget *parent, WFlags f) : GLControlWidget(parent,0,0,f) {
  setMinimumSize(100,100);
  init();
}

XOrsaOpenGL::XOrsaOpenGL(int w, int h, QWidget *parent, WFlags f) : GLControlWidget(parent,0,0,f) {
  setMinimumSize(w,h);  
  init();
}

void XOrsaOpenGL::init() {
  
  fov_base=1.0;
  zoom_scale_zero=1.0;
  // struct timezone tz;
  // gettimeofday(&tv_last,&tz);
  qtime.start();
  fps = 0.0;
  last_fps_frame_index = -1;
  fps_running_average_vector.resize(64);
  fps_running_average_index = 0;
  
  // setAnimationDelay(1);
  setAnimationDelay(40);
  
  // ellipse resolution
  // max_circle_index = 18; // never more than 20, eats too much memory
  // max_circle_index = 16; // never more than 20, eats too much memory
  // max_circle_index = 12; // never more than 20, eats too much memory
  max_circle_index = 12; // never more than 20, eats too much memory
  
  frame_changed_from_animation = false;
  evolution = 0;
  
  connect(this,SIGNAL(mouse_position(int,int)),this,SLOT(mouse_slot(int,int)));
  setMouseTracking(true);
  
  bool_animate               = false;
  bool_draw_orbits           = false;
  bool_draw_moid             = false;
  bool_draw_tracks           = false;
  bool_draw_labels           = false;
  bool_draw_reference_system = false;
  bool_draw_xy_grid          = false; 
  bool_draw_z_bars           = false;  
  bool_already_animating     = false;
  bool_absolute_tracks       = false;
  //
  bool_observe_from_direction_body = false;
  
  labels_mode = ALL;
  projection  = OGL_ORTHO;
  
  central_body_index = direction_body_index = 0;
  orbit_reference_body_index = -HEX_AUTO;
  moid_body_1 = moid_body_2 = 0;
  lagrange_body_1 = lagrange_body_2 = 0;
  xRot = yRot = zRot = 0.0;
  bool_use_direction_body = bool_bright_positive_z = bool_draw_lagrange = bool_draw_scale_bar = bool_draw_FPS = bool_light_time_correction = false;
  
#if !defined(_WIN32) && !defined(Q_OS_MACX)
  // default_font.setStyleStrategy(QFont::PreferBitmap); // NEEDED!! Otherwise TrueType fonts can be selected and the program will crash (at least using Qt 3.1.1 under X11)
  default_font.setFamily("fixed");
  default_font.setRawMode(true);
  default_font.setFamily("fixed");
  default_font.setStyleStrategy(QFont::OpenGLCompatible);
  default_font.setPixelSize(10);
#endif
  char err_msg[1024];
  sprintf(err_msg,"font raw name: %s",default_font.rawName().latin1());
  ORSA_ERROR_MACRO(err_msg);		   
}

void XOrsaOpenGL::initializeGL() {
  
#if !defined(_WIN32) && !defined(Q_OS_MACX)
  {
    // tests
    const GLubyte *vendor      = glGetString(GL_VENDOR);
    const GLubyte *renderer    = glGetString(GL_RENDERER);
    const GLubyte *version     = glGetString(GL_VERSION);
    const GLubyte *extensions  = glGetString(GL_EXTENSIONS);
    /*
      glInitNames();
      glPushName(0);
      glLoadName(1);
      glLoadName(2);
      glLoadName(3);
      //
      GLint names[4];
      glGetIntegerv(GL_NAME_STACK_DEPTH,names);
      glGetIntegerv(GL_NAME_STACK_DEPTH);
    */
    //
    Display *dpy = glXGetCurrentDisplay();
    int GLX_major, GLX_minor;
    glXQueryVersion(dpy,&GLX_major,&GLX_minor);
    //
    /* 
       cerr << "===== OpenGL =====" << endl;
       cerr << "Vendor:      " << vendor      << endl;
       cerr << "Renderer:    " << renderer    << endl;
       cerr << "Version:     " << version     << endl;
       cerr << "GLX version: " << GLX_major << "." << GLX_minor << endl;
       cerr << "Estensions:  " << extensions  << endl;
       cerr << "==================" << endl;
    */
    //
    char err_msg[4096];
    sprintf(err_msg,"OpenGL Vendor: %s",(char *)vendor);
    ORSA_ERROR_MACRO(err_msg);
    sprintf(err_msg,"OpenGL Renderer: %s",(char *)renderer);
    ORSA_ERROR_MACRO(err_msg); 
    sprintf(err_msg,"OpenGL Version: %s",(char *)version);
    ORSA_ERROR_MACRO(err_msg);
    sprintf(err_msg,"GLX Version: %i.%i",GLX_major,GLX_minor);
    ORSA_ERROR_MACRO(err_msg); 
    sprintf(err_msg,"OpenGL Estensions: %s",(char *)extensions);
    ORSA_ERROR_MACRO(err_msg);
    //
    /*
      cerr << "GL_NAME_STACK_DEPTH: " << names[0] << " " << names[1] << " " << names[2] << " " << names[3] << endl;
      cerr << "==================" << endl;
    */
  }
#endif
  
  {
    
    // Display *dpy = glXGetCurrentDisplay();
    
    /* int major,minor;
       glXQueryVersion(dpy,&major,&minor);
       cerr << "GLX version " << major << "." << minor << endl;
    */
    
    /* 
       { // useful test to check how many fonts are available, before to load one of them
       int found;
       int max_found=128;
       // char **fn = XListFonts(dpy,"-misc-fixed-medium-r-*-*-20-*-*-*-*-*-*-*",max_found,&found);
       // char **fn = XListFonts(dpy,"-*-fixed-*-*-*-*-*-*-*-*-*-*-*-*",max_found,&found);
       char **fn = XListFonts(dpy,"-*-fixed-*-*-*-*-10-*-*-*-*-*-*-*",max_found,&found);
       cerr << "test: fonts found: " << found << "  (max: " << max_found << ")" <<  endl;
       if (found) { 
       int j=0;
       while (j<found) {
       cerr << fn[j] << endl;
       j++;
       }
       }
       }
    */
    
    // Font font = XLoadFont(dpy,"-*-fixed-bold-*-*-*-*-*-*-*-*-*-*-*"); // -*-fixed-bold-*-*-*-*-*-*-*-*-*-*-*
    // Font font = XLoadFont(dpy,"-misc-fixed-medium-r-*-*-20-*-*-*-*-*-*-*"); // -*-fixed-bold-*-*-*-*-*-*-*-*-*-*-*
    /* 
       Font font = XLoadFont(dpy,"-*-fixed-*-*-*-*-10-*-*-*-*-*-iso8859-*");
       if (font==BadFont) cerr << "BadFont..." << endl;
       base_font = glGenLists(256);
       // cerr << "font: " << font << "   base_font: " << base_font << endl;
       glListBase(base_font);
       // glXUseXFont(font,base_font,128,base_font+1);
       glXUseXFont(font,0,256,base_font);
    */
    
    /* Font ledfont = XLoadFont(dpy,"-*-ledfixed-*-*-*-*-*-*-*-*-*-*-*-*");
       if (ledfont==BadFont) cerr << "BadFont..." << endl;
       base_ledfont = glGenLists(128);
       cerr << "ledfont: " << ledfont << "   base_ledfont: " << base_ledfont << endl;
       glListBase(base_ledfont);
       // glXUseXFont(ledfont,base_ledfont,128,base_ledfont+1);
       glXUseXFont(ledfont,0,128,base_ledfont);
    */
  }
  
  // glLightfv( GL_LIGHT0, GL_POSITION, pos );
  glEnable( GL_CULL_FACE );
  // glEnable( GL_LIGHTING );
  // glEnable( GL_LIGHT0 );
  glEnable( GL_DEPTH_TEST );
  
  // test planet
  /* 
     planet = glGenLists(1);
     glNewList(planet, GL_COMPILE);
     // glMaterialfv( GL_FRONT, GL_AMBIENT_AND_DIFFUSE, agreen );
     // glColor3d(1,0,0);
     GLUquadricObj *qobj = gluNewQuadric();
     gluQuadricDrawStyle(qobj,GLU_FILL); //GLU_LINE
     gluSphere(qobj,0.2,12,12);
     gluDeleteQuadric(qobj);
     glEndList();
  */
  
  //
  // glClearColor(1.0,1.0,1.0,1.0);
  
  // unitary X-Y circle
  /* 
     circle = glGenLists(1);
     glNewList(circle, GL_COMPILE);
     glBegin(GL_LINE_LOOP);
     // const int n_points = 256;
     // const int n_points = 1000;
     const int n_points = 512;
     double theta;
     double s,c;
     const double theta_incr = twopi/n_points;
     int count = 0;
     // while (theta < twopi) {
     while (count < n_points) {
     theta = count * theta_incr;
     s = sin(theta);
     c = cos(theta);
     // glVertex3d(cos(theta),sin(theta),0.0);
     glVertex3d(c,s,0);
     // cerr << c << "  " << s << "  " << theta << "  " << theta * (180/pi) << endl;
     count++;
     // theta += theta_incr;
     }
     glEnd();
     glEndList();
  */
  
  // unitary X-Y circles
  circle = glGenLists(max_circle_index);
  int list_index=0;
  while (list_index < max_circle_index) {
    glNewList(circle+list_index, GL_COMPILE);
    glBegin(GL_LINE_LOOP);
    // const int n_points = 256;
    // const int n_points = 1000;
    // int pow_base=2;
    // const int n_points = secure_pow(pow_base,list_index); // 2^k points
    const int n_points = 1<<list_index; // 2^k points
    // cerr << "k: " << list_index << " 2^k: " << n_points << endl;
    double theta;
    double s,c;
    const double theta_incr = twopi/n_points;
    int count = 0;
    while (count < n_points) {
      theta = count * theta_incr;
      s = sin(theta);
      c = cos(theta);
      // glVertex3d(cos(theta),sin(theta),0.0);
      glVertex3d(c,s,0);
      // cerr << c << "  " << s << "  " << theta << "  " << theta * (180/pi) << endl;
      count++;
      // theta += theta_incr;
    }
    glEnd();
    glEndList();
    ++list_index;
  }
  
  reference_system = glGenLists(1);
  glNewList(reference_system, GL_COMPILE);
  glLineWidth(2.0);
  gl2psLineWidth(2.0);
  // glCallList(circle);
  // glCallList(circle+10); // guessed the right one
  /* 
     if (max_circle_index > 10) 
     glCallList(circle+10);
     else
     glCallList(circle+max_circle_index-1);
  */
  if (max_circle_index > 8) 
    glCallList(circle+8);
  else
    glCallList(circle+max_circle_index-1);
  //
  glBegin(GL_LINES);
  glVertex3d(0,0,0);
  glVertex3d(0,0,1);
  glEnd();
  glBegin(GL_LINES);
  glVertex3d(0,0,0);
  glVertex3d(1,0,0);
  glEnd();
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  glEndList();
  
  glEnable(GL_NORMALIZE);
  
  glEnable(GL_POINT_SMOOTH); 
  
  // glEnable(GL_LINE_SMOOTH); // bad results (lines too big), maybe blending would solve?
  
  // X-Y grid
  xy_grid = glGenLists(1);
  glNewList(xy_grid, GL_COMPILE);
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  int l;
  for (l=-100;l<=100;++l) {
    glBegin(GL_LINES);
    glVertex3d(l,-100,0);
    glVertex3d(l,+100,0);
    glEnd();
    //
    glBegin(GL_LINES);
    glVertex3d(-100,l,0);
    glVertex3d(+100,l,0);
    glEnd();
  }
  glEndList();
  
  // earth
  earth = glGenLists(1);
  glNewList(earth, GL_COMPILE);
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  const lat_lon ll_new_poly = {NEW_POLY,NEW_POLY};
  const lat_lon ll_final_element = {FINAL_ELEMENT,FINAL_ELEMENT};
  lat_lon ll;
  double phi,delta;
  unsigned int world_index = 0;
  // glColor3d(0.1,0.2,0.9);
  glBegin(GL_LINE_STRIP);
  while ((ll=world[world_index]) != ll_final_element) {
    if (ll == ll_new_poly) {
      glEnd();
      glBegin(GL_LINE_STRIP);
    } else {
      phi   = (pi/180.0)*ll.lon;
      delta = (pi/180.0)*ll.lat;
      glVertex3d(cos(phi)*cos(delta),
		 sin(phi)*cos(delta),
		 sin(delta));
    }
    ++world_index;
  }
  glEnd();
  // internal sphere 
  /* 
     glColor3d(0.1,0.1,0.1);   
     GLUquadricObj *qobj = gluNewQuadric();
     // gluQuadricDrawStyle(qobj,GLU_SILHOUETTE);
     gluQuadricDrawStyle(qobj,GLU_FILL);
     // gluQuadricNormals(qobj,GLU_NONE);
     // gluSphere(qobj,evol_frame[j].radius(),10,10);
     // gluSphere(qobj,0.99,36,18);
     gluSphere(qobj,0.99,36,18);
     gluDeleteQuadric(qobj);
  */
  //
  glEndList();
}


void XOrsaOpenGL::resizeGL(int width, int height) {
  
  makeCurrent();
  
  GLdouble w,h;
  
  if (width < height) {
    w = 1.0;
    h = (double) height / (double) width;
  } else {
    w = (double) width / (double) height;
    h = 1.0;
  }
  
  glViewport( 0, 0, width, height);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  
  switch (projection) {
  case OGL_ORTHO:  
    // glOrtho(-fov*w, fov*w, -fov*h, fov*h, -fov*1.0e14, fov*1.0e14);
    // glOrtho(-fov*w, fov*w, -fov*h, fov*h, -10.0*zoom_scale_zero, 10.0*zoom_scale_zero);
    glOrtho(-fov*w, fov*w, -fov*h, fov*h, -3.0*zoom_scale_zero, 3.0*zoom_scale_zero);
    break;
  case OGL_PERSPECTIVE:
    // gluPerspective(fov,w/h,1.0e-3*zoom_scale_zero,zoom_scale_zero);
    // gluPerspective(fov,w/h,1.0e-6*zoom_scale_zero,3.0*zoom_scale_zero);
    //
    // gluPerspective(fov,w/h,1.0*zoom_scale_zero,3.0*zoom_scale_zero);
    // gluPerspective(fov,w/h,0.1*zoom_scale_zero,3.0*zoom_scale_zero);
    // glTranslated(0,0,-0.2*zoom_scale_zero);
    // gluPerspective(fov,w/h,distance*0.1,3.0*zoom_scale_zero);
    // gluPerspective(fov,w/h,distance*0.1,3.0*(distance+zoom_scale_zero));
    gluPerspective(fov,w/h,FromUnits(100,KM),3.0*(distance+zoom_scale_zero)); // the near distance should be dynamic...
    glTranslated(0,0,-distance);
    // glDepthFunc(GL_LEQUAL);
    //
    break;
  }
  
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

unsigned int XOrsaOpenGL::total_frames() const {
  if (evolution != 0) {
    return evolution->size();
  }
  return 0;
}

void XOrsaOpenGL::paintGL() {
  
  // here? probably there is a better position for this code
  if (ObservingFromDirectionBody()) {
    // ...from UpdateVectors() 
    {
      if (central_body_index == -HEX_ORIGIN)
	center.Set(0,0,0);
      else if (central_body_index == -HEX_CENTER_OF_MASS)
	center = evol_frame.CenterOfMass();
      else if ( (central_body_index >= 0) && 
		(central_body_index < (int)evol_frame.size()) )
	center = evol_frame[central_body_index].position();
      else
	center.Set(0,0,0);
      
      if (direction_body_index == -HEX_ORIGIN)
	dir_pos.Set(0,0,0);
      else if (direction_body_index == -HEX_CENTER_OF_MASS)
	dir_pos = evol_frame.CenterOfMass();
      else if ( (direction_body_index >= 0) && 
		(direction_body_index < (int)evol_frame.size()) )
	dir_pos = evol_frame[direction_body_index].position();
      else 
	dir_pos.Set(0,0,0);
      //
      dir_pos -= center;
    }
    
    distance = dir_pos.Length();
    
    // observer on Earth?
    if ( (direction_body_index >= 0) && 
	 (direction_body_index < (int)evol_frame.size()) ) {
      if (evol_frame[direction_body_index].name() == "Earth") {
	
	cerr << "observing from " << location.name << endl;
	
        Vector obspos(location.pxy*cos(location.lon),location.pxy*sin(location.lon),location.pz);
	
       	obspos.rotate(gmst(evol_frame).GetRad(),0,0);
	
	if (universe->GetReferenceSystem() == ECLIPTIC) {
	  EquatorialToEcliptic_J2000(obspos);
	}
	
	// overwrite distance
	distance = (dir_pos+obspos).Length();
      }
    }
    
    resizeGL(width(),height());
  }
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  
  glLoadIdentity();
  
  transform();  
  
  draw();
  
  const double dt = qtime.elapsed()/1000.0;
  qtime.restart();
  
  if (IsAnimating()) {
    
    if (evol_counter != last_fps_frame_index) {
      fps_running_average_vector[fps_running_average_index] = 1.0/dt;
      fps_running_average_index = (fps_running_average_index+1)%(fps_running_average_vector.size());
      
      last_fps_frame_index = evol_counter;
      
      fps = 0.0;
      unsigned int k;
      for (k=0;k<fps_running_average_vector.size();k++) {
	fps += fps_running_average_vector[k];
      }
      fps /= fps_running_average_vector.size();
    }
    
  }
  
}

void XOrsaOpenGL::draw() {
  
  // cerr << "XOrsaOpenGL::draw() --> central_body_index   = " << central_body_index << endl;
  // cerr << "XOrsaOpenGL::draw() --> direction_body_index = " << direction_body_index << endl;
  
  glPushMatrix();
  
  if (evolution) {
    if (evolution->size()) {
      // if (evol_it->size()) {
      
      UpdateVectors();
      
      if (DrawOrbits() || DrawMOID()) {
	if (orbit_reference_body_index==-HEX_AUTO) {
	  if (orbit_reference_body_index_vector[evol_counter].size()==0) {
	    // cerr << "filling orbit index vector for frame " << evol_counter << "..." << endl;
	    vector<int> ref_index;
	    AutoOrbitIndex(evol_frame,ref_index);
	    orbit_reference_body_index_vector[evol_counter] = ref_index;
	  }
	}
      }
      
      if ( UseDirectionBody() && 
	   ( (central_body_index == -HEX_ORIGIN) || 
	     (central_body_index == -HEX_CENTER_OF_MASS) || 
	     (central_body_index < (int)evol_frame.size()) ) ) {
	
	if (! dir_pos.IsZero()) {
	  internal_draw_direction_line(dir_pos);
	}
      }
      
      if ( DrawLagrange() &&
	   (lagrange_body_1 != -HEX_CENTER_OF_MASS) && 
	   (lagrange_body_1 != -HEX_ORIGIN) &&	 
	   (lagrange_body_2 != -HEX_CENTER_OF_MASS) && 
	   (lagrange_body_2 != -HEX_ORIGIN) &&
	   (evol_frame[lagrange_body_1].mass() != 0) &&
	   (evol_frame[lagrange_body_2].mass() != 0) && 
	   (lagrange_body_1 != lagrange_body_2) ) {
	
	Vector dir_pos_mod = -center;
	internal_draw_Lagrange_points(dir_pos_mod);
      }
      
      // note: first draw objects appear on top of the others
      
      internal_draw_time_label();
      
      if (DrawScaleBar()) {
	internal_draw_scalebar();
      }
      
      if (DrawFPS()) {
	internal_draw_FPS();
      }
      
      internal_draw_bodies_and_labels();
      
      if (DrawMOID()) {
	internal_draw_MOID();
      }
      
      if (DrawTracks()) {
	internal_draw_track();
      }
      
      if (DrawOrbits()) {
	internal_draw_orbit_ellipse();
	
	// projection of the orbit to the X-Y plane
	/* 
	   glPushMatrix();
	   glLineStipple(1, 0x3333);
	   glScaled(1.0,1.0,0.0);
	   glEnable(GL_LINE_STIPPLE);
	   gl2psEnable(GL2PS_LINE_STIPPLE);
	   //
	   internal_draw_orbit_ellipse();
	   //
	   glDisable(GL_LINE_STIPPLE);
	   gl2psDisable(GL2PS_LINE_STIPPLE);
	   //
	   glPopMatrix();
	*/
      }
      
      if (DrawReferenceSystem()) {
	internal_draw_reference_system();
      }
      
      if (DrawXYGrid() && (projection==OGL_ORTHO)) {
	internal_draw_xy_grid();
      }
      
      if (DrawZBars()) {
	internal_draw_z_bars();
      }
      
      // call again OSD, expecially for better PostScript output
      // is call order inverted in gl2ps?
      
      internal_draw_time_label();
      
      if (DrawScaleBar()) {
	internal_draw_scalebar();
      }
      
      if (DrawFPS()) {
	internal_draw_FPS();
      }
      
      if (projection==OGL_PERSPECTIVE) {
	internal_draw_distance();
      }
      
      // a test...
      if (DrawXYGrid() && (projection==OGL_PERSPECTIVE)) {    
	double m[16];
	
	/* 
	   glGetDoublev(GL_PROJECTION_MATRIX,m);
	   fprintf(stderr,
	   " projection matrix: %8.4g %8.4g %8.4g %8.4g\n"
	   "                    %8.4g %8.4g %8.4g %8.4g\n"
	   "                    %8.4g %8.4g %8.4g %8.4g\n"
	   "                    %8.4g %8.4g %8.4g %8.4g\n"
	   "                                           \n",
	   m[0],m[4],m[8],m[12],
	   m[1],m[5],m[9],m[13],
	   m[2],m[6],m[10],m[14],
	   m[3],m[7],m[11],m[15]);
	*/
	
	glGetDoublev(GL_MODELVIEW_MATRIX,m);
	
	/* 
	   fprintf(stderr,
	   " modelview matrix: %8.4g %8.4g %8.4g %8.4g\n"
	   "                   %8.4g %8.4g %8.4g %8.4g\n"
	   "                   %8.4g %8.4g %8.4g %8.4g\n"
	   "                   %8.4g %8.4g %8.4g %8.4g\n"
	   "                                          \n",
	   m[0],m[4],m[8],m[12],
	   m[1],m[5],m[9],m[13],
	   m[2],m[6],m[10],m[14],
	   m[3],m[7],m[11],m[15]);
	*/
	
	{
	  // update the inverted matrix
	  invert(m,inverted);
	  
	  double v[3];
	  //
	  v[0] = 0.0;
	  v[1] = 0.0;
	  v[2] = 1.0;
	  //
	  double w[3];
	  //
	  w[0] = v[0]*inverted[0]+v[1]*inverted[4]+v[2]*inverted[8];
	  w[1] = v[0]*inverted[1]+v[1]*inverted[5]+v[2]*inverted[9];
	  w[2] = v[0]*inverted[2]+v[1]*inverted[6]+v[2]*inverted[10];
	  //
	  /*
	    const double var_phi = secure_atan2(w[1],w[0]);
	    const double var_theta = secure_atan2(w[2],secure_sqrt(w[0]*w[0]+w[1]*w[1])); 
	    //
	    cerr << "var_phi: " << var_phi*(180.0/pi) << "    var_theta: " << var_theta*(180.0/pi) << endl;
	  */
	  //
	  if (universe->GetUniverseType() == Real) {
	    switch (universe->GetReferenceSystem()) {
	    case ECLIPTIC:
	      { 
		Vector v;
		v.Set(w[0],w[1],w[2]);
		//
		// v.rotate(0,obleq(evol_frame).GetRad(),0);
		/* 
		   v.rotate(0,0,delta_longitude(evol_frame).GetRad());
		   v.rotate(0,obleq(evol_frame).GetRad()+delta_obleq(evol_frame).GetRad(),0);
		*/
		// v.rotate(0,0,delta_longitude(evol_frame).GetRad());
		//
		// EclipticToEquatorial(v,evol_frame);
		EclipticToEquatorial_J2000(v);
		
		w[0] = v.x;
		w[1] = v.y;
		w[2] = v.z;
	      }
	      break;
	    case EQUATORIAL: break;
	    }
	  }
	  
	  /* 
	     // for future use...
	     double wz[3];
	     wz[0] = w[0];
	     wz[1] = w[1];
	     wz[2] = w[2];
	  */
	  
	  const double phi   = secure_atan2(w[1],w[0]);
	  const double theta = secure_atan2(w[2],secure_sqrt(w[0]*w[0]+w[1]*w[1])); 
	  
	  {
	    Angle angle;
	    double d,p,s,h,m;
	    
	    angle.SetRad(fmod(phi+pi,twopi));
	    angle.GetHMS(h,m,s);
	    fprintf(stderr,"R.A. %02gh %02gm %03.1fs\n",h,m,s);
	    
	    angle.SetRad(-theta);
	    angle.GetDPS(d,p,s);
	    fprintf(stderr,"DEC. %+02g %02g' %03.1f\"\n",d,p,s);
	  }
	  
	  {
	    // sky grid
	    
	    /* 
	       double angle;
	       
	       {
	       // search the projection angle
	       double ph, th;
	       
	       ph = phi + 0.001;
	       th = theta;
	       
	       w[0] = cos(ph)*sin(th);
	       w[1] = sin(ph)*sin(th);
	       w[2] = cos(th);
	       
	       {
	       Vector v;
	       v.Set(w[0],w[1],w[2]);
	       //
	       EquatorialToEcliptic(v,evol_frame);
	       //
	       w[0] = v.x;
	       w[1] = v.y;
	       w[2] = v.z;
	       }
	       
	       v[0] = w[0]*m[0]+w[1]*m[4]+w[2]*m[8];
	       v[1] = w[0]*m[1]+w[1]*m[5]+w[2]*m[9];
	       v[2] = w[0]*m[2]+w[1]*m[6]+w[2]*m[10];
	       
	       // cerr << "v: " << v[0] << " " << v[1] << " " << v[2] << endl;
	       
	       angle =  secure_atan2(v[1],v[0]);
	       
	       }
	    */
	    
	    // cerr << "phi: " << phi << "  angle: " << angle << endl;
	    
	    const double pixel_length = fov/height();    
	    const int min_bar_length = MIN(60,width()/5); // pixels
	    const double coefficient = coefficient_angle_label(min_bar_length*pixel_length);
	    
	    /* 
	       glPushMatrix();
	       glLoadIdentity();
	       glTranslated(0,0,-zoom_scale_zero);
	       glRotated(angle*(180.0/pi),0,0,1);
	       const double s = 2.0*zoom_scale_zero*tan((pi/180.0)*coefficient/2.0);
	       // cerr << "scale: " << s << endl;
	       glScaled(s,s,1);
	       glColor3d(0.3,0.5,0.4);
	       // glLineStipple(1, 0x00FF);
	       glLineStipple(1, 0x5555);
	       glEnable(GL_LINE_STIPPLE);
	       gl2psEnable(GL2PS_LINE_STIPPLE);
	       glCallList(xy_grid); 
	       glDisable(GL_LINE_STIPPLE);
	       gl2psDisable(GL2PS_LINE_STIPPLE);
	       glPopMatrix();
	    */
	    
	    glPushMatrix();
	    // glLoadIdentity();
	    // glTranslated(0,0,-2.0*zoom_scale_zero);
	    // glRotated(angle*(180.0/pi),0,0,1);
	    //
	    // glTranslated(center.x,center.y,center.z); // fix this translation
	    // 
	    // NEEDED!!!
	    switch (universe->GetReferenceSystem()) {
	    case ECLIPTIC:
	      {
		// test
		// glRotated(180.0,0,0,1);
		
	     	/*
		  glRotated(-obleq(evol_frame).GetRad()*(180.0/pi)-delta_obleq(evol_frame).GetRad()*(180.0/pi),1,0,0);
		  glRotated(-delta_longitude(evol_frame).GetRad()*(180.0/pi),0,0,1);
		*/
		//
		// glRotated(-obleq_J2000().GetRad()*(180.0/pi)-delta_obleq_J2000().GetRad()*(180.0/pi),1,0,0);
		// glRotated(-delta_longitude_J2000().GetRad()*(180.0/pi),0,0,1);
		//
		
		// if translating in XOrsaOpenGL::resizeGL,
		// we have to translate back
		// glTranslated(0,0,distance); // maybe not...
		
		glRotated(-obleq_J2000().GetRad()*(180.0/pi),1,0,0);
		
		// rotate-back of theta/phi
		// glRotated(-(phi+pi)*(180.0/pi),0,0,1);
		// glRotated(-theta*(180.0/pi),0,1,0);
		// glRotated(-theta*(180.0/pi),0,1,0);
		// glRotated(-phi*(180.0/pi),0,0,1);
		
	      }
	      break;
	    case EQUATORIAL: break;
	    }
	    //
	    /* 
	       // Z
	       glColor3d(1.0,0.5,0.0);
	       glBegin(GL_LINES);
	       glVertex3d(0,0,0);
	       glVertex3d(0,0,zoom_scale_zero);
	       glEnd();
	       // X
	       glColor3d(0,1,0);
	       glBegin(GL_LINES);
	       glVertex3d(0,0,0);
	       glVertex3d(zoom_scale_zero,0,0);
	       glEnd();
	       // Y
	       glColor3d(1,0,0);
	       glBegin(GL_LINES);
	       glVertex3d(0,0,0);
	       glVertex3d(0,zoom_scale_zero,0);
	       glEnd();  
	    */
	    //
	    glGetDoublev(GL_MODELVIEW_MATRIX,m);
	    double inverted[16];
	    invert(m,inverted);
	    //
	    v[0] = 0.0;
	    v[1] = 0.0;
	    v[2] = 1.0;
	    //
	    //
	    w[0] = v[0]*inverted[0]+v[1]*inverted[4]+v[2]*inverted[8];
	    w[1] = v[0]*inverted[1]+v[1]*inverted[5]+v[2]*inverted[9];
	    w[2] = v[0]*inverted[2]+v[1]*inverted[6]+v[2]*inverted[10];
	    //
	    const double var_phi_2   = secure_atan2(w[1],w[0]);
	    const double var_theta_2 = secure_atan2(w[2],secure_sqrt(w[0]*w[0]+w[1]*w[1])); 
	    //
	    // cerr << "var_phi_2: " << var_phi_2*(180.0/pi) << "    var_theta_2: " << var_theta_2*(180.0/pi) << endl;
	    //
	    //
	    /* 
	       renderText(zoom_scale_zero*cos(phi)*cos(theta),
	       zoom_scale_zero*cos(phi)*sin(theta),
	       zoom_scale_zero*sin(phi),
	       ".location.",
	       default_font);
	    */
	    //
	    /* 
	       renderText(zoom_scale_zero*cos(-phi-pi)*cos(theta),
	       zoom_scale_zero*cos(-phi-pi)*sin(theta),
	       zoom_scale_zero*sin(-phi-pi),
	       ".location.",
	       default_font);
	    */
	    //
	    /* 
	       renderText(zoom_scale_zero*cos(var_theta)*cos(var_phi+pi),
	       zoom_scale_zero*cos(var_theta)*sin(var_phi+pi),
	       zoom_scale_zero*sin(var_theta),
	       ".location.",
	       default_font);
	    */
	    //
	    /* 
	       renderText(zoom_scale_zero*cos(-theta)*cos(phi+pi),
	       zoom_scale_zero*cos(-theta)*sin(phi+pi),
	       zoom_scale_zero*sin(-theta),
	       ".location.",
	       default_font);
	    */
	    // *******************************************
	    // GOOD ONE when obleq rotation is excluded
	    /* renderText(zoom_scale_zero*cos(-var_theta)*cos(var_phi+pi),
	       zoom_scale_zero*cos(-var_theta)*sin(var_phi+pi),
	       zoom_scale_zero*sin(-var_theta),
	       ".location.",
	       default_font);
	    */
	    // ********************************************
	    //
	    /* // FINAL, GOOD ONE!
	       renderText(zoom_scale_zero*cos(-var_theta_2)*cos(var_phi_2+pi),
	       zoom_scale_zero*cos(-var_theta_2)*sin(var_phi_2+pi),
	       zoom_scale_zero*sin(-var_theta_2),
	       ".location.",
	       default_font);
	    */
	    //
	    // const double s = 2.0*zoom_scale_zero*tan((pi/180.0)*coefficient/2.0);
	    // cerr << "scale: " << s << endl;
	    // glScaled(s,s,1);
	    glColor3d(0.3,0.5,0.4);
	    // glLineStipple(1, 0x00FF);
	    glLineStipple(1, 0x5555);
	    glEnable(GL_LINE_STIPPLE);
	    gl2psEnable(GL2PS_LINE_STIPPLE);
	    // glCallList(xy_grid); 
	    //   
	    {
	      
	      /* 
		 double center_x, center_y;
		 if (coefficient >= 1) {
		 center_x = (pi/180.0)*coefficient*rint((-var_theta_2)/coefficient*(180.0/pi));
		 center_y = (pi/180.0)*coefficient*rint((var_phi_2+pi)/coefficient*(180.0/pi));
		 }	else if (coefficient >= 1.0/60.0) {
		 center_x = (pi/180.0)*coefficient*rint((-var_theta_2)/coefficient*(180.0/pi));
		 center_y = (pi/180.0)*coefficient*rint((var_phi_2+pi)/coefficient*(180.0/pi));
		 } else {
		 center_x = (pi/180.0)*coefficient*rint((-var_theta_2)/coefficient*(180.0/pi));
		 center_y = (pi/180.0)*coefficient*rint((var_phi_2+pi)/coefficient*(180.0/pi));
		 }
	      */
	      
	      const double center_x = (pi/180.0)*coefficient*rint((-var_theta_2)/coefficient*(180.0/pi));
	      const double center_y = (pi/180.0)*coefficient*rint((var_phi_2+pi)/coefficient*(180.0/pi));
	      
	      /* 
		 int lines = 60.0/coefficient; // degrees
		 if (lines > 10) lines = 10;
	      */
	      
	      int lx,ly;
	      double cfx,cfy;
	      double cfxx,cfyy;
	      for (lx=-10;lx<=10;++lx) {
		for (ly=-10;ly<=10;++ly) {
		  /* 
		     cfx  = (pi/180.0)*coefficient*lx;
		     cfy  = (pi/180.0)*coefficient*ly;
		     cfxx = (pi/180.0)*coefficient*(lx+1);
		     cfyy = (pi/180.0)*coefficient*(ly+1);
		  */
		  //
		  /* 
		     cfx  = -phi+pi+ (pi/180.0)*coefficient*lx;
		     cfy  = theta+ (pi/180.0)*coefficient*ly;
		     cfxx = -phi+pi+ (pi/180.0)*coefficient*(lx+1);
		     cfyy = theta+ (pi/180.0)*coefficient*(ly+1);
		  */
		  //
		  /* 
		     cfx  = theta+ (pi/180.0)*coefficient*lx;
		     cfy  = -phi-pi+ (pi/180.0)*coefficient*ly;
		     cfxx = theta+ (pi/180.0)*coefficient*(lx+1);
		     cfyy = -phi-pi+ (pi/180.0)*coefficient*(ly+1);
		  */
		  //
		  /* 
		     cfx  = -var_theta + (pi/180.0)*coefficient*lx;
		     cfy  = var_phi+pi + (pi/180.0)*coefficient*ly;
		     cfxx = -var_theta + (pi/180.0)*coefficient*(lx+1);
		     cfyy = var_phi+pi + (pi/180.0)*coefficient*(ly+1);
		  */
		  // GENERAL GOOD ONE, without center...
		  /* 
		     cfx  = -var_theta_2 + (pi/180.0)*coefficient*lx;
		     cfy  = var_phi_2+pi + (pi/180.0)*coefficient*ly;
		     cfxx = -var_theta_2 + (pi/180.0)*coefficient*(lx+1);
		     cfyy = var_phi_2+pi + (pi/180.0)*coefficient*(ly+1);
		  */
		  // GENERAL GOOD ONE, with center...
		  cfx  = center_x + (pi/180.0)*coefficient*lx;
		  cfy  = center_y + (pi/180.0)*coefficient*ly;
		  cfxx = center_x + (pi/180.0)*coefficient*(lx+1);
		  cfyy = center_y + (pi/180.0)*coefficient*(ly+1);
		  //
		  /* 
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*sin(cfx),
		     zoom_scale_zero*cos(cfx)*sin(cfy),
		     zoom_scale_zero*cos(cfx)*cos(cfy));
		     glVertex3d(zoom_scale_zero*sin(cfxx),
		     zoom_scale_zero*cos(cfxx)*sin(cfy),
		     zoom_scale_zero*cos(cfxx)*cos(cfy));
		     glEnd();
		     //
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*sin(cfx),
		     zoom_scale_zero*cos(cfx)*sin(cfy),
		     zoom_scale_zero*cos(cfx)*cos(cfy));
		     glVertex3d(zoom_scale_zero*sin(cfx),
		     zoom_scale_zero*cos(cfx)*sin(cfyy),
		     zoom_scale_zero*cos(cfx)*cos(cfyy));
		     glEnd();
		  */
		  //
		  /* 
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*sin(cfx),
		     zoom_scale_zero*cos(cfx)*sin(cfy),
		     zoom_scale_zero*cos(cfx)*cos(cfy));
		     glVertex3d(zoom_scale_zero*sin(cfxx),
		     zoom_scale_zero*cos(cfxx)*sin(cfy),
		     zoom_scale_zero*cos(cfxx)*cos(cfy));
		     glEnd();
		     //
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*sin(cfx),
		     zoom_scale_zero*cos(cfx)*sin(cfy),
		     zoom_scale_zero*cos(cfx)*cos(cfy));
		     glVertex3d(zoom_scale_zero*sin(cfx),
		     zoom_scale_zero*cos(cfx)*sin(cfyy),
		     zoom_scale_zero*cos(cfx)*cos(cfyy));
		     glEnd();
		  */
		  //
		  /* 
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*cos(cfx)*cos(cfy),
		     zoom_scale_zero*cos(cfx)*sin(cfy),
		     zoom_scale_zero*sin(cfx));
		     glVertex3d(zoom_scale_zero*cos(cfxx)*cos(cfy),
		     zoom_scale_zero*cos(cfxx)*sin(cfy),
		     zoom_scale_zero*sin(cfxx));
		     glEnd();
		     //
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*cos(cfx)*cos(cfy),
		     zoom_scale_zero*cos(cfx)*sin(cfy),
		     zoom_scale_zero*sin(cfx));
		     glVertex3d(zoom_scale_zero*cos(cfx)*cos(cfyy),
		     zoom_scale_zero*cos(cfx)*sin(cfyy),
		     zoom_scale_zero*sin(cfx));
		     glEnd();
		  */
		  //
		  glBegin(GL_LINES);
		  glVertex3d((zoom_scale_zero+distance)*cos(cfx)*cos(cfy),
			     (zoom_scale_zero+distance)*cos(cfx)*sin(cfy),
			     (zoom_scale_zero+distance)*sin(cfx));
		  glVertex3d((zoom_scale_zero+distance)*cos(cfxx)*cos(cfy),
			     (zoom_scale_zero+distance)*cos(cfxx)*sin(cfy),
			     (zoom_scale_zero+distance)*sin(cfxx));
		  glEnd();
		  //
		  glBegin(GL_LINES);
		  glVertex3d((zoom_scale_zero+distance)*cos(cfx)*cos(cfy),
			     (zoom_scale_zero+distance)*cos(cfx)*sin(cfy),
			     (zoom_scale_zero+distance)*sin(cfx));
		  glVertex3d((zoom_scale_zero+distance)*cos(cfx)*cos(cfyy),
			     (zoom_scale_zero+distance)*cos(cfx)*sin(cfyy),
			     (zoom_scale_zero+distance)*sin(cfx));
		  glEnd();
		  //
		  /* 
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*sin(cfy)*cos(cfx),
		     zoom_scale_zero*sin(cfy)*sin(cfx),
		     zoom_scale_zero*cos(cfy));
		     glVertex3d(zoom_scale_zero*sin(cfy)*cos(cfxx),
		     zoom_scale_zero*sin(cfy)*sin(cfxx),
		     zoom_scale_zero*cos(cfy));
		     glEnd();
		     //
		     glBegin(GL_LINES);
		     glVertex3d(zoom_scale_zero*sin(cfy)*cos(cfx),
		     zoom_scale_zero*sin(cfy)*sin(cfx),
		     zoom_scale_zero*cos(cfy));
		     glVertex3d(zoom_scale_zero*sin(cfyy)*cos(cfx),
		     zoom_scale_zero*sin(cfyy)*sin(cfx),
		     zoom_scale_zero*cos(cfyy));
		     glEnd();
		  */
		}
	      }
	    }
	    //
	    glDisable(GL_LINE_STIPPLE);
	    gl2psDisable(GL2PS_LINE_STIPPLE);
	    glPopMatrix();
	    
	    
	  }
	  
	}
      }
    }
  } 
  
  glPopMatrix();
}

void XOrsaOpenGL::internal_draw_direction_line(Vector &dir_pos) {
  
  const double phi   = secure_atan2(dir_pos.y,dir_pos.x);
  const double theta = secure_atan2(dir_pos.z,secure_sqrt(dir_pos.x*dir_pos.x+dir_pos.y*dir_pos.y)); 
  
  glRotated(-theta*(180.0/pi),0.0,1.0,0.0);
  glRotated(180.0-phi*(180.0/pi),0.0,0.0,1.0);
  
  if (projection == OGL_PERSPECTIVE) return; // no line and label,,,
  
  glColor3d(0,0,0.8);
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  glLineStipple(1, 0x00FF);
  glEnable(GL_LINE_STIPPLE);
  gl2psEnable(GL2PS_LINE_STIPPLE);
  glBegin(GL_LINES);
  glVertex3d(0,0,0);
  glVertex3d(dir_pos.x, dir_pos.y, dir_pos.z);
  glEnd();
  glDisable(GL_LINE_STIPPLE);
  gl2psDisable(GL2PS_LINE_STIPPLE);
  
  // label
  char t[1024];
  
  length_unit lu = AutoLengthUnit(dir_pos.Length());
  sprintf(t," %.3g %s",FromUnits(dir_pos.Length(),lu,-1),units->label(lu).c_str());
  
  Vector m = dir_pos/2;
  renderText(m.x,m.y,m.z,t,default_font); 
}
    
void XOrsaOpenGL::internal_draw_Lagrange_points(Vector &dir_pos_mod) {
  
  const double m_1 = evol_frame[lagrange_body_1].mass();
  const double m_2 = evol_frame[lagrange_body_2].mass();
  
  // assume m1 > m2
  double m1,m2;
  Vector tmp_x,tmp_y;
  Vector x1;
  if (m_1>m_2) {
    m1 = m_1;
    m2 = m_2;
    x1 = - dir_pos_mod - evol_frame[lagrange_body_1].position();   
    tmp_x  = evol_frame[lagrange_body_2].position() - evol_frame[lagrange_body_1].position();
    tmp_y  = evol_frame[lagrange_body_2].velocity() - evol_frame[lagrange_body_1].velocity();
    tmp_y /= tmp_y.Length();
    double tmp = tmp_y*tmp_x/tmp_x.Length();
    tmp_y -= tmp*tmp_x/tmp_x.Length();
    tmp_y /= tmp_y.Length();
    tmp_y *= tmp_x.Length();
    // cerr << " diff: " << tmp << " --> " << tmp_y*tmp_x << endl;
  } else {
    m1 = m_2;
    m2 = m_1;
    x1 = - dir_pos_mod - evol_frame[lagrange_body_2].position();
    tmp_x  = evol_frame[lagrange_body_1].position() - evol_frame[lagrange_body_2].position();
    tmp_y  = evol_frame[lagrange_body_1].velocity() - evol_frame[lagrange_body_2].velocity();
    tmp_y /= tmp_y.Length();
    double tmp = tmp_y*tmp_x/tmp_x.Length();
    tmp_y -= tmp*tmp_x/tmp_x.Length();
    tmp_y /= tmp_y.Length();
    tmp_y *= tmp_x.Length();
    // cerr << " diff: " << tmp << " --> " << tmp_y*tmp_x << endl;
  }
  
  const Vector x=tmp_x;
  const Vector y=tmp_y;
  
  const double mu_1  = m1/(m1+m2); 
  const double mu_2  = m2/(m1+m2); 
  const double mu_21 = mu_2/mu_1;
  const double alpha = secure_pow(mu_2/(3*mu_1),1.0/3.0);
  
  double r1,r2;
  
  // L1
  r2 = alpha - (1.0/3.0)*secure_pow(alpha,2) - (1.0/9.0)*secure_pow(alpha,3) - (23.0/81.0)*secure_pow(alpha,4); 
  r1 = 1.0 - r2;
  const Vector L1 = r1 * x - x1;
  
  // L2 
  r2 = alpha + (1.0/3.0)*secure_pow(alpha,2) - (1.0/9.0)*secure_pow(alpha,3) - (31.0/81.0)*secure_pow(alpha,4); 
  r1 = 1.0 + r2;
  const Vector L2 = r1 * x - x1;
  
  // L3
  const double beta = - (7.0/12.0)*mu_21 + (7.0/12.0)*secure_pow(mu_21,2) - (13223.0/20736.0)*secure_pow(mu_21,3);
  r1 = 1.0 + beta;
  r2 = 2.0 + beta;
  const Vector L3 = - r1 * x - x1;
  
  // L4
  const Vector L4 = (0.5-mu_2)*x + secure_sqrt(3.0)/2.0*y - x1;
  
  // L5
  const Vector L5 = (0.5-mu_2)*x - secure_sqrt(3.0)/2.0*y - x1;
  
  // render Lagrangian points
  string L_name;
  
  // glColor3d(0,0.7,0);
  // glColor3d(0.6,0.4,0.0);
  glColor3d(1.0,0.6,0.0);
  glPointSize(2);
  gl2psPointSize(2);
  
  QString text;
  
  L_name = (" L1");
  glBegin(GL_POINTS);
  glVertex3d(L1.x, L1.y, L1.z);
  glEnd();
  // glRasterPos3d(L1.x,L1.y,L1.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  renderText(L1.x, L1.y, L1.z,text,default_font); 
  
  L_name = (" L2");
  glBegin(GL_POINTS);
  glVertex3d(L2.x, L2.y, L2.z);
  glEnd();
  // glRasterPos3d(L2.x,L2.y,L2.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  renderText(L2.x, L2.y, L2.z,text,default_font); 
  
  L_name = (" L3");
  glBegin(GL_POINTS);
  glVertex3d(L3.x, L3.y, L3.z);
  glEnd();
  // glRasterPos3d(L3.x,L3.y,L3.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  renderText(L3.x, L3.y, L3.z,text,default_font); 
  
  L_name = (" L4");
  glBegin(GL_POINTS);
  glVertex3d(L4.x, L4.y, L4.z);
  glEnd();
  // glRasterPos3d(L4.x,L4.y,L4.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  renderText(L4.x, L4.y, L4.z,text,default_font); 
  
  L_name = (" L5");
  glBegin(GL_POINTS);
  glVertex3d(L5.x, L5.y, L5.z);
  glEnd();
  // glRasterPos3d(L5.x,L5.y,L5.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  renderText(L5.x, L5.y, L5.z,text,default_font); 
}

void XOrsaOpenGL::internal_draw_reference_system() {
  
  glPushMatrix();
  
  glColor3d(0,0,0.8);
  
  // sync with the scale_bar code!     
  double pixel_length = 2*fov/width();
  if (height()<width()) pixel_length = 2*fov/height();
  
  const int min_bar_length = MIN(60,width()/5); // pixels
  length_unit lu = AutoLengthUnit(coefficient_scale_label(min_bar_length*pixel_length));
  const double coefficient = FromUnits(coefficient_scale_label(FromUnits(min_bar_length*pixel_length,lu,-1)),lu);
  
  glScaled(coefficient,coefficient,coefficient);
  
  glCallList(reference_system); 
  glPopMatrix();
}

void XOrsaOpenGL::internal_draw_xy_grid() {
  
  // sync with the scale_bar code!     
  double pixel_length = 2*fov/width();
  if (height()<width()) pixel_length = 2*fov/height();
  
  const int min_bar_length = MIN(60,width()/5); // pixels
  length_unit lu = AutoLengthUnit(coefficient_scale_label(min_bar_length*pixel_length));
  const double coefficient = FromUnits(coefficient_scale_label(FromUnits(min_bar_length*pixel_length,lu,-1)),lu);
  
  glPushMatrix();
  glColor3d(0.3,0.5,0.4);
  glScaled(coefficient,coefficient,coefficient);
  glCallList(xy_grid); 
  glPopMatrix();
}

void XOrsaOpenGL::internal_draw_z_bars() {
  Vector p;
  unsigned int j;
  glColor3d(0.7,0.7,0.2);
  for (j=0;j<evol_frame.size();j++) {
    p  = evol_frame[j].position();
    p -= center;
    glPushMatrix();
    glBegin(GL_LINES);
    glVertex3d(p.x, p.y, p.z);
    glVertex3d(p.x, p.y, 0.0);
    glEnd();
    glPopMatrix();
  }
}

void XOrsaOpenGL::internal_draw_MOID() {
  
#ifdef HAVE_GSL	
  if (DrawOrbits())
    if (evolution != 0)
      if (evolution->size() != 0)
     	if ((evol_frame.size() >= 3) && (moid_body_1!=moid_body_2)) {
	  
	  if (orbit_reference_body_index==-HEX_AUTO) {
	    if (orbit_reference_body_index_vector[evol_counter].size()==0) return;
	    if (moid_body_1==orbit_reference_body_index_vector[evol_counter][moid_body_1]) return;
	    if (moid_body_2==orbit_reference_body_index_vector[evol_counter][moid_body_2]) return;
	  } else {
	    if (moid_body_1==orbit_reference_body_index) return;
	    if (moid_body_2==orbit_reference_body_index) return;
	  }
	  
	  Orbit  o_1, o_2;
	  Vector r_1, r_2;
	  
	  if (orbit_reference_body_index==-HEX_AUTO) {
	    o_1.Compute(evol_frame[moid_body_1],evol_frame[orbit_reference_body_index_vector[evol_counter][moid_body_1]]);
	    o_2.Compute(evol_frame[moid_body_2],evol_frame[orbit_reference_body_index_vector[evol_counter][moid_body_2]]);
	  } else {
	    o_1.Compute(evol_frame[moid_body_1],evol_frame[orbit_reference_body_index]);
	    o_2.Compute(evol_frame[moid_body_2],evol_frame[orbit_reference_body_index]);
	  }
	  
	  // really needed the eccentricity check?	
	  // if ((o_1.e < 1.0) && (o_2.e < 1.0)) {
	  
	  double moid;
	  // moid = MOID(o_1,o_2,r_1,r_2);
	  if ((orbit_reference_body_index==-HEX_AUTO) && (orbit_reference_body_index_vector[evol_counter][moid_body_1]!=orbit_reference_body_index_vector[evol_counter][moid_body_2])) { 
	    moid = MOID2RB(evol_frame[orbit_reference_body_index_vector[evol_counter][moid_body_1]].position(),
			   evol_frame[orbit_reference_body_index_vector[evol_counter][moid_body_2]].position(),
			   o_1,o_2,r_1,r_2);
	  } else {
	    moid = MOID(o_1,o_2,r_1,r_2);
	  }	
	  
	  // cerr << "MOID between bodies 1 and 2: " << moid << endl;
	  // cerr << "r1: " << r1.x << "  " << r1.y << "  " << r1.z << endl; 
	  // cerr << "r2: " << r2.x << "  " << r2.y << "  " << r2.z << endl; 
	  
	  Vector z1,z2;
	  if (orbit_reference_body_index==-HEX_AUTO) {
	    z1 = evol_frame[orbit_reference_body_index_vector[evol_counter][moid_body_1]].position() - center;
	    z2 = evol_frame[orbit_reference_body_index_vector[evol_counter][moid_body_2]].position() - center;
	  } else {
	    z1 = z2 = evol_frame[orbit_reference_body_index].position() - center;
	  }
	  
	  r_1 += z1;
	  r_2 += z2;
	  
	  glColor3d(0,1,1);
	  glLineWidth(1.0);
	  gl2psLineWidth(1.0);
	  // glLineStipple(1, 0x00FF);
	  // glEnable(GL_LINE_STIPPLE);
	  glBegin(GL_LINES);
	  glVertex3d(r_1.x,r_1.y,r_1.z);
	  glVertex3d(r_2.x,r_2.y,r_2.z);
	  glEnd();
	  // glDisable(GL_LINE_STIPPLE);
	  
	  char t[1024];
	  // sprintf(t," MOID: %.3g %s",moid,LengthLabel().c_str());
	  
	  length_unit lu = AutoLengthUnit(moid);
	  // sprintf(t," MOID: %.3g %s",moid,LengthLabel().c_str());
	  sprintf(t," MOID: %.3g %s",FromUnits(moid,lu,-1),units->label(lu).c_str());
	  
	  // string dpy_name = t;
	  Vector m = (r_1+r_2)/2;
	  // glRasterPos3d(m.x,m.y,m.z);
	  // glCallLists(dpy_name.size(),GL_BYTE,(GLubyte*) dpy_name.c_str());
	  renderText(m.x,m.y,m.z,t,default_font); 
	  
	  // } 
	}
#endif // HAVE_GSL  
}

void XOrsaOpenGL::internal_draw_bodies_and_labels() {
  
  Vector p;
  unsigned int j;
  for (j=0;j<evol_frame.size();j++) {
    
    glPushMatrix();
    
    if ( (projection==OGL_PERSPECTIVE) && LightTimeCorrection()) {
      // simple, linear approximation for the moment, to be improved and checked
      p = evol_frame[j].position() - evol_frame[j].velocity()*fabs((evol_frame[j].position()-center).Length()/GetC());
      
      /* 
	 fprintf(stderr,"LT: object: %s   distance = %g [AU]   time = %g [s]   velocity = %g [km/s]\n",
	 evol_frame[j].name().c_str(),
	 FromUnits((evol_frame[j].position()-center).Length(),AU,-1),
	 FromUnits((evol_frame[j].position()-center).Length()/GetC(),SECOND,-1),
	 FromUnits(FromUnits(evol_frame[j].velocity().Length(),KM,-1),SECOND));
      */
      
    } else {
      p = evol_frame[j].position();
    }
    
    p -= center;
    if (evol_frame[j].mass() != 0) {
      glPointSize(1);
      gl2psPointSize(1);
      glColor3d(1,0,0);
    } else {
      glPointSize(1);
      gl2psPointSize(1);
      glColor3d(0,0,1);
    }
    
    if (evol_frame[j].radius() > 0) {
      // debug  
      /* char t[1024];
	 length_unit lu = AutoLengthUnit((*evol_it)[j].radius());
	 sprintf(t,"%s radius: %.3g %s",(*evol_it)[j].name().c_str(),FromUnits((*evol_it)[j].radius(),lu,-1),units->label(lu).c_str());
	 cerr << t << endl;
      */
      //
      
      glTranslated(p.x, p.y, p.z);
      
      // Should rotate any planet...
      if (evol_frame[j].name() == "Earth") {
	
	//
	// glRotated(-(180.0/pi)*obleq_J2000().GetRad(),1,0,0);
	
	//
	// glRotated((180.0/pi)*gmst(evol_frame).GetRad(),0,0,1);
	
	/*
	  renderText(0.0,0.0,evol_frame[j].radius(), "N",   default_font); 
	  renderText(0.0,0.0,-evol_frame[j].radius(),"S",   default_font); 
	  renderText(evol_frame[j].radius(),0.0,0.0, "Eq-X",default_font); 
	*/
	
	// non-clipped Earth
      	/* 
	   glRotated(-(180.0/pi)*obleq_J2000().GetRad(),1,0,0);
	   glRotated((180.0/pi)*gmst(evol_frame).GetRad(),0,0,1);
	   //
	   glPushMatrix();
	   glScaled(evol_frame[j].radius(),evol_frame[j].radius(),evol_frame[j].radius());
	   glColor3d(0.1,0.2,0.9);
	   glCallList(earth);
	   glPopMatrix();
	   glLineWidth(1.0);
	*/
	
	// search Sun position
	Vector sunpos;
	{
	  unsigned int q=0;
	  while (q<evol_frame.size()) {
	    if (evol_frame[q].name() == "Sun") {
	      sunpos = evol_frame[q].position();
	      break;
	    }
	    ++q;
	  }
	}
	sunpos -= evol_frame[j].position(); // sunpos relative to Earth position
	// const double sl = sunpos.Length();
	sunpos.Normalize();
	
	const GLdouble equation_sun_plus[4]  = { sunpos.x, sunpos.y, sunpos.z,0.0};
	const GLdouble equation_sun_minus[4] = {-sunpos.x,-sunpos.y,-sunpos.z,0.0};
	//
	glClipPlane(GL_CLIP_PLANE0,equation_sun_plus);
	glClipPlane(GL_CLIP_PLANE1,equation_sun_minus);
	
	// viewing point
	double m[16];
	glGetDoublev(GL_MODELVIEW_MATRIX,m);
	//
	invert(m,inverted);
	//
	double v[3];
	//
	v[0] = 0.0;
	v[1] = 0.0;
	v[2] = 1.0;
	//
	double w[3];
	//
	w[0] = v[0]*inverted[0]+v[1]*inverted[4]+v[2]*inverted[8];
	w[1] = v[0]*inverted[1]+v[1]*inverted[5]+v[2]*inverted[9];
	w[2] = v[0]*inverted[2]+v[1]*inverted[6]+v[2]*inverted[10];
	//
	// cerr << "w[0]: " << w[0] << "   w[1]: " << w[1] << "   w[2]: " << w[2] << endl;
	//
	// const GLdouble equation_viewer[4]  = {w[0], w[1], w[2], 0.0};
	GLdouble equation_viewer[4] = {w[0], w[1], w[2], 0.0};
	if (projection == OGL_PERSPECTIVE) {
	  // TO BE FIXED: use distance, dir_pos.Length() or what?
	  equation_viewer[3] = -evol_frame[j].radius()*evol_frame[j].radius()/distance; // R*R/d
	  // equation_viewer[3] = -evol_frame[j].radius()*evol_frame[j].radius()/dir_pos.Length(); // R*R/d
	}
	//
	glClipPlane(GL_CLIP_PLANE2,equation_viewer);
	
	glEnable(GL_CLIP_PLANE2);
	glDepthFunc(GL_ALWAYS);
	//
	glPushMatrix();
	glScaled(evol_frame[j].radius(),evol_frame[j].radius(),evol_frame[j].radius());
	// internal sphere 
	glColor3d(0.1,0.1,0.1);   
	GLUquadricObj *qobj = gluNewQuadric();
       	gluQuadricDrawStyle(qobj,GLU_FILL);
	// gluSphere(qobj,0.99,36,18);
	// gluSphere(qobj,0.99,72,36);
	gluSphere(qobj,1.0,72,36);
	gluDeleteQuadric(qobj);
	// DAY
	glColor3d(0.5,0.5,0.2); 
	glEnable(GL_CLIP_PLANE0);
	glPushMatrix();
	glRotated(-(180.0/pi)*obleq_J2000().GetRad(),1,0,0);
	glRotated((180.0/pi)*gmst(evol_frame).GetRad(),0,0,1);
	glCallList(earth);
	glPopMatrix();
	glDisable(GL_CLIP_PLANE0);
	// NIGHT
	glColor3d(0.1,0.2,0.9);
	glEnable(GL_CLIP_PLANE1);
	glPushMatrix();
	glRotated(-(180.0/pi)*obleq_J2000().GetRad(),1,0,0);
	glRotated((180.0/pi)*gmst(evol_frame).GetRad(),0,0,1);
	glCallList(earth);
	glPopMatrix();
	glDisable(GL_CLIP_PLANE1);
	//
	glPopMatrix();
	glLineWidth(1.0);
	//
	glDepthFunc(GL_LESS);
	// glDepthFunc(GL_LEQUAL);
	glDisable(GL_CLIP_PLANE2);
	
	// observer locations
	if (0) {
	  Config conf;
	  OrsaConfigFile ocf(&conf);
	  ocf.Read();
	  const string obscode_path = conf.paths[OBSCODE]->GetValue().c_str();
	  
	  LocationFile lf;
	  lf.SetFileName(obscode_path);
	  lf.Read();
	  
	  // cerr << "locations read: " << lf.codes.size() << endl;
	  
	  if (lf.codes.size()) {
	    
	    glPushMatrix();
	    // glEnable(GL_CLIP_PLANE2);
	    glColor3d(0.2,0.8,0.2); 
	    glRotated(-(180.0/pi)*obleq_J2000().GetRad(),1,0,0);
	    glRotated((180.0/pi)*gmst(evol_frame).GetRad(),0,0,1);
	    
	    double x,y,z;
	    //
	    list<string>::iterator itr;
	    itr = lf.codes.begin();
	    //
	    while (itr != lf.codes.end()) {
	      
	      if ( (lf.locations[(*itr)].lon == 0.0) && 
		   (lf.locations[(*itr)].pz  == 0.0) &&
		   (lf.locations[(*itr)].pxy == 0.0) ) {
		// "bad" location
	      } else {    
		x = lf.locations[(*itr)].pxy*cos(lf.locations[(*itr)].lon);
		y = lf.locations[(*itr)].pxy*sin(lf.locations[(*itr)].lon);
		z = lf.locations[(*itr)].pz;
		
		// name
		// renderText(x,y,z,lf.locations[(*itr)].name,default_font); 
		
		// code
		renderText(x,y,z,(*itr),default_font); 
	      }
	      
	      ++itr;
	    }     
	    
	    // glDisable(GL_CLIP_PLANE2);
	    glPopMatrix();
	  }
	}
	
      } else if (evol_frame[j].name() == "Moon") {
	
	// search Sun position
	Vector sunpos;
	{
	  unsigned int q=0;
	  while (q<evol_frame.size()) {
	    if (evol_frame[q].name() == "Sun") {
	      sunpos = evol_frame[q].position();
	      break;
	    }
	    ++q;
	  }
	}
	sunpos -= evol_frame[j].position(); // sunpos relative to Earth position
	// const double sl = sunpos.Length();
	sunpos.Normalize();
	
	const GLdouble equation_sun_plus[4]  = { sunpos.x, sunpos.y, sunpos.z,0.0};
	const GLdouble equation_sun_minus[4] = {-sunpos.x,-sunpos.y,-sunpos.z,0.0};
	//
	glClipPlane(GL_CLIP_PLANE0,equation_sun_plus);
	glClipPlane(GL_CLIP_PLANE1,equation_sun_minus);
	
	// viewing point
	/* double m[16];
	   glGetDoublev(GL_MODELVIEW_MATRIX,m);
	   //
	   invert(m,inverted);
	   //
	   double v[3];
	   //
	   v[0] = 0.0;
	   v[1] = 0.0;
	   v[2] = 1.0;
	   //
	   double w[3];
	   //
	   w[0] = v[0]*inverted[0]+v[1]*inverted[4]+v[2]*inverted[8];
	   w[1] = v[0]*inverted[1]+v[1]*inverted[5]+v[2]*inverted[9];
	   w[2] = v[0]*inverted[2]+v[1]*inverted[6]+v[2]*inverted[10];
	   //
	   // cerr << "w[0]: " << w[0] << "   w[1]: " << w[1] << "   w[2]: " << w[2] << endl;
	   //
	   const GLdouble equation_viewer[4]  = {w[0], w[1], w[2], 0.0};
	   //
	   glClipPlane(GL_CLIP_PLANE2,equation_viewer);
	*/
	
	// glEnable(GL_CLIP_PLANE2);
	//
	glPushMatrix();
       	// DAY
	glColor3d(0.7,0.7,0.7); 
	glEnable(GL_CLIP_PLANE0);
	glPushMatrix();
	{	
	  GLUquadricObj *qobj;
	  qobj = gluNewQuadric();
	  // gluQuadricDrawStyle(qobj,GLU_SILHOUETTE);
	  gluQuadricDrawStyle(qobj,GLU_FILL);
	  // gluQuadricNormals(qobj,GLU_NONE);
	  // gluSphere(qobj,evol_frame[j].radius(),10,10);
	  gluSphere(qobj,evol_frame[j].radius(),360,180);
	  // gluSphere(qobj,evol_frame[j].radius(),144,180);
	  gluDeleteQuadric(qobj);
	}
	glPopMatrix();
	glDisable(GL_CLIP_PLANE0);
	// NIGHT
	glColor3d(0.2,0.2,0.2);
	glEnable(GL_CLIP_PLANE1);
	glPushMatrix();
	{
	  GLUquadricObj *qobj;
	  qobj = gluNewQuadric();
	  // gluQuadricDrawStyle(qobj,GLU_SILHOUETTE);
	  gluQuadricDrawStyle(qobj,GLU_FILL);
	  // gluQuadricNormals(qobj,GLU_NONE);
	  // gluSphere(qobj,evol_frame[j].radius(),10,10);
	  gluSphere(qobj,evol_frame[j].radius(),360,180);
	  // gluSphere(qobj,evol_frame[j].radius(),144,180);
	  gluDeleteQuadric(qobj);
	}
	glPopMatrix();
	glDisable(GL_CLIP_PLANE1);
	//
	glPopMatrix();
      	//
	// glDisable(GL_CLIP_PLANE2);
	
      } else if (evol_frame[j].name() == "Sun") {
	
	glColor3d(0.9,0.9,0.1);	
	GLUquadricObj *qobj;
	qobj = gluNewQuadric();
	gluQuadricDrawStyle(qobj,GLU_FILL);
	gluQuadricNormals(qobj,GLU_NONE);
     	gluSphere(qobj,evol_frame[j].radius(),36,18);
	gluDeleteQuadric(qobj);
	
      } else {
	
	GLUquadricObj *qobj;
	qobj = gluNewQuadric();
	gluQuadricDrawStyle(qobj,GLU_SILHOUETTE);
	// gluQuadricDrawStyle(qobj,GLU_FILL);
	gluQuadricNormals(qobj,GLU_NONE);
	// gluSphere(qobj,evol_frame[j].radius(),10,10);
	gluSphere(qobj,evol_frame[j].radius(),36,18);
	// gluSphere(qobj,evol_frame[j].radius(),144,180);
	gluDeleteQuadric(qobj);
	
      }	
      
      // draw the point too
      glBegin(GL_POINTS);
      glVertex3d(0,0,0);
      glEnd();
    } else {
      glBegin(GL_POINTS);
      glVertex3d(p.x, p.y, p.z);
      glEnd();
    }
    
    glPopMatrix();
    
    if (bool_draw_labels) {
      if (((evol_frame[j].mass() > 0) && (labels_mode == MASSIVE)) || 
	  (labels_mode == ALL)) {
	glColor3d(0,0.7,0);
	
       	QString text;
	text.sprintf(" %s",evol_frame[j].name().c_str());
	renderText(p.x, p.y, p.z, text,default_font); 
      }
    }
  }
  
  /* 
     if (0) {
     // test!
     QChar uc_mu(0x03BC);
     QChar uc_nu(0x03BD);
     // QString s_F_mn("F<sub></sub>");
     QString s_F_mn("F           ");
     // QString s_F_mn("Origin...");
     s_F_mn.insert(6,uc_mu);
     s_F_mn.insert(7,uc_nu);
     // QFont font(default_font);
     QFont font;
     font.setStyleStrategy(QFont::PreferBitmap); // NEEDED!! Otherwise TrueType fonts can be selected and the program will crash (at least using Qt 3.1.1 under X11)
     renderText(0.0,0.0,0.0,s_F_mn,font);
     }
  */
  
  /* 
     // test: If watching from Earth, plot sky grids
     if ( (central_body_index >= 0) && 
     (central_body_index < (int)evol_frame.size()) ) {
     if ( (projection==OGL_PERSPECTIVE) &&
     (evol_frame[central_body_index].name() == "Earth") ) {
     
     // cerr << "looking from earth!!!!!" << endl;
     
     { // RA/DEC 
     
     glPushMatrix();
     
     glColor3d(1,1,1);
     //
     // cerr << "obleq: " << (180/pi)*obleq(evol_frame).GetRad() << endl;
     //
     glRotated((180/pi)*obleq(evol_frame).GetRad(),1,0,0);
     //
     GLUquadricObj *qobj;
     qobj = gluNewQuadric();
     gluQuadricDrawStyle(qobj,GLU_SILHOUETTE);
     // gluQuadricDrawStyle(qobj,GLU_FILL);
     gluQuadricNormals(qobj,GLU_NONE);
     // gluSphere(qobj,evol_frame[j].radius(),10,10);
     gluSphere(qobj,zoom_scale_zero,36,18);
     gluDeleteQuadric(qobj);
     
     glPopMatrix();
     }
     
     }
     }
  */
  
  // another test
  /* 
     if ( (central_body_index >= 0) && 
     (central_body_index < (int)evol_frame.size()) ) {
     if ( (projection==OGL_PERSPECTIVE) &&
     (evol_frame[central_body_index].name() == "Earth") ) {
     const double pixel_length = fov/height();   
     const int min_bar_length = MIN(60,width()/5); // pixels
     double coefficient = coefficient_angle_label(min_bar_length*pixel_length);
     double bar_length = coefficient / pixel_length;
     }
     }
  */
  
}

void XOrsaOpenGL::internal_draw_track() {
  
  glColor3d(0.7,0.7,0.4);
  glPointSize(1);
  gl2psPointSize(1);
  
  const int skip_track_frames = 1; // > 0
  const int max_frames = 200;
  int count = 0;
  unsigned int j;
  Vector p;
  
  const double phi_zero     = secure_atan2(dir_pos_zero.y,dir_pos_zero.x);
  const double theta_zero   = secure_atan2(dir_pos_zero.z,secure_sqrt(dir_pos_zero.x*dir_pos_zero.x+dir_pos_zero.y*dir_pos_zero.y)); 
  
  if ( UseDirectionBody() && 
       (direction_body_index < (int)evol_frame.size()) &&
       ( (central_body_index == -HEX_ORIGIN) || 
	 (central_body_index == -HEX_CENTER_OF_MASS) || 
	 (central_body_index < (int)evol_frame.size()) ) && 
       (!AbsoluteTracks())) {
    
    glPushMatrix(); 
    
    glRotated(phi_zero*(180/pi)-180,0,0,1);
    glRotated(theta_zero*(180/pi),0,1,0);
    
    Vector tmp_center, tmp_dir_pos;
    
    int local_counter = evol_counter;
    // Frame *f;
    
    while (local_counter >= 0) {
      
      const Frame * f = &((*evolution)[local_counter]);
      
      if (central_body_index == -HEX_ORIGIN)
	tmp_center.Set(0,0,0);
      else if (central_body_index == -HEX_CENTER_OF_MASS)
	tmp_center = f->CenterOfMass();
      else
	tmp_center = (*f)[central_body_index].position();
      
      if (direction_body_index == -HEX_ORIGIN)
	tmp_dir_pos.Set(0,0,0);
      else if (direction_body_index == -HEX_CENTER_OF_MASS)
	tmp_dir_pos = f->CenterOfMass();
      else
	tmp_dir_pos = (*f)[direction_body_index].position();
      //
      tmp_dir_pos -= tmp_center;
      
      if (! tmp_dir_pos.IsZero()) {
	
	const double phi   = secure_atan2(tmp_dir_pos.y,tmp_dir_pos.x);
	const double theta = secure_atan2(tmp_dir_pos.z,secure_sqrt(tmp_dir_pos.x*tmp_dir_pos.x+tmp_dir_pos.y*tmp_dir_pos.y)); 
	
	glPushMatrix(); 
	
	glRotated(-theta*(180/pi),0,1,0);
	glRotated(180-phi*(180/pi),0,0,1);
	
	for (j=0;j<f->size();j++) {
	  
	  if (central_body_index == -HEX_ORIGIN)
	    p = (*f)[j].position();
	  else if (central_body_index == -HEX_CENTER_OF_MASS)
	    p = (*f)[j].position() - f->CenterOfMass();
	  else
	    p = (*f)[j].position() - (*f)[central_body_index].position();
	  
	  glBegin(GL_POINTS);
	  glVertex3d(p.x,p.y,p.z);
	  glEnd();
	}
	
	glPopMatrix();
	
      }
      
      local_counter -= skip_track_frames;
      ++count;
      if (count == max_frames) break;
    }
    
    glPopMatrix();
    
  } else {
    
    int local_counter = evol_counter;
    // Frame *f;
    
    while (local_counter >= 0) {
      
      const Frame * f = &((*evolution)[local_counter]);
      
      for (j=0;j<f->size();j++) {
	
	if (central_body_index == -HEX_ORIGIN)
	  p = (*f)[j].position();
	else if (central_body_index == -HEX_CENTER_OF_MASS)
	  p = (*f)[j].position() - f->CenterOfMass();
	else
	  p = (*f)[j].position() - (*f)[central_body_index].position();
	
	glBegin(GL_POINTS);
	glVertex3d(p.x,p.y,p.z);
	glEnd();
	
      }
      
      local_counter -= skip_track_frames;
      ++count;
      if (count == max_frames) break;
    }
    
  }
  
}

void XOrsaOpenGL::internal_draw_orbit_ellipse() {
  
  glColor3d(0.5,0.5,0.5);
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  
  unsigned int j;
  Orbit orbit;
  
  glPushMatrix();
  
  for (j=0;j<evol_frame.size();++j) {
    
    if (orbit_reference_body_index==-HEX_AUTO) {
      if (((int)j)==orbit_reference_body_index_vector[evol_counter][j]) continue;
    } else {
      if (((int)j)==orbit_reference_body_index) continue;
    }
    
    Vector v_diff;
    if (orbit_reference_body_index==-HEX_AUTO) {
      v_diff = evol_frame[orbit_reference_body_index_vector[evol_counter][j]].position()-center;
    } else {
      v_diff = evol_frame[orbit_reference_body_index].position()-center;
    }
    
    const GLdouble equation_z_plus[4]  = {0,0, 1,-v_diff.z};
    const GLdouble equation_z_minus[4] = {0,0,-1, v_diff.z};
    //
    if (BrightPositiveZ()) {
      glClipPlane(GL_CLIP_PLANE0,equation_z_plus);
      glClipPlane(GL_CLIP_PLANE1,equation_z_minus);
    }
    
    glPushMatrix();
    
    glTranslated(v_diff.x,v_diff.y,v_diff.z);
    
    int ref_body_index;
    if (orbit_reference_body_index==-HEX_AUTO) {
      ref_body_index = orbit_reference_body_index_vector[evol_counter][j];
    } else {
      ref_body_index = orbit_reference_body_index;
    }
    
    if ( (orbit_cache_vector[evol_counter][j].set) &&
	 (orbit_cache_vector[evol_counter][j].ref_body_index == ref_body_index) ) {
      
      // cerr << "usign orbit cache!" << endl;
      
      orbit = orbit_cache_vector[evol_counter][j].orbit;
      
    } else {
      
      OrbitCache orbit_cache;
      
      orbit.Compute(evol_frame[j],evol_frame[ref_body_index]);
      
      orbit_cache.orbit          = orbit;
      orbit_cache.ref_body_index = ref_body_index;
      orbit_cache.set            = true;
      
      orbit_cache_vector[evol_counter][j] = orbit_cache;
    }
    
    if (orbit.e<1.0) {
      
      glPushMatrix();	      
      glRotated(orbit.omega_node*(180.0/pi),0,0,1);
      glRotated(orbit.i*(180.0/pi),1,0,0);
      glRotated(orbit.omega_pericenter*(180.0/pi),0,0,1);
      glTranslated(-orbit.a*orbit.e,0.0,0.0);
      
      // glPushMatrix();
      glScaled(orbit.a,orbit.a*secure_sqrt(1-orbit.e*orbit.e),orbit.a);
      
      // test: a rotation to keep a point on the body: better accuracy when zooming?
      double E = orbit.M;
      double tmp_E;
      int count = 0;
      do {
        tmp_E = E;
	E = orbit.e * sin(E) + orbit.M;
        count++;
      } while ( (fabs((E-tmp_E)/E) > 1.0e-15) && (count < 100) );
      //
      glRotated(E*(180.0/pi),0,0,1);
      
      // select the appropriate circle
      /* double pixel_length = 2*fov/width();
	 if (height()<width()) pixel_length = 2*fov/height();
      */
      //
      double pixel_length=0.0;
      switch(projection) {
      case OGL_ORTHO:
	pixel_length = 2*fov/width();
	if (height()<width()) pixel_length = 2*fov/height();
	break;
      case OGL_PERSPECTIVE:
	pixel_length = (pi/180.0)*fov/height();
	break;
      }
      //
      const int max_pixels_per_circle_element = 4; // should be tuned; the smaller this value, the better quality and more CPU
      // const int max_pixels_per_circle_element = 6; // should be tuned; the smaller this value, the better quality and more CPU
      // const int max_pixels_per_circle_element = 10; // should be tuned; the smaller this value, the better quality and more CPU
      //
      
      // int appropriate_circle_index = (int)rint(secure_log(orbit.a*(1.0+orbit.e)*pi/(max_pixels_per_circle_element*pixel_length))/secure_log(2.0)); // log base 2
      int appropriate_circle_index=0;
      switch(projection) {
      case OGL_ORTHO:
	appropriate_circle_index = (int)rint(secure_log(orbit.a*(1.0+orbit.e)*pi/(max_pixels_per_circle_element*pixel_length))/secure_log(2.0)); // log base 2
      	break;
      case OGL_PERSPECTIVE:
	// appropriate_circle_index = (int)rint(secure_log(orbit.a*(1.0+orbit.e)*pi/(max_pixels_per_circle_element*pixel_length))/secure_log(2.0)); // log base 2
	// appropriate_circle_index = 10; // to be replaced by a dynamic value
	appropriate_circle_index = (int)rint(secure_log(orbit.a*(1.0+orbit.e)/((evol_frame[ref_body_index].position()-evol_frame[j].position()).Length())/(max_pixels_per_circle_element*pixel_length))/secure_log(2.0)); // log base 2
	// appropriate_circle_index = (int)rint(secure_log(orbit.a*(1.0+orbit.e)/MIN(((evol_frame[ref_body_index].position()-evol_frame[j].position()).Length()),distance)/(max_pixels_per_circle_element*pixel_length))/secure_log(2.0)); // log base 2
	
	if ((int)j==central_body_index) appropriate_circle_index=max_circle_index-1;
	
	break;
      }
      
      if (appropriate_circle_index <  0) appropriate_circle_index=0;
      if (appropriate_circle_index >= max_circle_index) appropriate_circle_index=max_circle_index-1;
      // 
      // cerr << (*evol_it)[j].name() << " ellipse index: " << appropriate_circle_index << endl;
      
      if (appropriate_circle_index > 1) {
	
	const GLint appropriate_circle = circle+appropriate_circle_index;
	
	if (BrightPositiveZ()) {
	  glColor3d(0.5,0.5,0.5);
	  glEnable(GL_CLIP_PLANE0);
	  glCallList(appropriate_circle);
	  glDisable(GL_CLIP_PLANE0);
	  
	  glColor3d(0.3,0.3,0.3);
	  glEnable(GL_CLIP_PLANE1);
	  glCallList(appropriate_circle);
	  glDisable(GL_CLIP_PLANE1);
	} else {
	  glColor3d(0.5,0.5,0.5);
	  glCallList(appropriate_circle);
	}
	
      }
      
      glPopMatrix();
    }
    
    glPopMatrix();
  } 
  
  glPopMatrix();
}

void XOrsaOpenGL::internal_draw_time_label() {
  glColor3d(0,0.7,0);  
  QString time_label;
  //
  switch (universe->GetUniverseType()) {
  case Real: {
    // Date date = (*evol_it).GetDate();
    /*  int y,m,d;
	date.GetGregor(y,m,d);
	double fd = date.GetDayFraction();
	fd += d;
	sprintf(t,"time: %s %6.3d, %i",MonthNameShort[m],fd,y);
    */
    FineDate(time_label,evol_frame.GetDate());
    time_label.prepend("time: ");
    break;
  }
  case Simulated: 
    // sprintf(t,"time: %-6g %s",(*evol_it).Time(),TimeLabel().c_str());
    time_label.sprintf("time: %-6g %s",evol_frame.Time(),TimeLabel().c_str());
    break;
  }
  //
  renderText(10,20,time_label,default_font); 
}

void XOrsaOpenGL::internal_draw_scalebar() {
  glColor3d(0,0.7,0);
  
  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  gluOrtho2D(0.0,(GLdouble) width(),0.0, (GLdouble) height());
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity(); 
  
  switch(projection) {
  case OGL_ORTHO:
    renderText(10,height()-15,"scale: ",default_font); 
    break;
  case OGL_PERSPECTIVE:
    renderText(10,height()-15,"fov: ",default_font); 
    break;
  }
  
  GLdouble current_raster_position[4];
  glGetDoublev(GL_CURRENT_RASTER_POSITION,current_raster_position);
  // cerr << "current raster position: " << current_raster_position[0] << "  " << current_raster_position[1] << "  " << current_raster_position[2] << "  " << current_raster_position[3] << endl; 
  
  double pixel_length=0.0;
  switch(projection) {
  case OGL_ORTHO:
    pixel_length = 2.0*fov/width();
    if (height()<width()) pixel_length = 2.0*fov/height();
    break;
  case OGL_PERSPECTIVE:
    pixel_length = fov/height();
    break;
  }
  
  length_unit lu=AU;
  double bar_length=0.0;
  double coefficient=0.0;
  const int min_bar_length = MIN(60,width()/5); // pixels
  switch(projection) {
  case OGL_ORTHO:
    {
      // cerr << "min_bar_length: " << min_bar_length << "  width(): " << width() << endl;
      lu = AutoLengthUnit(coefficient_scale_label(min_bar_length*pixel_length));
      coefficient = FromUnits(coefficient_scale_label(FromUnits(min_bar_length*pixel_length,lu,-1)),lu);
      // const double bar_length  = coefficient / pixel_length;
      bar_length = coefficient / pixel_length;
    }
    break;
  case OGL_PERSPECTIVE:
    // coefficient = coefficient_scale_label(min_bar_length*pixel_length);
    coefficient = coefficient_angle_label(min_bar_length*pixel_length);
    bar_length = coefficient / pixel_length;
    break;
  }
  
  // cerr << "pixel_length: " << pixel_length << "  coefficient: " << coefficient << "  bar_length: " << bar_length << endl;
  
  const int tick_length = 3;
  
  const GLdouble bp[2] = {current_raster_position[0],current_raster_position[1]+3};
  
  glLineWidth(2.0);
  gl2psLineWidth(2.0);
  glBegin(GL_LINES);
  glVertex2d(bp[0],bp[1]);
  glVertex2d(bp[0]+bar_length,bp[1]);
  glEnd();
  // cosmetics
  glBegin(GL_LINES);
  glVertex2d(bp[0],bp[1]-tick_length);
  glVertex2d(bp[0],bp[1]+tick_length);
  glEnd();
  //
  glBegin(GL_LINES);
  glVertex2d(bp[0]+bar_length,bp[1]-tick_length);
  glVertex2d(bp[0]+bar_length,bp[1]+tick_length);
  glEnd();
  
  char t[1024];
  switch(projection) {
  case OGL_ORTHO:
    sprintf(t," = %.3g %s",FromUnits(coefficient,lu,-1),units->label(lu).c_str());
    renderText((int)rint(current_raster_position[0]+bar_length+5),height()-15,t,default_font); 
    break;
  case OGL_PERSPECTIVE:
    if (coefficient >= 1.0) {
      sprintf(t," = %02g 00' 00\"",coefficient);
    } else if (coefficient >= 1.0/60.0) {
      sprintf(t," = 00 %02g' 00\"",coefficient*60.0);
    } else {
      sprintf(t," = 00 00' %02g\"",coefficient*3600.0);
    }
    renderText((int)rint(current_raster_position[0]+bar_length+5),height()-15,t,default_font); 
    break;
  }
  
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  
  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);
  
}

void XOrsaOpenGL::internal_draw_FPS() {
  glColor3d(0,0.7,0);
  QString text;
  text.sprintf("FPS: %.0f",fps);
  renderText(10,height()-45,text,default_font); 
}

void XOrsaOpenGL::internal_draw_distance() {
  glColor3d(0,0.7,0);
  
  // label
  char t[1024];
  
  length_unit lu = AutoLengthUnit(distance);
  sprintf(t,"distance from %s: %.3g %s",center_body_name().c_str(),FromUnits(distance,lu,-1),units->label(lu).c_str());
  
  renderText(10,height()-60,t,default_font); 
}

void XOrsaOpenGL::animate() {
  
  if (bool_animate==false) return;
  
  if (bool_already_animating) {
    return;
  }
  
  bool_already_animating = true;
  
  if (evolution == 0) return;
  
  // evol_it++; 
  evol_counter++;
  if (evol_counter == (int)evolution->size()) {
    evol_counter = 0;
  }
  
  evol_frame = (*evolution)[evol_counter];
  
  // updateGL();
  
  /*   // cerr << "Frame time: " << (*evol_it).time << "  Energy: " << KineticEnergy(*evol_it) + evolution->integrator->interaction->PotentialEnergy(*evol_it) << endl;
       const double sys_K_energy = KineticEnergy(*evol_it);
       const double sys_P_energy = evolution->integrator->interaction->PotentialEnergy(*evol_it);
       printf("frame time: %8f   total system energy:   %16.12f  (K=  %16.12f + P=  %16.12f)\n",
       (*evol_it).time,sys_K_energy+sys_P_energy,sys_K_energy,sys_P_energy);
  */
  
  orbit_reference_body_index_vector.resize(evolution->size());
  
  orbit_cache_vector.resize(evolution->size());
  
  updateGL();
  
  frame_changed_from_animation = true;
  emit evolution_size_changed(evolution->size()); // update from evolution size, for multi-thread purposes
  emit frame_changed(evol_counter);
  frame_changed_from_animation = false;
  bool_already_animating = false;
}

// void XOrsaOpenGL::export_file(const string filename, const OpenGL_Export_File file_type) {
void XOrsaOpenGL::export_file(const QString &s_filename, const OpenGL_Export_File file_type) {
  
  // FILE *fp = fopen(filename.c_str(), "wb");
  FILE *fp = fopen(s_filename.latin1(), "wb");
  GLint buffsize = 0, state = GL2PS_OVERFLOW;
  GLint viewport[4];
  
  glGetIntegerv(GL_VIEWPORT, viewport);
  
  while (state==GL2PS_OVERFLOW) { 
    buffsize += 1024*1024;
    
    gl2psBeginPage ("ORSA OpenGL view","ORSA - http://orsa.sourceforge.net",viewport,
		    file_type, GL2PS_BSP_SORT,
		    GL2PS_SILENT | GL2PS_DRAW_BACKGROUND | GL2PS_BEST_ROOT | GL2PS_OCCLUSION_CULL,
		    GL_RGBA, 0, 0, 0, 0, 0, buffsize,
		    fp, s_filename.latin1());
    
    paintGL();
    
    state = gl2psEndPage();
  }
  
  fclose(fp);
}

void XOrsaOpenGL::UpdateVectors() {
  
  // cerr << "XOrsaOpenGL::UpdateVectors() --> central_body_index   = " << central_body_index << endl;
  // cerr << "XOrsaOpenGL::UpdateVectors() --> direction_body_index = " << direction_body_index << endl;
  
  if (central_body_index == -HEX_ORIGIN)
    center.Set(0,0,0);
  else if (central_body_index == -HEX_CENTER_OF_MASS)
    center = evol_frame.CenterOfMass();
  else
    center = evol_frame[central_body_index].position();
  
  if (direction_body_index == -HEX_ORIGIN)
    dir_pos.Set(0,0,0);
  else if (direction_body_index == -HEX_CENTER_OF_MASS)
    dir_pos = evol_frame.CenterOfMass();
  else
    dir_pos = evol_frame[direction_body_index].position();
  //
  dir_pos -= center;
  
  /* 
     if (central_body_index == -HEX_ORIGIN)
     dir_pos_mod.Set(0,0,0);
     else if (central_body_index == -HEX_CENTER_OF_MASS)
     dir_pos_mod = - evol_it->CenterOfMass();
     else
     dir_pos_mod = - (*evol_it)[central_body_index].position();
  */
  //
  dir_pos_mod = - dir_pos;
  
  // p0 = (*evol_it)[orbit_center_body_index].position();
  
  // dir_pos_zero is used only when plotting the track  
  /* 
     if ( (UseDirectionBody()) &&
     (direction_body_index < evol_it->size()) &&
     ( (central_body_index == -HEX_ORIGIN) || 
     (central_body_index == -HEX_CENTER_OF_MASS) || 
     (central_body_index < evol_it->size()) ) ) {
     
     cerr << "inside dir_pos_zero [*****]..." << endl;
     
     if (central_body_index == -HEX_ORIGIN)
     dir_pos_zero = (*evol_it)[direction_body_index].position();
     else if (central_body_index == -HEX_CENTER_OF_MASS)
     dir_pos_zero = (*evol_it)[direction_body_index].position() - evol_it->CenterOfMass();
     else
     dir_pos_zero = (*evol_it)[direction_body_index].position() - (*evol_it)[central_body_index].position();    
     
     } else {
     dir_pos_zero = dir_pos;
     }
  */
  //
  // usual value
  dir_pos_zero =  dir_pos;
  
}

void XOrsaOpenGL::setFOV(int i_fov) {
  // fov_base=secure_pow(1.02,(i_fov-900));
  switch (projection) {
  case OGL_ORTHO:       fov_base=secure_pow(1.02,(i_fov-900)); break;
  case OGL_PERSPECTIVE: fov_base=secure_pow(1.014,(i_fov-1000)); break;
  }
  updateFOV();
  resizeGL(width(),height());
  updateGL();
}

void XOrsaOpenGL::UpdateLocation(const Location &l) {
  location = l;
  resizeGL(width(),height());
  updateGL();
}

void XOrsaOpenGL::setDistance(int i_distance) {
  distance = zoom_scale_zero*secure_pow(1.02,(i_distance-900));
  resizeGL(width(),height());
  updateGL();
}

void XOrsaOpenGL::SetEvolution(const Evolution *evol_in) {
  if (evol_in == 0) {
    return;
  } else {
    evolution = evol_in;
    // evol_it   = evolution->begin();
    evol_frame   = (*evolution)[0];
    evol_counter = 0;
    
    orbit_reference_body_index_vector.clear();
    orbit_reference_body_index_vector.resize(evolution->size());
    
    orbit_cache_vector.clear();
    orbit_cache_vector.resize(evolution->size());
    
    // create the labels for all the bodies
    /* {
       if (evolution != 0)
       if (evolution->size() != 0) 
       if (evolution[0].size() != 0) {
       
       unsigned int body_index;
       string display_name;
       base_names = glGenLists(evolution[0][0].size());
       for (body_index=0;body_index<evolution[0][0].size();body_index++) {
       display_name = " " + evolution[0][0][body_index].name();
       glNewList(base_names+body_index, GL_COMPILE);
       glCallLists(display_name.size(),GL_BYTE,(GLubyte*) display_name.c_str());
       glEndList();
       // cerr << "created body name: " << display_name << endl;
       }
       
       }
       }
    */
    
    emit evolution_size_changed(evolution->size());
    emit evolution_changed();
  }
}

void XOrsaOpenGL::renderText(int x, int y, const QString &str, const QFont &fnt) {
  QGLWidget::renderText(x,y,"",fnt);
  const int fontsize=MAX(fnt.pixelSize(),fnt.pointSize());
  gl2psText(str,"Times-Roman",fontsize);
  QGLWidget::renderText(x,y,str,fnt);
}

void XOrsaOpenGL::renderText(double x, double y, double z, const QString &str, const QFont &fnt) {
  QGLWidget::renderText(x,y,z,"",fnt);
  const int fontsize=MAX(fnt.pixelSize(),fnt.pointSize());
  gl2psText(str,"Times-Roman",fontsize);
  QGLWidget::renderText(x,y,z,str,fnt);
}

std::string XOrsaOpenGL::center_body_name() const {
  if ( (central_body_index >= 0) &&
       (central_body_index < (int)evol_frame.size()) ) {
    return evol_frame[central_body_index].name();
  } else if (central_body_index == -HEX_ORIGIN) {
    return std::string("origin");
  } else if (central_body_index == -HEX_CENTER_OF_MASS) {
    return std::string("center of mass");
  } else {
    return std::string("");
  }
}

std::string XOrsaOpenGL::direction_body_name() const {
  if ( (direction_body_index >= 0) &&
       (direction_body_index < (int)evol_frame.size()) ) {
    return evol_frame[direction_body_index].name();
  } else if (direction_body_index == -HEX_ORIGIN) {
    return std::string("origin");
  } else if (direction_body_index == -HEX_CENTER_OF_MASS) {
    return std::string("center of mass");
  } else {
    return std::string("");
  }
}

void XOrsaOpenGL::mouse_slot(int x,int y) {
  
  if (projection!=OGL_PERSPECTIVE) return;   
  if (!DrawXYGrid()) return;
  
  mouse_x = x;
  mouse_y = y;
  
  // cerr << "mouse position: " << x << " " << y << endl;
  
  if (projection!=OGL_PERSPECTIVE) return;
  
  const double pixel_length = fov/height();   
  /* 
     const int min_bar_length = MIN(60,width()/5); // pixels
     double coefficient = coefficient_angle_label(min_bar_length*pixel_length);
     double bar_length = coefficient / pixel_length;
  */
  
  /* const double dx = (pi/180.0)*pixel_length*(x-(width()/2));
     const double dy = (pi/180.0)*pixel_length*((height()/2)-y);
  */
  //
  const double dx = (pi/180.0)*pixel_length*((width()/2)-x);
  const double dy = (pi/180.0)*pixel_length*(y-(height()/2));
  
  // cerr << "dx: " << dx << "   dy: " << dy << endl;
  
  const double ph = secure_atan2(dy,dx);
  const double th = secure_sqrt(dx*dx+dy*dy);
  
  double v[3];
  //
  v[0] = cos(ph)*sin(th);
  v[1] = sin(ph)*sin(th);
  v[2] = cos(th);
  
  // v[3]: CHECK HERE, maybe the old version commented out is better
  
  /* 
     double v[3];
     //
     v[0] = sin(dx);
     v[1] = cos(dx)*sin(dy);
     v[2] = cos(dx)*cos(dy);
  */
  
  double w[3];
  //
  w[0] = v[0]*inverted[0]+v[1]*inverted[4]+v[2]*inverted[8];
  w[1] = v[0]*inverted[1]+v[1]*inverted[5]+v[2]*inverted[9];
  w[2] = v[0]*inverted[2]+v[1]*inverted[6]+v[2]*inverted[10];
  //
  switch (universe->GetReferenceSystem()) {
  case ECLIPTIC:
    { 
      Vector v;
      v.Set(w[0],w[1],w[2]);
      
      // EclipticToEquatorial(v,evol_frame);
      EclipticToEquatorial_J2000(v);
      
      w[0] = v.x;
      w[1] = v.y;
      w[2] = v.z;
    }
    break;
  case EQUATORIAL: break;
  }  
  //
  const double phi   = secure_atan2(w[1],w[0]);
  const double theta = secure_atan2(w[2],secure_sqrt(w[0]*w[0]+w[1]*w[1])); 
  
  {
    Angle angle;
    double d,p,s,h,m;
    char msg[1024];
    
    angle.SetRad(fmod(phi+pi,twopi));
    angle.GetHMS(h,m,s);
    sprintf(msg,"MOUSE R.A. %02gh %02gm %03.1fs",h,m,s);
    ORSA_ERROR_MACRO(msg);
    
    angle.SetRad(-theta);
    angle.GetDPS(d,p,s);
    sprintf(msg,"MOUSE DEC. %+02g %02g' %03.1f\"",d,p,s);
    ORSA_ERROR_MACRO(msg);
    
  }
  
}

void XOrsaOpenGL::set_frame(int position) {
 
  if (frame_changed_from_animation) return;
  
  if (evolution != 0) {
    if (evolution->size() != 0) {
      if (position < 0) {
       	evol_counter = 0;
      } else if (position >= (int)evolution->size()) {
       	evol_counter = evolution->size()-1;
      } else {
	const bool keep_bool_animate = bool_animate;
	bool_animate = false;
	const int diff = position - evol_counter;
	if (diff < 0) {
	  while (evol_counter != position) {
	    --evol_counter;
	    }
	} else if (diff > 0) {
	  while (evol_counter != position) {
	    ++evol_counter;
	  }
	}
	bool_animate = keep_bool_animate;
      }
    } 
  }
  evol_frame = (*evolution)[evol_counter];
  updateGL();
}

// XOrsaOpenGLEvolutionTool

XOrsaOpenGLEvolutionTool::XOrsaOpenGLEvolutionTool(QWidget *parent) : QMainWindow(parent,0,Qt::WType_TopLevel | Qt::WDestructiveClose) {
  opengl = new XOrsaOpenGLEvolutionWidget(this);
  setCentralWidget(opengl);
  init_toolbars();
}

XOrsaOpenGLEvolutionTool::XOrsaOpenGLEvolutionTool(int w, int h, QWidget *parent) : QMainWindow(parent,0,Qt::WType_TopLevel | Qt::WDestructiveClose) {
  opengl = new XOrsaOpenGLEvolutionWidget(w,h,this);
  setCentralWidget(opengl);
  init_toolbars();
}

XOrsaOpenGLEvolutionTool::~XOrsaOpenGLEvolutionTool() {
  delete opengl;
}

void XOrsaOpenGLEvolutionTool::SetEvolution(const Evolution *e) {
  
  opengl->SetEvolution(e);
  
  oc_orbits->Set(opengl->bodies(),true);
  oc_orbits->SetObject(HEX_AUTO);
  
  oc_moid_1->Set(opengl->bodies(),false);
  oc_moid_2->Set(opengl->bodies(),false);
}

void XOrsaOpenGLEvolutionTool::init_toolbars() {
 
  /* 
     {
     // animation
     QToolBar * animationTools = new QToolBar(this);
     animationTools->setLabel("Animation Tools");
     //
     animate_tb = new XOrsaBoolToolButton(&opengl->bool_animate,QIconSet(tmp_xpm),"Animate",animationTools); 
     
     TimeSlider * time_slider = new TimeSlider(&opengl->evol_counter,opengl,animationTools);
     time_slider->setOrientation(Qt::Horizontal);
     QToolTip::add(time_slider,"Time Control");
     }
  */
  
  new XOrsaAnimationToolBar(opengl,this);
  
  new XOrsaLagrangePointsToolBar(opengl,this);
  
  /* 
     {
     // camera
     QToolBar * cameraTools = new QToolBar(this);
     cameraTools->setLabel("Camera Tools");
     
     projection_cb = new QComboBox(false,cameraTools);
     //
     projection_cb->insertItem("Orthographic");
     projection_cb->insertItem("Perspective");
     //
     connect(projection_cb,SIGNAL(activated(const QString&)),this,SLOT(slot_set_projection(const QString&)));
     //
     QToolTip::add(projection_cb,"Camera Projection");
     
     FOV_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->FOV,cameraTools);
     FOV_slider->setOrientation(Qt::Horizontal);
     
     near_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->near,cameraTools);
     near_slider->setOrientation(Qt::Horizontal);
     
     far_slider = new XOrsaDoubleObjectWithLimitsSliderTool(&opengl->far,cameraTools);
     far_slider->setOrientation(Qt::Horizontal);
     }
  */
  
  new XOrsaCameraToolBar(opengl,this);

  {
    // export images
    QToolBar * exportTools = new QToolBar(this);
    exportTools->setLabel("Export Tools");
    
    new QToolButton(QIconSet(png_xpm),"Export Image to PNG File",
		    QString::null,opengl,SLOT(export_png()),   exportTools);
    
    new QToolButton(QIconSet(ps_xpm),"Export Image to PostScript File",
		    QString::null,opengl,SLOT(export_ps()),    exportTools);
    
    new QToolButton(QIconSet(pdf_xpm),"Export Image to PDF File",
		    QString::null,opengl,SLOT(export_pdf()),   exportTools);
    
    new QToolButton(QIconSet(film_xpm),"Export All the Frame PNG Images to a Directory (movie)",
		    QString::null,opengl,SLOT(export_movie()), exportTools);
  }
  
  {
    // orbit
    QToolBar * orbitTools = new QToolBar(this);
    orbitTools->setLabel("Orbit Tools");
    //
    // plot_orbits_tb = new XOrsaBoolToolButton(&opengl->draw_orbits,QIconSet(tmp_xpm),"Plot Orbits",orbitTools); 
    plot_orbits_tb = new XOrsaBoolToolButton(&opengl->draw_orbits,QIconSet(orbits_xpm),"Plot Orbits",orbitTools); 
    //
    // oc_orbits = new XOrsaAutoObjectsCombo(orbitTools); 
    oc_orbits = new XOrsaImprovedObjectsCombo(HEX_AUTO,orbitTools);
    //
    bright_positive_z_tb = new XOrsaBoolToolButton(&opengl->bright_positive_Z,QIconSet(bright_positive_z_xpm),"Bright Positive Z",orbitTools); 
    //
    connect(oc_orbits,SIGNAL(ObjectChanged(int)),this,SLOT(slot_set_orbit_reference_body_index(int)));
    //
    // connect(plot_orbits_tb,SIGNAL(toggled(bool)),oc_orbits,SLOT(setEnabled(bool)));
    // connect(plot_orbits_tb,SIGNAL(toggled(bool)),bright_positive_z_tb,SLOT(setEnabled(bool)));
    // connect(plot_orbits_tb,SIGNAL(toggled(bool)),bright_positive_z_tb,SLOT(widgets_enabler()));
    //
    oc_orbits->setEnabled(false);
    //
    bright_positive_z_tb->setEnabled(false);
    //
    QToolTip::add(oc_orbits,"Orbits Reference Body");
    
    orbitTools->addSeparator();
    
    // MOID
    // ..remember: MOID needs GSL!
    moid_tb = new XOrsaBoolToolButton(&opengl->draw_MOID,QIconSet(moid_xpm),"MOID - Minimum Orbital Intersection Distance",orbitTools); 
    //
    oc_moid_1 = new XOrsaImprovedObjectsCombo(orbitTools);
    oc_moid_2 = new XOrsaImprovedObjectsCombo(orbitTools);
    //
    connect(oc_moid_1,SIGNAL(ObjectChanged(int)),this,SLOT(slot_set_moid_body_1_index(int)));
    connect(oc_moid_2,SIGNAL(ObjectChanged(int)),this,SLOT(slot_set_moid_body_2_index(int)));
    
    moid_tb->setEnabled(false);
    
    oc_moid_1->setEnabled(false);
    oc_moid_2->setEnabled(false);
    
    QToolTip::add(oc_moid_1,"MOID Body 1");
    QToolTip::add(oc_moid_2,"MOID Body 2");
    
    connect(plot_orbits_tb, SIGNAL(toggled(bool)),this,SLOT(widgets_enabler()));
    connect(moid_tb,        SIGNAL(toggled(bool)),this,SLOT(widgets_enabler()));
    
  }
}

void XOrsaOpenGLEvolutionTool::widgets_enabler() {
  
  // orbits
  if (plot_orbits_tb->isOn()) {
    oc_orbits->setEnabled(true);
    bright_positive_z_tb->setEnabled(true);
#ifdef HAVE_GSL	
    moid_tb->setEnabled(true);
#endif // HAVE_GSL	
  } else {
    oc_orbits->setEnabled(false);
    bright_positive_z_tb->setEnabled(false);
    moid_tb->setEnabled(false);
  }
  
  // moid
  if (moid_tb->isEnabled() && moid_tb->isOn()) {
#ifdef HAVE_GSL	
    oc_moid_1->setEnabled(true);
    oc_moid_2->setEnabled(true);
#endif // HAVE_GSL
  } else {
    oc_moid_1->setEnabled(false);
    oc_moid_2->setEnabled(false);
  }
}

void XOrsaOpenGLEvolutionTool::slot_set_orbit_reference_body_index(int i) {
  opengl->orbit_reference_body_index = i;
}

void XOrsaOpenGLEvolutionTool::slot_set_moid_body_1_index(int i) {
  opengl->MOID_body_1 = i;
}

void XOrsaOpenGLEvolutionTool::slot_set_moid_body_2_index(int i) {
  opengl->MOID_body_2 = i;
}





// XOrsaExtendedDockOpenGL

XOrsaExtendedDockOpenGL::XOrsaExtendedDockOpenGL(QWidget *parent) : QMainWindow(parent,0,Qt::WType_TopLevel | Qt::WDestructiveClose) {
  QVBox *vbox = new QVBox(this);
  setCentralWidget(vbox);
  opengl = new XOrsaOpenGL(vbox);
  init_bottom(statusBar());
  init_dock();
  //
  connect(opengl,SIGNAL(projection_changed()),this,SLOT(projection_changed_slot()));
  //
  resize(400,400);
}

XOrsaExtendedDockOpenGL::XOrsaExtendedDockOpenGL(int w, int h, QWidget *parent) : QMainWindow(parent,0,Qt::WType_TopLevel | Qt::WDestructiveClose) {
  QVBox *vbox = new QVBox(this);
  setCentralWidget(vbox);
  opengl = new XOrsaOpenGL(w,h,vbox);
  init_bottom(statusBar());
  init_dock();
  //
  connect(opengl,SIGNAL(projection_changed()),this,SLOT(projection_changed_slot()));
  //
  resize(400,400); 
}

XOrsaExtendedDockOpenGL::~XOrsaExtendedDockOpenGL() {
  delete opengl;
}

void XOrsaExtendedDockOpenGL::init_bottom(QStatusBar *sb) {
  
  QFrame *buttons_frame = new QFrame(sb);
  buttons_frame->setFrameStyle( QFrame::Box | QFrame::Raised);
  
  QHBoxLayout *buttons_lay = new QHBoxLayout(buttons_frame,5);
  
  fov_label = new QLabel("scale:",buttons_frame);
  buttons_lay->addWidget(fov_label);
  fov_slider = new QSlider(Qt::Horizontal,buttons_frame);
  fov_slider->setMinimumWidth(60);
  buttons_lay->addWidget(fov_slider);
  connect(fov_slider,SIGNAL(valueChanged(int)),this,SLOT(fov_slider_changed(int)));
  fov_slider->setMinValue(0);
  fov_slider->setMaxValue(1000);
  fov_slider->setValue(900);
  opengl->setFOV(900);
  fov_slider_bool = false;
  connect(opengl,SIGNAL(evolution_changed()),this,SLOT(update_zoom_scale_zero()));
  connect(fov_slider,SIGNAL(sliderPressed()),this,SLOT(fov_slider_pressed()));
  //
  /* 
     distance_label = new QLabel("distance:",buttons_frame);
     buttons_lay->addWidget(distance_label);
     distance_slider = new QSlider(Qt::Horizontal,buttons_frame);
     distance_slider->setMinimumWidth(60);
     buttons_lay->addWidget(distance_slider);
     connect(distance_slider,SIGNAL(valueChanged(int)),this,SLOT(distance_slider_changed(int)));
     distance_slider->setMinValue(0);
     distance_slider->setMaxValue(1000);
     distance_slider->setValue(900);
     distance_slider_bool = false;
     connect(distance_slider,SIGNAL(sliderPressed()),this,SLOT(distance_slider_pressed()));
     distance_label->hide();
     distance_slider->hide();
  */
  //
  buttons_lay->addWidget(new QLabel("time:",buttons_frame));
  time_slider = new QSlider(Qt::Horizontal,buttons_frame);
  time_slider->setMinimumWidth(60);
  time_slider->setMinValue(0);
  time_slider->setMaxValue(0);
  buttons_lay->addWidget(time_slider);
  connect(opengl,SIGNAL(frame_changed(int)),time_slider,SLOT(setValue(int)));
  connect(opengl,SIGNAL(evolution_size_changed(int)),this,SLOT(setTimeSliderMaxValue(int)));
  connect(time_slider,SIGNAL(valueChanged(int)),opengl,SLOT(set_frame(int)));
  //
  connect(time_slider,SIGNAL(sliderPressed()),this,SLOT(time_slider_pressed()));
  connect(time_slider,SIGNAL(sliderReleased()),this,SLOT(time_slider_released()));
  //
  play_pb = new QPushButton("play",buttons_frame);
  play_pb->setToggleButton(true);
  connect(play_pb,SIGNAL(toggled(bool)),this,SLOT(play_pb_toggled(bool)));
  buttons_lay->addWidget(play_pb);
  //
  xy_pb = new QPushButton("XY",buttons_frame);
  buttons_lay->addWidget(xy_pb);
  connect(xy_pb,SIGNAL(clicked()),opengl,SLOT(XY()));
  //
  xz_pb = new QPushButton("XZ",buttons_frame);
  buttons_lay->addWidget(xz_pb);
  connect(xz_pb,SIGNAL(clicked()),opengl,SLOT(XZ()));
  //
  yz_pb = new QPushButton("YZ",buttons_frame);
  buttons_lay->addWidget(yz_pb);
  connect(yz_pb,SIGNAL(clicked()),opengl,SLOT(YZ()));
  //
  // buttons_lay->addStretch();
  
  QSizePolicy csp = buttons_frame->sizePolicy();
  csp.setVerData(QSizePolicy::Fixed);
  buttons_frame->setSizePolicy(csp);
  
  sb->addWidget(buttons_frame,1,true);
}

void XOrsaExtendedDockOpenGL::init_dock() {
  setDockMenuEnabled(false);
  dock = new QDockWindow(QDockWindow::InDock,this);
  sidebar = new XOrsaOpenGLSidebarWidget(opengl,dock);
  dock->setResizeEnabled(true);
  dock->setWidget(sidebar);
  dock->setCloseMode(QDockWindow::Never);
  // addDockWindow(dock,Qt::DockRight);
  addDockWindow(dock,Qt::DockBottom);
  dock->setCaption("sidebar");
  // dock->setFixedExtentWidth(200);
}

void XOrsaExtendedDockOpenGL::projection_changed_slot() {
  switch(opengl->GetProjection()) {
  case OGL_ORTHO:
    fov_label->setText("scale:");
    //
    xy_pb->show();
    xz_pb->show();
    yz_pb->show();
    //
    // distance_label->hide();
    // distance_slider->hide();
    break;
  case OGL_PERSPECTIVE:
    fov_label->setText("fov:");
    //
    xy_pb->hide();
    xz_pb->hide();
    yz_pb->hide();
    //
    // distance_label->show();
    // distance_slider->show();
    break;
  }
}

void XOrsaExtendedDockOpenGL::setTimeSliderMaxValue(int size) {
  time_slider->setMaxValue(size-1);
}

void XOrsaExtendedDockOpenGL::update_zoom_scale_zero() {
  // Evolution::iterator ev_it=opengl->evolution->begin();
  Evolution::const_iterator ev_it=opengl->evolution->begin();
  double x_min,x_max,y_min,y_max,z_min,z_max;
  double x_mean,y_mean,z_mean;
  //
  x_mean=y_mean=z_mean=0;
  //
  x_min=x_max=ev_it[0][0].position().x;
  y_min=y_max=ev_it[0][0].position().y;
  z_min=z_max=ev_it[0][0].position().z;
  //
  unsigned int body;
  while (ev_it!=opengl->evolution->end()) {
    for(body=0;body<ev_it->size();body++) {
      if ((*ev_it)[body].position().x<x_min) x_min=(*ev_it)[body].position().x;
      if ((*ev_it)[body].position().x>x_max) x_max=(*ev_it)[body].position().x;
      if ((*ev_it)[body].position().y<y_min) y_min=(*ev_it)[body].position().y;
      if ((*ev_it)[body].position().y>y_max) y_max=(*ev_it)[body].position().y;
      if ((*ev_it)[body].position().z<z_min) z_min=(*ev_it)[body].position().z;
      if ((*ev_it)[body].position().z>z_max) z_max=(*ev_it)[body].position().z;
    }
    ev_it++;
  }
  //
  double delta_x=x_max-x_min;
  double delta_y=y_max-y_min;
  double delta_z=z_max-z_min;
  //
  double delta_max=delta_x;
  if (delta_y>delta_max) delta_max=delta_y;
  if (delta_z>delta_max) delta_max=delta_z;
  //
  opengl->setZoomScaleZero(delta_max);
}

void XOrsaExtendedDockOpenGL::fov_slider_pressed() {
  focusWidget()->clearFocus();
  fov_slider->setFocus();
}

void XOrsaExtendedDockOpenGL::fov_slider_changed(int) {
  if (!fov_slider_bool) { 
    fov_slider_bool=true;
    opengl->setFOV(fov_slider->value());
    fov_slider_bool=false;
  }
}

void XOrsaExtendedDockOpenGL::update_fov_slider(int l) {
  if (!fov_slider_bool) { 
    fov_slider_bool=true;
    fov_slider->setValue(l);
    fov_slider_bool=false;
  }
}

/* 
   void XOrsaExtendedDockOpenGL::distance_slider_pressed() {
   focusWidget()->clearFocus();
   distance_slider->setFocus();
   }
   
   void XOrsaExtendedDockOpenGL::distance_slider_changed(int) {
   if (!distance_slider_bool) { 
   distance_slider_bool=true;
   opengl->setDistance(distance_slider->value());
   distance_slider_bool=false;
   }
   }
*/

void XOrsaExtendedDockOpenGL::time_slider_pressed() {
  focusWidget()->clearFocus();
  time_slider->setFocus();
  time_slider_old_animation = opengl->IsAnimating();
  opengl->StopAnimation();
}

void XOrsaExtendedDockOpenGL::time_slider_released() {
  if (time_slider_old_animation==true) opengl->StartAnimation();
}

void XOrsaExtendedDockOpenGL::play_pb_toggled(bool play) {
  if (play) {
    opengl->StartAnimation();
    play_pb->setText("stop");
  } else {
    opengl->StopAnimation();
    play_pb->setText("play");
  }
}

void XOrsaExtendedDockOpenGL::keyPressEvent(QKeyEvent *e) {
  if (e->key() == Qt::Key_F) {
    QWidget *w = opengl;
    while (w->parentWidget(true)) {
      w = w->parentWidget(true);
    }
#if (QT_VERSION>0x030300)
    if (w->isFullScreen()) {
      w->setWindowState(w->windowState() & ~WindowFullScreen);
    } else {
      w->setWindowState(w->windowState() ^ WindowFullScreen);
    } 
#endif
    e->accept();
  } else if (e->key() == Qt::Key_Escape) {
    QWidget *w = opengl;
    while (w->parentWidget(true)) {
      w = w->parentWidget(true);
    }
#if (QT_VERSION>0x030300)
    if (w->isFullScreen()) {
      w->setWindowState(w->windowState() & ~WindowFullScreen);
    } 
#endif 
    e->accept();
  } else {
    e->ignore();
  }
}

// XOrsaOpenGLSidebarWidget: the content of the XOrsaExtendedDockOpenGL sidebar

XOrsaOpenGLSidebarWidget::XOrsaOpenGLSidebarWidget(XOrsaOpenGL *ogl, QWidget *parent) : QTabWidget(parent), opengl(ogl) {
  
  connect(opengl,SIGNAL(evolution_changed()),this,SLOT(content_update()));
  
  {
    QWidget *w_main = new QWidget(this);
    QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
    
    {
      QHBox *hb_labels = new QHBox(w_main);
      hb_labels->setSpacing(3);
      //
      QCheckBox *cb_labels = new QCheckBox("labels",hb_labels);
      XOrsaLabelsModeCombo *lm_labels = new XOrsaLabelsModeCombo(hb_labels);
      //
      connect(cb_labels,SIGNAL(toggled(bool)),opengl,SLOT(DrawLabels(bool)));
      connect(cb_labels,SIGNAL(toggled(bool)),lm_labels,SLOT(setEnabled(bool)));
      cb_labels->setChecked(false);
      lm_labels->setEnabled(false);
      //
      connect(lm_labels,SIGNAL(ModeChanged(OpenGL_Labels_Mode)),opengl,SLOT(SetLabelsMode(OpenGL_Labels_Mode)));
      lm_labels->SetLabelsMode(ALL);  
      //
      vlay->addWidget(hb_labels);
    }
    
    {
      QHBox *hb_scale = new QHBox(w_main);
      hb_scale->setSpacing(3);
      //
      QCheckBox *cb_scale = new QCheckBox("scale bar",hb_scale);
      // lc_scale = new LengthCombo(hb_scale);
      //
      connect(cb_scale,SIGNAL(toggled(bool)),opengl,SLOT(DrawScaleBar(bool)));
      // connect(cb_scale,SIGNAL(toggled(bool)),lc_scale,SLOT(setEnabled(bool)));
      cb_scale->setChecked(false);
      // lc_scale->setEnabled(false);
      //
      // lc_scale->SetUnit(units->GetLengthBaseUnit());
      //
      vlay->addWidget(hb_scale);
    }
    
    {
      QCheckBox *cb_FPS = new QCheckBox("FPS",w_main);
      //
      connect(cb_FPS,SIGNAL(toggled(bool)),opengl,SLOT(DrawFPS(bool)));
      //
      cb_FPS->setChecked(false);
      //
      vlay->addWidget(cb_FPS);
    }
    
    /* 
       {
       QCheckBox *cb_tracks = new QCheckBox("track",w_main);
       //
       connect(cb_tracks,SIGNAL(toggled(bool)),opengl,SLOT(DrawTracks(bool)));
       //
       cb_tracks->setChecked(false);
       //
       vlay->addWidget(cb_tracks);
       }
    */
     
    {
      QHBox *hb_track = new QHBox(w_main);
      hb_track->setSpacing(3);
      //
      // QCheckBox *cb_tracks = new QCheckBox("track",w_main);
      QCheckBox *cb_tracks = new QCheckBox("track",hb_track);
      QCheckBox *cb_absolute_tracks = new QCheckBox("absolute track",hb_track);
      //
      connect(cb_tracks,SIGNAL(toggled(bool)),opengl,SLOT(DrawTracks(bool)));
      connect(cb_absolute_tracks,SIGNAL(toggled(bool)),opengl,SLOT(AbsoluteTracks(bool)));
      connect(cb_tracks,SIGNAL(toggled(bool)),cb_absolute_tracks,SLOT(setEnabled(bool)));
      //
      cb_tracks->setChecked(false);
      cb_absolute_tracks->setChecked(false);
      cb_absolute_tracks->setEnabled(false);
      //
      // vlay->addWidget(cb_tracks);
      vlay->addWidget(hb_track);
    }
    
    {
      QCheckBox *cb_show_ref_sys = new QCheckBox("show reference system",w_main);
      //
      connect(cb_show_ref_sys,SIGNAL(toggled(bool)),opengl,SLOT(DrawReferenceSystem(bool)));
      //
      cb_show_ref_sys->setChecked(false);
      //
      vlay->addWidget(cb_show_ref_sys);
    }
    
    {
      cb_show_grid = new QCheckBox("show X-Y grid",w_main);
      //
      connect(cb_show_grid,SIGNAL(toggled(bool)),opengl,SLOT(DrawXYGrid(bool)));
      //
      cb_show_grid->setChecked(false);
      //
      vlay->addWidget(cb_show_grid);
    }
    
    {
      QCheckBox *cb_show_z_bars = new QCheckBox("show Z bars",w_main);
      //
      connect(cb_show_z_bars,SIGNAL(toggled(bool)),opengl,SLOT(DrawZBars(bool)));
      //
      cb_show_z_bars->setChecked(false);
      //
      vlay->addWidget(cb_show_z_bars);
    }
    
    {
      QHBoxLayout *hb_fs = new QHBoxLayout(vlay);
      //
      QPushButton *pb_fs = new QPushButton("full screen on/off",w_main);
      // pb_fs->setToggleButton(true);
      hb_fs->addWidget(pb_fs);
      hb_fs->addStretch();
      //
      // connect(pb_fs,SIGNAL(toggled(bool)),this,SLOT(full_screen_slot(bool)));
      connect(pb_fs,SIGNAL(clicked()),this,SLOT(full_screen_slot()));
    }  
    
    vlay->addStretch(1);
    
    addTab(w_main,"general");
  }
  
  /* 
     {
     QWidget *w_main = new QWidget(this);
     QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
     
     {
     QHBox *hb_center = new QHBox(w_main);
     hb_center->setSpacing(3);
     //
     center_label = new QLabel("center",hb_center);
     eoc_center = new XOrsaExtendedObjectsCombo(hb_center);
     //
     connect(eoc_center,SIGNAL(ObjectChanged(int)),opengl,SLOT(center_on_body(int)));
     //
     vlay->addWidget(hb_center);
     }
     
     {
     QHBox *hb_direction = new QHBox(w_main);
     hb_direction->setSpacing(3);
     //
     cb_direction = new QCheckBox("direction",hb_direction);
     eoc_direction = new XOrsaExtendedObjectsCombo(hb_direction);
     //
     connect(cb_direction,SIGNAL(toggled(bool)),opengl,SLOT(UseDirectionBody(bool)));
     connect(cb_direction,SIGNAL(toggled(bool)),eoc_direction,SLOT(setEnabled(bool)));
     cb_direction->setChecked(false);
     eoc_direction->setEnabled(false);
     connect(eoc_direction,SIGNAL(ObjectChanged(int)),opengl,SLOT(direction_body(int)));
     //
     vlay->addWidget(hb_direction);
     }
     
     {
     QCheckBox *cb_show_ref_sys = new QCheckBox("show reference system",w_main);
     //
     connect(cb_show_ref_sys,SIGNAL(toggled(bool)),opengl,SLOT(DrawReferenceSystem(bool)));
     //
     cb_show_ref_sys->setChecked(false);
     //
     vlay->addWidget(cb_show_ref_sys);
     }
     
     {
     cb_show_grid = new QCheckBox("show X-Y grid",w_main);
     //
     connect(cb_show_grid,SIGNAL(toggled(bool)),opengl,SLOT(DrawXYGrid(bool)));
     //
     cb_show_grid->setChecked(false);
     //
     vlay->addWidget(cb_show_grid);
     }
     
     {
     QCheckBox *cb_show_z_bars = new QCheckBox("show Z bars",w_main);
     //
     connect(cb_show_z_bars,SIGNAL(toggled(bool)),opengl,SLOT(DrawZBars(bool)));
     //
     cb_show_z_bars->setChecked(false);
     //
     vlay->addWidget(cb_show_z_bars);
     }
     
     vlay->addStretch(1);
     
     addTab(w_main,"reference system");
     }	
  */
  
  {
    QWidget *w_main = new QWidget(this);
    QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
    
    {
      QComboBox *projection_cb = new QComboBox(false,w_main);
      //
      projection_cb->insertItem("orthographic");
      projection_cb->insertItem("perspective");
      //
      connect(projection_cb,SIGNAL(activated(const QString&)),this,SLOT(set_projection(const QString&)));
      //
      vlay->addWidget(projection_cb);
    }
    
    {
      QHBox *hb_center = new QHBox(w_main);
      hb_center->setSpacing(3);
      //
      center_label = new QLabel("center",hb_center);
      // eoc_center = new XOrsaExtendedObjectsCombo(hb_center);
      eoc_center = new XOrsaImprovedObjectsCombo(HEX_ORIGIN | HEX_CENTER_OF_MASS, hb_center);
      //
      connect(eoc_center,SIGNAL(ObjectChanged(int)),opengl,SLOT(center_on_body(int)));
      //
      vlay->addWidget(hb_center);
    }
    
    {
      QHBox *hb_direction = new QHBox(w_main);
      hb_direction->setSpacing(3);
      //
      cb_direction = new QCheckBox("frame direction",hb_direction);
      // eoc_direction = new XOrsaExtendedObjectsCombo(hb_direction);
      eoc_direction = new XOrsaImprovedObjectsCombo(HEX_ORIGIN | HEX_CENTER_OF_MASS, hb_direction);
      //
      connect(cb_direction,SIGNAL(toggled(bool)),opengl,SLOT(UseDirectionBody(bool)));
      connect(cb_direction,SIGNAL(toggled(bool)),eoc_direction,SLOT(setEnabled(bool)));
      cb_direction->setChecked(false);
      eoc_direction->setEnabled(false);
      connect(eoc_direction,SIGNAL(ObjectChanged(int)),opengl,SLOT(direction_body(int)));
      //
      vlay->addWidget(hb_direction);
    }
    
    {
      observer_gb = new QGroupBox(2,Qt::Vertical,"observer",w_main);
      
      QHBox *hb_free = new QHBox(observer_gb);
      rb_free = new QRadioButton(hb_free);
      connect(opengl,SIGNAL(center_changed(int)),this,SLOT(slot_center_changed()));
      //
      distance_slider = new QSlider(Qt::Horizontal,hb_free);
      distance_slider->setMinimumWidth(60);
      connect(distance_slider,SIGNAL(valueChanged(int)),this,SLOT(distance_slider_changed(int)));
      distance_slider->setMinValue(0);
      distance_slider->setMaxValue(1000);
      distance_slider->setValue(900);
      distance_slider_bool = false;
      connect(distance_slider,SIGNAL(sliderPressed()),this,SLOT(distance_slider_pressed()));
      //
      connect(rb_free,SIGNAL(toggled(bool)),distance_slider,SLOT(setEnabled(bool)));
      
      QHBox *hb_object = new QHBox(observer_gb);
      rb_object = new QRadioButton("",hb_object);
      connect(opengl,SIGNAL(direction_changed(int)),this,SLOT(slot_direction_changed()));
      loc_pb = new XOrsaLocationPushButton(hb_object);
      connect(loc_pb,SIGNAL(LocationChanged()),this,SLOT(UpdateLocation()));
      //
      // connect(rb_object,SIGNAL(toggled(bool)),loc_pb,SLOT(setEnabled(bool)));
      // connect(rb_object,SIGNAL(toggled(bool)),cb_direction,SLOT(setChecked(bool)));
      connect(rb_object,SIGNAL(toggled(bool)),this,SLOT(slot_rb_object_toggled(bool)));
      
      QButtonGroup *bg = new QButtonGroup(w_main);
      bg->setRadioButtonExclusive(true);
      bg->hide();
      bg->insert(rb_free);
      bg->insert(rb_object);
      
      // init 
      rb_free->setChecked(true);
      rb_object->setChecked(true);
      rb_free->setChecked(true);
      
      vlay->addWidget(observer_gb);
    }
    
    {
      QHBoxLayout *hb_lt = new QHBoxLayout(vlay);
      //
      QPushButton *pb_lt = new QPushButton("light-time delay correction",w_main);
      pb_lt->setToggleButton(true);
      hb_lt->addWidget(pb_lt);
      hb_lt->addStretch();
      //
      connect(pb_lt,SIGNAL(toggled(bool)),opengl,SLOT(LightTimeCorrection(bool)));
    }
    
    vlay->addStretch(1);
    
    // init
    projection_changed_slot();
    
    addTab(w_main,"projection");
  }	
  
  {
    QWidget *w_main = new QWidget(this);
    QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
    
    {
      QHBox *hb_orbits = new QHBox(w_main);
      hb_orbits->setSpacing(3);
      //
      cb_orbits = new QCheckBox("reference body",hb_orbits);
      // oc_orbits = new XOrsaAutoObjectsCombo(hb_orbits);
      oc_orbits = new XOrsaImprovedObjectsCombo(HEX_AUTO,hb_orbits);
      //
      connect(cb_orbits,SIGNAL(toggled(bool)),opengl,SLOT(DrawOrbits(bool)));
      connect(oc_orbits,SIGNAL(ObjectChanged(int)),opengl,SLOT(orbit_reference_body(int)));
      oc_orbits->setEnabled(false);
      connect(cb_orbits,SIGNAL(toggled(bool)),oc_orbits,SLOT(setEnabled(bool)));
      cb_orbits->setChecked(false);
      oc_orbits->setEnabled(false);
      //
      vlay->addWidget(hb_orbits);
    }
    
    {
      QHBox *hb_moid = new QHBox(w_main);
      hb_moid->setSpacing(3);
      //
      cb_moid = new QCheckBox("moid",hb_moid);
      oc_moid_1 = new XOrsaImprovedObjectsCombo(hb_moid);
      oc_moid_2 = new XOrsaImprovedObjectsCombo(hb_moid);
      //
      connect(cb_moid,SIGNAL(toggled(bool)),opengl,SLOT(DrawMOID(bool)));
      connect(oc_moid_1,SIGNAL(ObjectChanged(int)),opengl,SLOT(MoidBody1(int)));
      connect(oc_moid_2,SIGNAL(ObjectChanged(int)),opengl,SLOT(MoidBody2(int)));
      connect(cb_moid,SIGNAL(toggled(bool)),oc_moid_1,SLOT(setEnabled(bool)));
      connect(cb_moid,SIGNAL(toggled(bool)),oc_moid_2,SLOT(setEnabled(bool)));
      cb_moid->setChecked(false);
      cb_moid->setEnabled(false);
      oc_moid_1->setEnabled(false);
      oc_moid_2->setEnabled(false);
      //
#ifdef HAVE_GSL
      connect(cb_orbits,SIGNAL(toggled(bool)),cb_moid,SLOT(setEnabled(bool))); // no moid without orbits, and enable moid only if it is available (have GSL)
#endif
      //
      vlay->addWidget(hb_moid);
    }
    
    {
      QHBox *hb_lagrange = new QHBox(w_main);
      hb_lagrange->setSpacing(3);
      //
      QCheckBox *cb_lagrange = new QCheckBox("Lagrange points",hb_lagrange);
      oc_lagrange_1 = new XOrsaImprovedObjectsCombo(hb_lagrange);
      oc_lagrange_2 = new XOrsaImprovedObjectsCombo(hb_lagrange);
      //
      connect(cb_lagrange,SIGNAL(toggled(bool)),opengl,SLOT(DrawLagrange(bool)));
      connect(oc_lagrange_1,SIGNAL(ObjectChanged(int)),opengl,SLOT(LagrangeBody1(int)));
      connect(oc_lagrange_2,SIGNAL(ObjectChanged(int)),opengl,SLOT(LagrangeBody2(int)));
      connect(cb_lagrange,SIGNAL(toggled(bool)),oc_lagrange_1,SLOT(setEnabled(bool)));
      connect(cb_lagrange,SIGNAL(toggled(bool)),oc_lagrange_2,SLOT(setEnabled(bool)));
      cb_lagrange->setChecked(false);
      oc_lagrange_1->setEnabled(false);
      oc_lagrange_2->setEnabled(false);
      //
      vlay->addWidget(hb_lagrange);
    }
      
    {
      QCheckBox *cb_bright = new QCheckBox("bright positive Z",w_main);
      //
      connect(cb_bright,SIGNAL(toggled(bool)),opengl,SLOT(BrightPositiveZ(bool)));
      connect(cb_orbits,SIGNAL(toggled(bool)),cb_bright,SLOT(setEnabled(bool))); // no moid without orbits
      //
      cb_bright->setEnabled(false);
      cb_bright->setChecked(false);
      //
      vlay->addWidget(cb_bright);
    }
    
    vlay->addStretch(1);
    
    addTab(w_main,"orbits");
  }
  
  /* 
     {
     QWidget *w_main = new QWidget(this);
     QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
     
     {
     
     }
     
     vlay->addStretch(1);
     
     addTab(w_main,"trails");
     }	
  */
  
  /* 
     {
     QWidget *w_main = new QWidget(this);
     QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
     
     {
     
     }
     
     vlay->addStretch(1);
     
     addTab(w_main,"colors");
     }	
     
     {
     QWidget *w_main = new QWidget(this);
     QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
     
     {
     
     }
     
     vlay->addStretch(1);
     
     addTab(w_main,"animation");
     }	
     
     {
     QWidget *w_main = new QWidget(this);
     QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
     
     {
     
     }
     
     vlay->addStretch(1);
     
     addTab(w_main,"fonts");
     }	
  */
  
  {
    QWidget *w_main = new QWidget(this);
    QVBoxLayout *vlay = new QVBoxLayout(w_main,3,3);
    
    {
      QPushButton *pb_acquire_single_PS_image = new QPushButton("save single image to PostScript file",w_main);
      //
      connect(pb_acquire_single_PS_image,SIGNAL(clicked()),this,SLOT(slot_acquire_single_PS_image()));
      //
      vlay->addWidget(pb_acquire_single_PS_image);
    }
    
    {
      QPushButton *pb_acquire_single_PDF_image = new QPushButton("save single image to PDF file",w_main);
      //
      connect(pb_acquire_single_PDF_image,SIGNAL(clicked()),this,SLOT(slot_acquire_single_PDF_image()));
      //
      vlay->addWidget(pb_acquire_single_PDF_image);
    }
    
    {
      QPushButton *pb_acquire_single_PNG_image = new QPushButton("grab single PNG image",w_main);
      //
      connect(pb_acquire_single_PNG_image,SIGNAL(clicked()),this,SLOT(slot_acquire_single_PNG_image()));
      //
      vlay->addWidget(pb_acquire_single_PNG_image);
    }
    
    {
      QPushButton *pb_acquire_movie = new QPushButton("grab a PNG image for each frame of the integration",w_main);
      //
      connect(pb_acquire_movie,SIGNAL(clicked()),this,SLOT(slot_acquire_movie()));
      //
      vlay->addWidget(pb_acquire_movie);
    }
    
    vlay->addStretch(1);
    
    addTab(w_main,"grab");
  }	
  
  connect(opengl,SIGNAL(projection_changed()),this,SLOT(projection_changed_slot())); 
}

void XOrsaOpenGLSidebarWidget::set_projection(const QString &s) {
  if (s=="orthographic") {
    opengl->SetProjection(OGL_ORTHO);
  } else if (s=="perspective") {
    opengl->SetProjection(OGL_PERSPECTIVE);
  } else {
    char err_msg[4096];
    sprintf(err_msg,"XOrsaOpenGLSidebarWidget::set_projection() can't handle [%s] projection",s.latin1());
    ORSA_ERROR_MACRO(err_msg);
  }
}

void XOrsaOpenGLSidebarWidget::content_update() {
  // lc_scale->SetUnit(units->GetLengthBaseUnit());
  eoc_center->Set(opengl->bodies());
  eoc_direction->Set(opengl->bodies());
  oc_orbits->Set(opengl->bodies(),true);
  oc_moid_1->Set(opengl->bodies(),false);
  oc_moid_2->Set(opengl->bodies(),false);
  oc_lagrange_1->Set(opengl->bodies(),true);
  oc_lagrange_2->Set(opengl->bodies(),true);
}

void XOrsaOpenGLSidebarWidget::projection_changed_slot() {
  switch(opengl->GetProjection()) {
  case OGL_ORTHO:
    center_label->setText("center");
    // cb_direction->setText("frame direction");
    cb_show_grid->setText("show X-Y grid");
    observer_gb->setEnabled(false);
    break;
  case OGL_PERSPECTIVE:
    /*
      center_label->setText("observer position");
      cb_direction->setText("direction");
      cb_show_grid->setText("show ra/dec grid");
    */
    //
    center_label->setText("center");
    // cb_direction->setText("observer position");
    cb_show_grid->setText("show ra/dec grid");
    observer_gb->setEnabled(true);
    break;
  }
}

void XOrsaOpenGLSidebarWidget::full_screen_slot() {
  QWidget *w = opengl;
  while (w->parentWidget()) {
    w = w->parentWidget();
  }
#if QT_VERSION>0x030300
  if (w->isFullScreen()) {
    w->setWindowState(w->windowState() & ~WindowFullScreen);
  } else {
    w->setWindowState(w->windowState() ^ WindowFullScreen);
  }
#endif
}

void XOrsaOpenGLSidebarWidget::slot_center_changed() {
  QString s;
  s.sprintf("distance from %s: ",opengl->center_body_name().c_str());
  rb_free->setText(s);
}

void XOrsaOpenGLSidebarWidget::slot_direction_changed() {
  QString s;
  s.sprintf("observe from %s.",opengl->direction_body_name().c_str());
  rb_object->setText(s);
  //
  if (opengl->direction_body_name() == "Earth") {
    loc_pb->setEnabled(true);
  } else {
    loc_pb->setEnabled(false);
  }
}

void XOrsaOpenGLSidebarWidget::slot_rb_object_toggled(bool b) {
  if (b) {
    cb_direction->setChecked(true);
    if (opengl->direction_body_name() == "Earth") {
      loc_pb->setEnabled(true);
    }
  } else {
    loc_pb->setEnabled(false); 
  }
  
  opengl->ObserveFromDirectionBody(b);
}

void XOrsaOpenGLSidebarWidget::distance_slider_pressed() {
  focusWidget()->clearFocus();
  distance_slider->setFocus();
}

void XOrsaOpenGLSidebarWidget::distance_slider_changed(int) {
  if (!distance_slider_bool) { 
    distance_slider_bool=true;
    opengl->setDistance(distance_slider->value());
    distance_slider_bool=false;
  }
}

/* 
   void XOrsaOpenGLSidebarWidget::keyPressEvent(QKeyEvent *e) {
   if (e->key() == Qt::Key_F) {
   QWidget *w = opengl;
   while (w->parentWidget(false)) {
   w = w->parentWidget(false);
   }
   if (w->isFullScreen()) {
   w->setWindowState(w->windowState() & ~WindowFullScreen);
   } else {
   w->setWindowState(w->windowState() ^ WindowFullScreen);
   } 
   e->accept();
   } else if (e->key() == Qt::Key_Escape) {
   QWidget *w = opengl;
   while (w->parentWidget(true)) {
   w = w->parentWidget(true);
   }
   if (w->isFullScreen()) {
   w->setWindowState(w->windowState() & ~WindowFullScreen);
   } 
   e->accept();
   } else {
   e->ignore();
   }
   }
*/

void XOrsaOpenGLSidebarWidget::UpdateLocation() {
  opengl->UpdateLocation(loc_pb->location);
}

void XOrsaOpenGLSidebarWidget::slot_acquire_single_PS_image() {
  
  QString s = QFileDialog::getSaveFileName(QString::null,QString::null,this,0,"save image as:");
  
  if (!(s.isEmpty())) {
    // opengl->export_file(s.latin1(), OGL_PS_FILE);
    opengl->export_file(s, OGL_PS_FILE);
  } 
}

void XOrsaOpenGLSidebarWidget::slot_acquire_single_EPS_image() {
  
  QString s = QFileDialog::getSaveFileName(QString::null,QString::null,this,0,"save image as:");
  
  if (!(s.isEmpty())) {
    // opengl->export_file(s.latin1(), OGL_EPS_FILE);
    opengl->export_file(s, OGL_EPS_FILE);
  } 
}

void XOrsaOpenGLSidebarWidget::slot_acquire_single_PDF_image() {
  
  QString s = QFileDialog::getSaveFileName(QString::null,QString::null,this,0,"save image as:");
  
  if (!(s.isEmpty())) {
    // opengl->export_file(s.latin1(), OGL_PDF_FILE);
    opengl->export_file(s, OGL_PDF_FILE);
  } 
}


void XOrsaOpenGLSidebarWidget::slot_acquire_single_PNG_image() {

  // Make GL Context current
  opengl->makeCurrent();
  
  // Copy from OpenGL
  QImage *tmpImage = new QImage(opengl->grabFrameBuffer());
  
  if (!tmpImage) {
    // error...
    return;
  }
  
  QString s = QFileDialog::getSaveFileName(QString::null,QString::null,this,0,"save image as:");
  
  if (!(s.isEmpty())) {
    tmpImage->save(s,"PNG");
  }
  
  // Cleaning memory
  delete tmpImage;
  
  // opengl->doneCurrent(); 
}


void XOrsaOpenGLSidebarWidget::slot_acquire_movie() {
  
  const QString s_dir = QFileDialog::getExistingDirectory(QString::null,this,0,"save images into:",true);
  
  if (!(s_dir.isEmpty())) {
    
    opengl->makeCurrent();
    
    QString s_filename;
    unsigned int j=0;
    while (j < opengl->total_frames()) {
      
      // Make GL Context current
      // opengl->makeCurrent();
      // better to call directly paintGL()
      // paintGL();
      // everything is updated by the set_frame call;
      
      s_filename.sprintf("frame-%06i.png",j);
      s_filename.prepend(s_dir);
      
      opengl->set_frame(j);
      
      // Copy from OpenGL
      QImage *tmpImage = new QImage(opengl->grabFrameBuffer());
      
      tmpImage->save(s_filename,"PNG");
      
      // Cleaning memory
      delete tmpImage;
      
      ++j;
      // j += 50;
    }
    
    // opengl->doneCurrent();
  }
}

////////////

// XOrsaOpenGLWidget: a rewrite for XOrsaOpenGL

XOrsaOpenGLWidget::XOrsaOpenGLWidget(QWidget *parent, WFlags f) : QGLWidget(parent,0,0,f) {
  setMinimumSize(300,200);
  init();  
}

XOrsaOpenGLWidget::XOrsaOpenGLWidget(int w, int h, QWidget *parent, WFlags f) : QGLWidget(parent,0,0,f) {
  setMinimumSize(w,h);
  init();
}

void XOrsaOpenGLWidget::init() {
  
  bool_animate = false;
  animation_delay_ms = 40;
  bool_already_animating = false;
  last_fps_frame_index = 0;
  fps.reset();
  
  projection = OGL_PERSPECTIVE;
  
  center_X_rotation = center_Y_rotation = center_Z_rotation = 0.0;
  eye_X_rotation = eye_Y_rotation = eye_Z_rotation = 0.0;
  center_rotation_impulse = 0.1;
  eye_rotation_impulse    = 0.1;
  
  near_and_far_limit_on_distance = true;
  
  bool_near_internal_change = bool_far_internal_change = false;
  bool_near_limits_internal_change = bool_far_limits_internal_change = false;
  
  FOV.SetMinMax(1.0/3600.0,150.0);
  
  center_X_rotation.SetMinMax(0.0,360.0);
  center_Y_rotation.SetMinMax(0.0,360.0);
  center_Z_rotation.SetMinMax(0.0,360.0);
  
  eye_X_rotation.SetMinMax(0.0,360.0);
  eye_Y_rotation.SetMinMax(0.0,360.0);
  eye_Z_rotation.SetMinMax(0.0,360.0);
  
  // fonts
  
#if !defined(_WIN32) && !defined(Q_OS_MACX)
  QFont default_font = font();
  default_font.setFamily("fixed");
  default_font.setRawMode(true);
  default_font.setFamily("fixed");
  default_font.setStyleStrategy(QFont::OpenGLCompatible);
  default_font.setPixelSize(10);
  setFont(default_font);
#endif
  
  char err_msg[1024];
  sprintf(err_msg,"font raw name: %s",font().rawName().latin1());
  ORSA_ERROR_MACRO(err_msg);
  
  /* 
     const QGLContext * gl_context = context(); 
     if (gl_context) {
     listBase = glGenLists(256);
     gl_context->generateFontDisplayLists(default_font,listBase); 
     }
  */
  
  init_signals_slots();
}

void XOrsaOpenGLWidget::init_signals_slots() {
  // animation
  connect(&animation_timer,    SIGNAL(timeout()), this, SLOT(animate()));
  connect(&bool_animate,       SIGNAL(changed()), this, SLOT(slot_bool_animate_changed()));
  connect(&animation_delay_ms, SIGNAL(changed()), this, SLOT(slot_animation_delay_changed()));
  // camera
  connect(&distance,       SIGNAL(changed()),this,SLOT(updateGL()));  
  connect(&distance,       SIGNAL(changed()),this,SLOT(slot_near_and_far_limit_on_distance_changed()));  
  //
  connect(&FOV,            SIGNAL(changed()),this,SLOT(slot_update_eye_rotation_impulse()));
  connect(&FOV,            SIGNAL(changed()),this,SLOT(slot_call_resizeGL()));
  connect(&FOV,            SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&near,           SIGNAL(limits_changed()),this,SLOT(slot_near_limits_changed()));
  connect(&near,           SIGNAL(changed()),this,SLOT(slot_call_resizeGL()));
  connect(&near,           SIGNAL(changed()),this,SLOT(slot_near_changed()));
  //
  connect(&far,            SIGNAL(limits_changed()),this,SLOT(slot_far_limits_changed()));
  connect(&far,            SIGNAL(changed()),this,SLOT(slot_call_resizeGL()));
  connect(&far,            SIGNAL(changed()),this,SLOT(slot_far_changed()));
  //
  connect(&near_and_far_limit_on_distance, SIGNAL(changed()),this,SLOT(slot_near_and_far_limit_on_distance_changed()));  
  //
  connect(&projection,         SIGNAL(changed()),this,SLOT(slot_call_resizeGL()));
  //
  connect(&ortho_xy_scale,     SIGNAL(changed()),this,SLOT(slot_call_resizeGL()));
  //
  connect(&ortho_z_near_scale, SIGNAL(changed()),this,SLOT(slot_call_resizeGL()));
  //
  connect(&ortho_z_far_scale,  SIGNAL(changed()),this,SLOT(slot_call_resizeGL()));
  // center rotation
  connect(&center_X_rotation, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&center_Y_rotation, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&center_Z_rotation, SIGNAL(changed()),this,SLOT(updateGL()));
  // eye rotation
  connect(&eye_X_rotation, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&eye_Y_rotation, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&eye_Z_rotation, SIGNAL(changed()),this,SLOT(updateGL()));
}

void XOrsaOpenGLWidget::slot_bool_animate_changed() {
  if (bool_animate) {
    animation_timer.start(animation_delay_ms);
  } else {
    animation_timer.stop();
  }
}

void XOrsaOpenGLWidget::slot_animation_delay_changed() {
  // cerr << "new animation delay: " << animation_delay_ms << endl;
  fps.reset();
  animation_timer.stop();
  if (bool_animate) {
    animation_timer.start(animation_delay_ms);
  }
}

void XOrsaOpenGLWidget::slot_update_eye_rotation_impulse() {
  eye_rotation_impulse = FOV/MIN(width(),height());
}

void XOrsaOpenGLWidget::slot_near_changed() {
  if (!near_and_far_limit_on_distance) {
    if (!bool_near_internal_change) {
      bool_near_internal_change = true;
      far.SetMin(near);
      bool_near_internal_change = false;
    }
  }
}

void XOrsaOpenGLWidget::slot_far_changed() {
  if (!near_and_far_limit_on_distance) {
    if (!bool_far_internal_change) {
      bool_far_internal_change = true;
      near.SetMax(far);
      bool_far_internal_change = false;
    }
  }
}

void XOrsaOpenGLWidget::slot_near_limits_changed() {
  if (near_and_far_limit_on_distance) {
    if (!bool_near_limits_internal_change) {
      bool_near_limits_internal_change = true;
      near.SetMax(distance);
      bool_near_limits_internal_change = false;
    }
  }
}

void XOrsaOpenGLWidget::slot_far_limits_changed() {
  if (near_and_far_limit_on_distance) {
    if (!bool_far_limits_internal_change) {
      bool_far_limits_internal_change = true;
      far.SetMin(distance);
      bool_far_limits_internal_change = false;
    }
  }
}

void XOrsaOpenGLWidget::slot_near_and_far_limit_on_distance_changed() {
  if (near_and_far_limit_on_distance) {
    near.SetMax(distance);
    far.SetMin(distance);
  } else {
    near.SetMax(far);
    far.SetMin(near);
  }
}

void XOrsaOpenGLWidget::initializeGL() {
  
#if !defined(_WIN32) && !defined(Q_OS_MACX)
  {    
    const GLubyte *vendor      = glGetString(GL_VENDOR);
    const GLubyte *renderer    = glGetString(GL_RENDERER);
    const GLubyte *version     = glGetString(GL_VERSION);
    const GLubyte *extensions  = glGetString(GL_EXTENSIONS);
    
    Display * dpy = glXGetCurrentDisplay();
    int GLX_major, GLX_minor;
    glXQueryVersion(dpy,&GLX_major,&GLX_minor);
    
    char err_msg[4096];
    sprintf(err_msg,"OpenGL Vendor: %s",(char *)vendor);
    ORSA_ERROR_MACRO(err_msg);
    sprintf(err_msg,"OpenGL Renderer: %s",(char *)renderer);
    ORSA_ERROR_MACRO(err_msg); 
    sprintf(err_msg,"OpenGL Version: %s",(char *)version);
    ORSA_ERROR_MACRO(err_msg);
    sprintf(err_msg,"GLX Version: %i.%i",GLX_major,GLX_minor);
    ORSA_ERROR_MACRO(err_msg); 
    sprintf(err_msg,"OpenGL Estensions: %s",(char *)extensions);
    ORSA_ERROR_MACRO(err_msg);
  }
#endif
  
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_NORMALIZE);
  
  // point anti-aliasing
  glEnable(GL_POINT_SMOOTH); 
  glHint(GL_POINT_SMOOTH_HINT,GL_NICEST);
  
  /* 
     { 
     // debug
     GLfloat values[2];
     glGetFloatv (GL_LINE_WIDTH_GRANULARITY, values);
     printf ("GL_LINE_WIDTH_GRANULARITY value is %3.1f\n", values[0]); 
     glGetFloatv (GL_LINE_WIDTH_RANGE, values); 
     printf ("GL_LINE_WIDTH_RANGE values are %3.1f %3.1f\n", values[0], values[1]);
     }	
  */
  
  // line anti-aliasing
  /* glEnable(GL_LINE_SMOOTH); // bad results (lines too big), maybe blending would solve?
     glEnable(GL_BLEND);
     // glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
     // the right one has to be called when needed, not here
     // glBlendFunc(GL_SRC_ALPHA,GL_ONE);
     glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
  */
  
  // glHint(GL_POLYGON_SMOOTH_HINT,GL_NICEST);
}

void XOrsaOpenGLWidget::resizeGL(int width, int height) {
  
  // cerr << "inside resizeGL()..." << endl;
  
  // This call to makeCurrent() is needed to avoid 
  // some odd problems when using more than one XOrsaOpenGLWidget
  // from the same application
  makeCurrent();
  
  glViewport(0,0,width,height);
  
  GLdouble w, h;
  
  if (width < height) {
    w = 1.0;
    h = (double) height / (double) width;
  } else {
    w = (double) width / (double) height;
    h = 1.0;
  }
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  //
  switch (projection) {
  case OGL_ORTHO:  
    glOrtho(-ortho_xy_scale*w, ortho_xy_scale*w, -ortho_xy_scale*h, ortho_xy_scale*h, -ortho_z_near_scale, ortho_z_far_scale);
    break;
  case OGL_PERSPECTIVE:
    gluPerspective(FOV,w/h,near,far);
    break;
  }
  
  slot_update_eye_rotation_impulse();
}

void XOrsaOpenGLWidget::paintGL() {
  
  // cerr << "inside paintGL()..." << endl;
  
  glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
  
  glMatrixMode(GL_MODELVIEW);
  
  glLoadIdentity();
  
  glRotated(eye_X_rotation, 1.0, 0.0, 0.0 );
  glRotated(eye_Y_rotation, 0.0, 1.0, 0.0 );
  glRotated(eye_Z_rotation, 0.0, 0.0, 1.0 );
  
  if (projection==OGL_PERSPECTIVE) glTranslated(0,0,-distance);
  // if (projection==OGL_PERSPECTIVE) glTranslatef(0,0,-(float)distance);
  
  glRotated(center_X_rotation, 1.0, 0.0, 0.0 );
  glRotated(center_Y_rotation, 0.0, 1.0, 0.0 );
  glRotated(center_Z_rotation, 0.0, 0.0, 1.0 );
  
  draw();
  
  if (bool_animate) {
    fps.sample();
    // cerr << "estimated FPS: " << fps << endl;
  }
  
  // cerr << "out of paintGL()..." << endl;
}

void XOrsaOpenGLWidget::animate() {
  
}

void XOrsaOpenGLWidget::showEvent(QShowEvent *e) {
  if (bool_animate && (!animation_timer.isActive()))
    animation_timer.start(animation_delay_ms);
  QGLWidget::showEvent(e);
}

void XOrsaOpenGLWidget::hideEvent(QHideEvent *e) {
  bool_animate = animation_timer.isActive();
  animation_timer.stop();
  QGLWidget::hideEvent(e);
}

void XOrsaOpenGLWidget::mousePressEvent(QMouseEvent *e) {
  e->accept();
  old_mouse_position = e->pos();
}

void XOrsaOpenGLWidget::mouseReleaseEvent(QMouseEvent *e) {
  e->accept();
  old_mouse_position = e->pos();
}

void XOrsaOpenGLWidget::mouseMoveEvent(QMouseEvent *e) {
  
  const double dx = e->x() - old_mouse_position.x();
  const double dy = e->y() - old_mouse_position.y();
  
  e->accept();
  
  old_mouse_position = e->pos();
  
  /* 
     const double rx = dx / width();
     const double ry = dy / height();
  */
  //
  /* 
     const double window_parameter = MAX(1.0,MIN(width(),height()));
     //
     const double rx = dx / window_parameter;
     const double ry = dy / window_parameter;
  */
  //
  const double rx = dx;
  const double ry = dy;
  
  if (e->state() & Qt::LeftButton) {
    // rotate the center
    if (e->state() & Qt::KeyButtonMask) {
      center_X_rotation += center_rotation_impulse * ry;
      center_Y_rotation += center_rotation_impulse * rx;
    } else {
      center_X_rotation += center_rotation_impulse * ry;
      center_Z_rotation += center_rotation_impulse * rx; 
    }
  } else if (e->state() & Qt::RightButton) {
    switch (projection) {
    case OGL_ORTHO:  
      // rotate the center
      if (e->state() & Qt::KeyButtonMask) {
	center_X_rotation += center_rotation_impulse * ry;
	center_Z_rotation += center_rotation_impulse * rx;
      } else {
	center_X_rotation += center_rotation_impulse * ry;
	center_Y_rotation += center_rotation_impulse * rx; 
      }
      break;
    case OGL_PERSPECTIVE:
      // rotate the eye
      if (e->state() & Qt::KeyButtonMask) {
	eye_X_rotation -= eye_rotation_impulse * ry;
	eye_Z_rotation -= eye_rotation_impulse * rx;
      } else {
	eye_X_rotation -= eye_rotation_impulse * ry;
	eye_Y_rotation -= eye_rotation_impulse * rx;
      }
      break;
    }
  } 
}

void XOrsaOpenGLWidget::wheelEvent(QWheelEvent *e) {
  e->accept();
  const double factor = pow(1.05,(e->delta()/120));
  switch (projection) {
  case OGL_ORTHO:  
    ortho_xy_scale *= factor;
    break;
  case OGL_PERSPECTIVE:
    FOV *= factor;
    break;
  }
}

/* 
   void XOrsaOpenGLWidget::renderText(int x, int y, const QString &str, const QFont &fnt) const {
   ORSA_ERROR_MACRO("XOrsaOpenGLWidget::renderText(...) code not ready...");
   //
   QGLWidget::renderText(x,y,"",fnt);
   const int fontsize=MAX(fnt.pixelSize(),fnt.pointSize());
   gl2psText(str,"Times-Roman",fontsize);
   QGLWidget::renderText(x,y,str,fnt);
   }
*/

/* 
   void XOrsaOpenGLWidget::renderText(double x, double y, double z, const QString &str, const QFont &fnt) const {
   ORSA_ERROR_MACRO("XOrsaOpenGLWidget::renderText(...) code not ready...");
   //
   QGLWidget::renderText(x,y,z,"",fnt);
   const int fontsize=MAX(fnt.pixelSize(),fnt.pointSize());
   gl2psText(str,"Times-Roman",fontsize);
   QGLWidget::renderText(x,y,z,str,fnt);
   }
*/

void XOrsaOpenGLWidget::export_file(const QString &s_filename, const OpenGL_Export_File file_type) {
  
  FILE *fp = fopen(s_filename.latin1(), "wb");
  GLint buffsize = 0, state = GL2PS_OVERFLOW;
  GLint viewport[4];
  
  glGetIntegerv(GL_VIEWPORT, viewport);
  
  while (state==GL2PS_OVERFLOW) { 
    
    buffsize += 1024*1024;
    
    gl2psBeginPage ("ORSA OpenGL view","ORSA - http://orsa.sourceforge.net",viewport,
		    file_type, GL2PS_BSP_SORT,
		    GL2PS_SILENT | GL2PS_DRAW_BACKGROUND | GL2PS_BEST_ROOT | GL2PS_OCCLUSION_CULL,
		    GL_RGBA, 0, 0, 0, 0, 0, buffsize,
		    fp, s_filename.latin1());
    
    paintGL();
    
    state = gl2psEndPage();
  }
  
  fclose(fp);
}

void XOrsaOpenGLWidget::export_png() {
  
  makeCurrent();
  
  QImage *tmpImage = new QImage(grabFrameBuffer());
  
  if (!tmpImage) {
    ORSA_ERROR_MACRO("XOrsaOpenGLWidget::export_png(): unexpected problem...");
    return;
  }
  
  const QString s = QFileDialog::getSaveFileName(QString::null,QString::null,this,0,"save image as:");
  
  if (!(s.isEmpty())) {
    tmpImage->save(s,"PNG");
  }
  
  delete tmpImage;
}

void XOrsaOpenGLWidget::export_ps() {
  const QString s = QFileDialog::getSaveFileName(QString::null,QString::null,this,0,"save PostScript file as:");
  if (!(s.isEmpty())) {
    export_file(s, OGL_PS_FILE);
  } 
}

void XOrsaOpenGLWidget::export_pdf() {
  const QString s = QFileDialog::getSaveFileName(QString::null,QString::null,this,0,"save PDF file as:");
  if (!(s.isEmpty())) {
    export_file(s, OGL_PDF_FILE);
  } 
}

////////////

// XOrsaOpenGLEvolutionWidget

XOrsaOpenGLEvolutionWidget::XOrsaOpenGLEvolutionWidget(QWidget *parent, WFlags f) : XOrsaOpenGLWidget(parent,f) {
  init();  
}

XOrsaOpenGLEvolutionWidget::XOrsaOpenGLEvolutionWidget(int w, int h, QWidget *parent, WFlags f) : XOrsaOpenGLWidget(w,h,parent,f) {
  init();
}

void XOrsaOpenGLEvolutionWidget::init() {
  
  max_circle_index = 10; // never more than 20, eats too much memory
  
  frame_changed_from_animation = false;
  
  // signals_and_slots
  connect(&range,        SIGNAL(changed()),this,SLOT(range_changed()));
  //
  connect(&distance,     SIGNAL(changed()),this,SLOT(distance_changed()));
  //
  connect(&evol_counter, SIGNAL(changed()),this,SLOT(evol_counter_changed()));
  connect(&evol_counter, SIGNAL(changed()),this,SLOT(update_distance()));
  connect(&evol_counter, SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&center_body,  SIGNAL(changed()),this,SLOT(center_body_changed()));
  connect(&center_body,  SIGNAL(changed()),this,SLOT(update_distance()));
  connect(&center_body,  SIGNAL(changed()),this,SLOT(distance_changed()));
  connect(&center_body,  SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&eye_body,     SIGNAL(changed()),this,SLOT(eye_body_changed()));
  connect(&eye_body,     SIGNAL(changed()),this,SLOT(update_distance()));
  connect(&eye_body,     SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&eye_on_body,  SIGNAL(changed()),this,SLOT(eye_on_body_changed()));
  connect(&eye_on_body,  SIGNAL(changed()),this,SLOT(update_distance()));
  connect(&eye_on_body,  SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&rotation_body,             SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&rotate_with_rotation_body, SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&draw_orbits,                SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&orbit_reference_body_index, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&bright_positive_Z,          SIGNAL(changed()),this,SLOT(updateGL()));
  //  
  connect(&draw_MOID,   SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&MOID_body_1, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&MOID_body_2, SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&draw_Lagrange_points, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&Lagrange_body_1,      SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&Lagrange_body_2,      SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&draw_tracks,     SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&absolute_tracks, SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&draw_reference_system, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&draw_XY_grid,          SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&draw_ra_dec_grid,      SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&draw_Z_bars,           SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&draw_scale_bar,        SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&light_time_correction, SIGNAL(changed()),this,SLOT(updateGL()));
  //
  connect(&draw_labels, SIGNAL(changed()),this,SLOT(updateGL()));
  connect(&labels_mode, SIGNAL(changed()),this,SLOT(updateGL())); 
  
  evolution = 0;
  //
  center_body = 0;
  //
  eye_body = 0;
  eye_on_body = false;
  //
  rotation_body = 0;
  rotate_with_rotation_body = false;
  
  evol_counter.SetSize(1);
  evol_counter = 0;
  //
  range = 0.0;
  last_size_checked = 0;
  
  projection = OGL_ORTHO;
  //
  bool_animate = false;
  //
  center_rotation_impulse.SetMinMax(0.01,100.0);
  center_rotation_impulse = 0.1;
  //
  eye_rotation_impulse.SetMinMax(1.0e-8,100.0); // small minimum limit, useful with big zooms...
  eye_rotation_impulse    = 0.1;
  //
  distance.SetMinMax(FromUnits(0.1,AU),FromUnits(20,AU));
  distance = FromUnits(4.0,AU);
  //
  FOV.SetMinMax(1.0/3600.0,150.0);
  FOV = 50.0;
  //
  near.SetMinMax(FromUnits(0.1,AU),FromUnits(20,AU));
  near = FromUnits(0.1,AU);
  //
  far.SetMinMax(FromUnits(0.1,AU),FromUnits(20,AU)); 
  far = FromUnits(20.0,AU);
  //
  ortho_xy_scale.SetMinMax(FromUnits(0.1,AU),FromUnits(20,AU)); 
  ortho_xy_scale = FromUnits(3.0,AU);
  //
  ortho_z_near_scale.SetMinMax(FromUnits(0.1,AU),FromUnits(20,AU)); 
  ortho_z_near_scale  = FromUnits(10.0,AU);
  //
  ortho_z_far_scale.SetMinMax(FromUnits(0.1,AU),FromUnits(20,AU)); 
  ortho_z_far_scale  = FromUnits(10.0,AU);
  
  // object_names_base_list = 0;
}

void XOrsaOpenGLEvolutionWidget::initializeGL() {
  
  XOrsaOpenGLWidget::initializeGL();
  
  // unitary XY circles
  circle = glGenLists(max_circle_index);
  int list_index=0;
  while (list_index < max_circle_index) {
    glNewList(circle+list_index, GL_COMPILE);
    glBegin(GL_LINE_LOOP);
    const int n_points = 1<<list_index; // 2^k points
    double theta;
    double s,c;
    const double theta_incr = twopi/n_points;
    int count = 0;
    while (count < n_points) {
      theta = count * theta_incr;
#ifdef HAVE_SINCOS
      sincos(theta,&s,&c);
#else // HAVE_SINCOS
      s = sin(theta);
      c = cos(theta);
#endif // HAVE_SINCOS
      glVertex3d(c,s,0);
      ++count;
    }
    glEnd();
    glEndList();
    ++list_index;
  }
  
  reference_system = glGenLists(1);
  glNewList(reference_system, GL_COMPILE);
  glLineWidth(2.0);
  gl2psLineWidth(2.0);
  if (max_circle_index > 8) {
    glCallList(circle+8);
  } else {
    glCallList(circle+max_circle_index-1);
  }
  glBegin(GL_LINES);
  glVertex3d(0,0,0);
  glVertex3d(0,0,1);
  glEnd();
  glBegin(GL_LINES);
  glVertex3d(0,0,0);
  glVertex3d(1,0,0);
  glEnd();
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  glEndList();
  
  // X-Y grid
  xy_grid = glGenLists(1);
  glNewList(xy_grid, GL_COMPILE);
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  int l;
  for (l=-100;l<=100;++l) {
    glBegin(GL_LINES);
    glVertex3d(l,-100,0);
    glVertex3d(l,+100,0);
    glEnd();
    //
    glBegin(GL_LINES);
    glVertex3d(-100,l,0);
    glVertex3d(+100,l,0);
    glEnd();
  }
  glEndList();
  
  // earth vector
  earth_vector = glGenLists(1);
  glNewList(earth_vector, GL_COMPILE);
  glLineWidth(1.0);
  gl2psLineWidth(1.0);
  const lat_lon ll_new_poly = {NEW_POLY,NEW_POLY};
  const lat_lon ll_final_element = {FINAL_ELEMENT,FINAL_ELEMENT};
  lat_lon ll;
  double phi,delta;
  unsigned int world_index = 0;
  glBegin(GL_LINE_STRIP);
  while ((ll=world[world_index]) != ll_final_element) {
    if (ll == ll_new_poly) {
      glEnd();
      glBegin(GL_LINE_STRIP);
    } else {
      phi   = (pi/180.0)*ll.lon;
      delta = (pi/180.0)*ll.lat;
      glVertex3d(cos(phi)*cos(delta),
		 sin(phi)*cos(delta),
		 sin(delta));
    }
    ++world_index;
  }
  glEnd();
  glEndList();
  
  {
    // earth
    
    QImage buffer;
    
    Config conf;
    //
    {
      OrsaConfigFile ocf(&conf);
      ocf.Read();
      ocf.Close();
    }
    //
    if (!buffer.load(conf.paths[TEXTURE_EARTH]->GetValue().c_str())) {
      ORSA_ERROR_MACRO("no Earth texture file....");
      QImage dummy(128,128,32);
      dummy.fill(Qt::blue.rgb());
      // dummy.fill(Qt::white.rgb());
      buffer = dummy;
    }
    
    QImage earth_texture = QGLWidget::convertToGLFormat(buffer);  // flipped 32bit RGBA
    
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    
    earth = glGenLists(1);
    glNewList(earth, GL_COMPILE);
    //
    glEnable(GL_TEXTURE_2D);
    glTexImage2D( GL_TEXTURE_2D, 0, 3, earth_texture.width(), earth_texture.height(), 0,
		  GL_RGBA, GL_UNSIGNED_BYTE, earth_texture.bits() );
    //
    GLUquadricObj * qobj = gluNewQuadric();
    //
    gluQuadricTexture(qobj,GL_TRUE);
    gluSphere(qobj,radius(EARTH),180,90);
    gluDeleteQuadric(qobj);
    glDisable(GL_TEXTURE_2D);
    glEndList();
    
  }
  
  {
    // moon
    
    QImage buffer;
    
    Config conf;
    //
    {
      OrsaConfigFile ocf(&conf);
      ocf.Read();
      ocf.Close();
    }
    //
    if (!buffer.load(conf.paths[TEXTURE_MOON]->GetValue().c_str())) {
      ORSA_ERROR_MACRO("no Moon texture file....");
      QImage dummy(128,128,32);
      // dummy.fill(Qt::blue.rgb());
      dummy.fill(Qt::white.rgb());
      buffer = dummy;
    }
    
    QImage moon_texture = QGLWidget::convertToGLFormat(buffer);  // flipped 32bit RGBA
    
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );
    glTexParameterf( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
    
    moon = glGenLists(1);
    glNewList(moon, GL_COMPILE);
    //
    glEnable(GL_TEXTURE_2D);
    glTexImage2D( GL_TEXTURE_2D, 0, 3, moon_texture.width(), moon_texture.height(), 0,
		  GL_RGBA, GL_UNSIGNED_BYTE, moon_texture.bits() );
    //
    GLUquadricObj * qobj = gluNewQuadric();
    //
    gluQuadricTexture(qobj,GL_TRUE);
    gluSphere(qobj,radius(MOON),180,90);
    gluDeleteQuadric(qobj);
    glDisable(GL_TEXTURE_2D);
    glEndList();
  }
  
}

void XOrsaOpenGLEvolutionWidget::draw() {
  
  // the eye position vector must be updated here and only here
  update_EyePosition();
  
  // cerr << "XOrsaOpenGLEvolutionWidget::draw() called..." << endl;
  
  // cerr << "distance: " << distance << endl;
  
  /* 
     { 
     //debug
     const Vector eye_position = EyePosition();
     cerr << "eye position: " << eye_position.x << " " << eye_position.y << " " << eye_position.z << endl;
     }
  */
  
  glPushMatrix();
  
  if (evolution) {
    if (evolution->size()) {
      
      // reference frame rotation HERE!
      if (rotate_with_rotation_body) {
	
	const Vector dir_pos = RotationBodyPosition() - CenterBodyPosition();
	
	const double phi   = secure_atan2(dir_pos.y,dir_pos.x);
	const double theta = secure_atan2(dir_pos.z,secure_sqrt(dir_pos.x*dir_pos.x+dir_pos.y*dir_pos.y)); 
	
	glRotated(-theta*(180.0/pi),   0.0,1.0,0.0);
	glRotated(180.0-phi*(180.0/pi),0.0,0.0,1.0);
      }	
      
      /* 
	 {
	 // debug: how is transate really working?
	 glPushMatrix();
	 //
	 double m_pre[16];
	 glGetDoublev(GL_MODELVIEW_MATRIX,m_pre);
	 //
	 double inverted_pre[16];
	 invert(m_pre,inverted_pre);
	 //
	 double v[3];
	 //
	 v[0] = -CenterBodyPosition().x;
	 v[1] = -CenterBodyPosition().y;
	 v[2] = -CenterBodyPosition().z;
	 //
	 fprintf(stderr,"v: %20.16f %20.16f %20.16f\n",v[0],v[1],v[2]);
	 //
	 glTranslated(v[0],v[1],v[2]);
	 //
	 double m_post[16];
	 glGetDoublev(GL_MODELVIEW_MATRIX,m_post);
	 //
	 double inverted_post[16];
	 invert(m_post,inverted_post);
	 //
	 // -----------------
	 // output now!
	 // cerr << "translation in m_pre:  " << m_pre[12]  << " " << m_pre[13]  << " " << m_pre[14]  << endl;
	 // cerr << "translation in m_post: " << m_post[12] << " " << m_post[13] << " " << m_post[14] << endl;
	 fprintf(stderr,"translation in m_pre:  %20.16f %20.16f %20.16f\n",m_pre[12], m_pre[13], m_pre[14]);
	 fprintf(stderr,"translation in m_post: %20.16f %20.16f %20.16f\n",m_post[12],m_post[13],m_post[14]);
	 //
	 double w[3];
	 //
	 w[0] = v[0]*m_pre[0]+v[1]*m_pre[4]+v[2]*m_pre[8];
	 w[1] = v[0]*m_pre[1]+v[1]*m_pre[5]+v[2]*m_pre[9];
	 w[2] = v[0]*m_pre[2]+v[1]*m_pre[6]+v[2]*m_pre[10];
	 //
	 fprintf(stderr,"w: %20.16f %20.16f %20.16f\n",w[0],w[1],w[2]);
	 // back to the original v = z
	 double z[3];
	 //
	 w[0] = m_post[12]-m_pre[12];
	 w[1] = m_post[13]-m_pre[13];
	 w[2] = m_post[14]-m_pre[14];
	 //
	 z[0] = w[0]*inverted_pre[0]+w[1]*inverted_pre[4]+w[2]*inverted_pre[8];
	 z[1] = w[0]*inverted_pre[1]+w[1]*inverted_pre[5]+w[2]*inverted_pre[9];
	 z[2] = w[0]*inverted_pre[2]+w[1]*inverted_pre[6]+w[2]*inverted_pre[10];
	 //
	 fprintf(stderr,"z: %20.16f %20.16f %20.16f\n",z[0],z[1],z[2]);
	 //
	 fprintf(stderr,"v-z: %20.16f %20.16f %20.16f\n",v[0]-z[0],v[1]-z[1],v[2]-z[2]);
	 // note: v-z is the difference between the translation requested and 
	 //       the real OpenGL translation: the two differ for many reasons,
	 //       especially because OpenGL internally uses floats... :(
	 //
	 glPopMatrix();
	 }
      */
      
      // [x]
      // glTranslate(-CenterBodyPosition());
      
      if (draw_orbits || draw_MOID) {
	if (orbit_reference_body_index==-HEX_AUTO) {
	  if (orbit_reference_body_index_vector[evol_counter].size()==0) {
	    vector<int> ref_index;
	    AutoOrbitIndex(evol_frame,ref_index);
	    orbit_reference_body_index_vector[evol_counter] = ref_index;
	  }
	}
      }
      
      internal_draw_bodies_and_labels();
      
      // transparent lines rendering should be the last one
      if (draw_orbits) {
	internal_draw_orbit_ellipse();
      }
      
      if (draw_Lagrange_points) {
	internal_draw_Lagrange_points();
      }
      
      if (draw_MOID) {
	internal_draw_MOID();
      }
      
      // last one...
      internal_draw_OSD();
      
      // test: transparent reference plane
      /* 
	 {
	 glPushMatrix();
	 glDepthMask(GL_FALSE); // depth-buffer read-only
	 // glBlendFunc(GL_SRC_ALPHA,GL_ONE);
	 glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	 // glBlendFunc(GL_SRC_ALPHA,GL_DST_COLOR);
	 // glBlendFunc(GL_ONE,GL_ZERO);
	 // glBlendFunc(GL_SRC_ALPHA,GL_ZERO);
	 // glColor4d(0.2,0.4,1.0,0.5);
	 glColor4d(0.2,0.4,1.0,0.6);
	 // wto quads are needed!
	 // visible from z>0
	 glBegin(GL_QUADS);
	 glVertex3d(+ortho_z_scale,+ortho_z_scale,0);
	 glVertex3d(-ortho_z_scale,+ortho_z_scale,0);
	 glVertex3d(-ortho_z_scale,-ortho_z_scale,0);
	 glVertex3d(+ortho_z_scale,-ortho_z_scale,0);
	 glEnd();
	 //
	 glColor4d(0.2,0.4,1.0,0.2);	
	 //
	 // visible from z<0
	 glBegin(GL_QUADS);
	 glVertex3d(+ortho_z_scale,+ortho_z_scale,0);
	 glVertex3d(+ortho_z_scale,-ortho_z_scale,0);
	 glVertex3d(-ortho_z_scale,-ortho_z_scale,0);
	 glVertex3d(-ortho_z_scale,+ortho_z_scale,0);
	 glEnd();
	 //
	 // adding lighting can be nice...
	 //
	 glDepthMask(GL_TRUE);
	 glPopMatrix(); 
	 }
      */
      
    }
  } 
  
  glPopMatrix();
}

void XOrsaOpenGLEvolutionWidget::animate() {
  
  if (bool_animate==false) return;
  
  if (bool_already_animating) {
    return;
  }
  
  bool_already_animating = true;
  
  if (evolution == 0) return;
  
  // cerr << "increasing evol_counter..." << endl;
  //
  ++evol_counter;
  //
  // cerr << "evol_counter increased... " << endl;
  
  // IMPORTANT! This call to qApp->processEvents()
  // prevents a lot of problems when using more
  // than one OpenGL window in the same session
  // qApp->processEvents(10000);
  // qApp->processEvents(10);
  qApp->processEvents(1);
  
  /* check not needed if evol_counter is a SizeObject...
     if (evol_counter == (int)evolution->size()) {
     evol_counter = 0;
     } 
  */
  // evol_frame should be automatically updated
  // evol_frame = (*evolution)[evol_counter];
  
  orbit_reference_body_index_vector.resize(evolution->size());
  orbit_cache_vector.resize(evolution->size());
  
  // automatically called...
  // updateGL();
  
  frame_changed_from_animation = true;
  emit frame_changed(evol_counter);
  frame_changed_from_animation = false;
  bool_already_animating = false;
}

void XOrsaOpenGLEvolutionWidget::SetEvolution(const Evolution *evol_in) {
  if (evol_in == 0) {
    return;
  } else {
    evolution = evol_in;
    const unsigned int e_size = evolution->size();
    evol_counter.SetSize(e_size);
    evol_counter = 0;
    // need to call evol_counter_changed() manually 
    // here, because maybe evol_counter was already
    // equal to zero, so no signal is emitted and
    // evol_counter_changed() called.
    //
    evol_counter_changed();
    
    last_size_checked = 0;
    
    {
      const XOrsaEvolution * const oe = dynamic_cast <const XOrsaEvolution * const> (evolution);
      if (oe) {
	oe->add_event_receiver(integration_step_done_event_type,this);
	oe->add_event_receiver(integration_started_event_type,this);
	oe->add_event_receiver(integration_finished_event_type,this);
	oe->add_event_receiver(evolution_modified_event_type,this);
      } else {
	ORSA_ERROR_MACRO("I don't know why...");
      }
    }
    
    orbit_reference_body_index_vector.clear();
    orbit_reference_body_index_vector.resize(e_size);
    
    orbit_cache_vector.clear();
    orbit_cache_vector.resize(e_size);
    
    // generate_object_names();
    
    // needed
    update_range();
    
    emit evolution_changed();
  }
}

void XOrsaOpenGLEvolutionWidget::customEvent(QCustomEvent *e) {
  switch (e->type()) {
  case integration_step_done_event_type: 
    break;
  case integration_started_event_type:
    break;
  case integration_finished_event_type: 
    break;
  case evolution_modified_event_type: 
    update_sizes();
    update_range();
    break;
  case universe_modified_event_type:
    break;
  default:
    break;
  }
}

Vector XOrsaOpenGLEvolutionWidget::BodyPosition(const int i) const {
  if (i >= 0) {
    if (i < (int)evol_frame.size()) {
      return evol_frame[i].position();
    } else {
      ORSA_ERROR_MACRO("problem in XOrsaOpenGLEvolutionWidget::BodyPosition()...");
      return Vector(0,0,0);
    }	
  } else {
    if (i == -HEX_ORIGIN) {
      return Vector(0,0,0);
    } else if (i == -HEX_CENTER_OF_MASS) {
      return evol_frame.Barycenter();
    } else {
      ORSA_ERROR_MACRO("problem in XOrsaOpenGLEvolutionWidget::BodyPosition()...");
      return Vector(0,0,0);
    }
  }
}

string XOrsaOpenGLEvolutionWidget::BodyName(const int i) const {
  if (i >= 0) {
    if (i < (int)evol_frame.size()) {
      return evol_frame[i].name();
    } else {
      ORSA_ERROR_MACRO("problem in XOrsaOpenGLEvolutionWidget::BodyName()...");
      return "";
    }	
  } else {
    if (i == -HEX_ORIGIN) {
      return "";
    } else if (i == -HEX_CENTER_OF_MASS) {
      return "center of mass";
    } else {
      ORSA_ERROR_MACRO("problem in XOrsaOpenGLEvolutionWidget::BodyName()...");
      return "";
    }
  }
}

Vector XOrsaOpenGLEvolutionWidget::CenterBodyPosition() const {
  return BodyPosition(center_body);
}

string XOrsaOpenGLEvolutionWidget::CenterBodyName() const {
  return BodyName(center_body);
}

Vector XOrsaOpenGLEvolutionWidget::EyeBodyPosition() const {
  if (!eye_on_body) ORSA_ERROR_MACRO("Hmmm... something funny here...");
  return BodyPosition(eye_body);
}

string XOrsaOpenGLEvolutionWidget::EyeBodyName() const {
  if (!eye_on_body) ORSA_ERROR_MACRO("Hmmm... something funny here...");
  return BodyName(eye_body);
}

Vector XOrsaOpenGLEvolutionWidget::RotationBodyPosition() const {
  if (!rotate_with_rotation_body) ORSA_ERROR_MACRO("Hmmm... something funny here...");
  return BodyPosition(rotation_body);
}

string XOrsaOpenGLEvolutionWidget::RotationBodyName() const {
  if (!rotate_with_rotation_body) ORSA_ERROR_MACRO("Hmmm... something funny here...");
  return BodyName(rotation_body);
}

void XOrsaOpenGLEvolutionWidget::update_EyePosition() {
  if (projection != OGL_PERSPECTIVE) return;
  //
  double m[16];
  glGetDoublev(GL_MODELVIEW_MATRIX,m);
  //
  double inverted[16];
  invert(m,inverted);
  //
  double v[3];
  //
  v[0] = -m[12];
  v[1] = -m[13];
  v[2] = -m[14];
  //
  double w[3];
  //
  w[0] = v[0]*inverted[0]+v[1]*inverted[4]+v[2]*inverted[8];
  w[1] = v[0]*inverted[1]+v[1]*inverted[5]+v[2]*inverted[9];
  w[2] = v[0]*inverted[2]+v[1]*inverted[6]+v[2]*inverted[10];
  //
  eye_position.x = w[0];
  eye_position.y = w[1];
  eye_position.z = w[2];
  //
  eye_position += CenterBodyPosition();
}

void XOrsaOpenGLEvolutionWidget::update_sizes() {
  const unsigned int e_size = evolution->size();
  //
  orbit_reference_body_index_vector.resize(e_size);
  orbit_cache_vector.resize(e_size);
  evol_counter.SetSize(e_size);
}

void XOrsaOpenGLEvolutionWidget::update_range() {
  
  double x_min,x_max,y_min,y_max,z_min,z_max;
  //
  x_min=x_max=(*evolution)[0][0].position().x;
  y_min=y_max=(*evolution)[0][0].position().y;
  z_min=z_max=(*evolution)[0][0].position().z;
  //
  const unsigned int current_evolution_size = evolution->size();
  for (unsigned int j=last_size_checked;j<current_evolution_size;++j) {
    const Frame f = (*evolution)[j];
    for(unsigned int body=0;body<f.size();++body) {
      const double _x = f[body].position().x;
      x_min = MIN(x_min,_x);
      x_max = MAX(x_max,_x);
      const double _y = f[body].position().y;
      y_min = MIN(y_min,_y);
      y_max = MAX(y_max,_y);
      const double _z = f[body].position().z;
      z_min = MIN(z_min,_z);
      z_max = MAX(z_max,_z);
    }
  }
  //
  const double delta_x = x_max - x_min;
  const double delta_y = y_max - y_min;
  const double delta_z = z_max - z_min;
  //
  const double old_range = range;
  range = MAX(old_range,MAX(delta_x,MAX(delta_y,delta_z)));
  //
  last_size_checked = current_evolution_size;
}

void XOrsaOpenGLEvolutionWidget::export_movie() {
  
  const QString s_dir = QFileDialog::getExistingDirectory(QString::null,this,0,"save images into:",true);
  
  if (!(s_dir.isEmpty())) {
    
    makeCurrent();
    
    const unsigned int old_evol_counter = evol_counter;
    
    QString s_filename;
    unsigned int j=0;
    while (j < total_frames()) {
      
      s_filename.sprintf("frame-%06i.png",j);
      s_filename.prepend(s_dir);
      
      // set_frame(j);
      evol_counter = j;
      
      QImage *tmpImage = new QImage(grabFrameBuffer());
      
      if (!tmpImage) {
	ORSA_ERROR_MACRO("XOrsaOpenGLWidget::export_png(): unexpected problem...");
	return;
      }
      
      tmpImage->save(s_filename,"PNG");
      
      delete tmpImage;
      
      ++j;
    }
    
    evol_counter = old_evol_counter;
  }
}

void XOrsaOpenGLEvolutionWidget::internal_draw_bodies_and_labels() {
  
  for (unsigned int j=0;j<evol_frame.size();++j) {
    
    glPushMatrix();
    
    // IMPORTANT: this difference (pos - center) is REALLY needed:
    // allows to have the central body in (0,0,0), avoiding
    // some annoying shaking effects.
    //
    const Vector p = evol_frame[j].position() - CenterBodyPosition();
    
    if (evol_frame[j].mass() != 0) {
      glPointSize(1);
      gl2psPointSize(1);
      glColor3d(1,0,0);
    } else {
      glPointSize(1);
      gl2psPointSize(1);
      glColor3d(0,0,1);
    }
    
    if (evol_frame[j].radius() > 0) {
      
      // [x]
      // glTranslate(p);
      // glTranslate(p-CenterBodyPosition());
      glTranslate(p);
      
      if (evol_frame[j].name() == "Earth") {
	
	GLfloat whiteDir[4] = {2.0, 2.0, 2.0, 1.0};
	GLfloat whiteAmb[4] = {1.0, 1.0, 1.0, 1.0};
	//
	// GLfloat lightPos[4] = {30.0, 30.0, 30.0, 1.0};
       	GLfloat lightPos[4]; /* relative position of the Sun
			      * with respect to the planet */
	{
	  // light-time delay?
	  Vector sunpos;
	  {
	    unsigned int q = 0;
	    while (q < evol_frame.size()) {
	      if (evol_frame[q].name() == "Sun") {
		sunpos = evol_frame[q].position();
		break;
	      }
	      ++q;
	    }
	  }
	  sunpos -= evol_frame[j].position(); // sunpos relative to Earth position
	  //
	  lightPos[0] = sunpos.x;
	  lightPos[1] = sunpos.y;
	  lightPos[2] = sunpos.z;
	  //
	  lightPos[3] = 1.0;
	}
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, whiteAmb);
	
        glMaterialfv(GL_FRONT, GL_DIFFUSE, whiteDir);
	glMaterialfv(GL_FRONT, GL_SPECULAR, whiteDir);
	glMaterialf(GL_FRONT, GL_SHININESS, 20.0);
	
	glLightfv(GL_LIGHT0, GL_DIFFUSE, whiteDir);  // enable diffuse
	glLightfv(GL_LIGHT0, GL_SPECULAR, whiteDir);	// enable specular
	glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
	
	glPushMatrix();	
	//
	if (universe->GetReferenceSystem() == ECLIPTIC) {
	  glRotated(-(180.0/pi)*obleq_J2000().GetRad(),1,0,0);
	}
	glRotated((180.0/pi)*gmst(evol_frame).GetRad(),0,0,1);
	glRotated(90,0,0,1); // ADD an extra rotation of 90 degrees when using a texture object!!
	glCallList(earth); 
	// 
	glPopMatrix();
	
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHTING);
	
	
      } else if (evol_frame[j].name() == "Moon") {
	
	GLfloat whiteDir[4] = {2.0, 2.0, 2.0, 1.0};
	GLfloat whiteAmb[4] = {1.0, 1.0, 1.0, 1.0};
	//
	// GLfloat lightPos[4] = {30.0, 30.0, 30.0, 1.0};
       	GLfloat lightPos[4]; /* relative position of the Sun
			      * with respect to the planet */
	{
	  // light-time delay?
	  Vector sunpos;
	  {
	    unsigned int q = 0;
	    while (q < evol_frame.size()) {
	      if (evol_frame[q].name() == "Sun") {
		sunpos = evol_frame[q].position();
		break;
	      }
	      ++q;
	    }
	  }
	  sunpos -= evol_frame[j].position(); // sunpos relative to Moon position
	  //
	  lightPos[0] = sunpos.x;
	  lightPos[1] = sunpos.y;
	  lightPos[2] = sunpos.z;
	  //
	  lightPos[3] = 1.0;
	}
	
	Vector EarthPosition;
	{
	  unsigned int q = 0;
	  while (q < evol_frame.size()) {
	    if (evol_frame[q].name() == "Earth") {
	      EarthPosition = evol_frame[q].position();
	      break;
	    }
	    ++q;
	  }
	  EarthPosition -= evol_frame[j].position(); // sunpos relative to Moon position
	}
	
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_FALSE);
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, whiteAmb);
	
	glMaterialfv(GL_FRONT, GL_DIFFUSE, whiteDir);
	glMaterialfv(GL_FRONT, GL_SPECULAR, whiteDir);
	glMaterialf(GL_FRONT, GL_SHININESS, 20.0);
	
	glLightfv(GL_LIGHT0, GL_DIFFUSE, whiteDir);  // enable diffuse
	glLightfv(GL_LIGHT0, GL_SPECULAR, whiteDir);	// enable specular
	glLightfv(GL_LIGHT0, GL_POSITION, lightPos);
	
	glPushMatrix();	
	//
	// NOTE: all the Moon rotations have to be checked!!!
	// the angles seem to be quite good, but probably
	// some fine-tuning is needed
	//
	{
	  Angle inclination;
	  inclination.SetDPS(1,32,32.7);
	  glRotated(-(180.0/pi)*inclination.GetRad(),1,0,0);
	}
	//
	if (universe->GetReferenceSystem() == EQUATORIAL) { 
	  glRotated((180.0/pi)*obleq_J2000().GetRad(),1,0,0);
	}
	//
	{
	  const double phi = secure_atan2(EarthPosition.y,EarthPosition.x);
	  glRotated(90+(180.0/pi)*phi,0,0,1); // 90 for the texture...
	}
	//
	glCallList(moon);
	//
	glPopMatrix();
	
	glDisable(GL_LIGHT0);
	glDisable(GL_LIGHTING);
	
      } else {
	
	// standard code...
	GLUquadricObj *qobj = gluNewQuadric();
	gluQuadricDrawStyle(qobj,GLU_SILHOUETTE);
	// gluQuadricDrawStyle(qobj,GLU_FILL);
	gluQuadricNormals(qobj,GLU_SMOOTH);
	gluSphere(qobj,evol_frame[j].radius(),36,18);
	gluDeleteQuadric(qobj);
	
      }
      
      glBegin(GL_POINTS);
      glVertex3d(0,0,0);
      glEnd();
      
    } else {
      glBegin(GL_POINTS);
      glVertex(p);
      glEnd();
    }
    
    glPopMatrix();
    
    /* if (draw_labels) {
       if (((evol_frame[j].mass() > 0) && (labels_mode == MASSIVE)) || 
       (labels_mode == ALL)) {
    */
    // glColor3d(0,0.7,0);
    QString text;
    text.sprintf(" %s",evol_frame[j].name().c_str());
    // renderText(p,text,default_font); 
    //
    //renderText(p,text); 
    //
    if ((int)j==center_body) {
      QFont fnt(font());
      // fnt.setBold(true);
      // fnt.setUnderline(true);
      fnt.setPixelSize(14);
      glColor3d(0.1,0.9,0.1);
      renderText(p,text,fnt);
    } else {
      glColor3d(0,1.0,0);
      renderText(p,text); 
    }
    //
    /* }
       }
    */
  }
}

void XOrsaOpenGLEvolutionWidget::internal_draw_orbit_ellipse() {
  
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE);
  glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
  glDepthMask(GL_FALSE);
  
  Orbit orbit;
  
  glPushMatrix();
  
  for (unsigned int j=0;j<evol_frame.size();++j) {
    
    if (evol_frame[j].mass() > 0.0) {
      glLineWidth(2.5);
      gl2psLineWidth(2.5);
    } else {
      glLineWidth(1.5);
      gl2psLineWidth(1.5);
    }
    
    if (orbit_reference_body_index==-HEX_AUTO) {
      if (((int)j)==orbit_reference_body_index_vector[evol_counter][j]) continue;
    } else {
      if (((int)j)==orbit_reference_body_index) continue;
    }
    
    Vector v_diff;
    if (orbit_reference_body_index==-HEX_AUTO) {
      // v_diff = evol_frame[orbit_reference_body_index_vector[evol_counter][j]].position()-center;
      v_diff = evol_frame[orbit_reference_body_index_vector[evol_counter][j]].position();
    } else {
      // v_diff = evol_frame[orbit_reference_body_index].position()-center;
      v_diff = evol_frame[orbit_reference_body_index].position();
    }
    
    const GLdouble equation_z_plus[4]  = {0,0, 1,-v_diff.z};
    const GLdouble equation_z_minus[4] = {0,0,-1, v_diff.z};
    //
    if (bright_positive_Z) {
      glClipPlane(GL_CLIP_PLANE0,equation_z_plus);
      glClipPlane(GL_CLIP_PLANE1,equation_z_minus);
    }
    
    glPushMatrix();
    
    // [x]
    // glTranslate(v_diff);
    glTranslate(v_diff-CenterBodyPosition());
    
    int ref_body_index;
    if (orbit_reference_body_index==-HEX_AUTO) {
      ref_body_index = orbit_reference_body_index_vector[evol_counter][j];
    } else {
      ref_body_index = orbit_reference_body_index;
    }
    
    if ( (orbit_cache_vector[evol_counter][j].set) &&
	 (orbit_cache_vector[evol_counter][j].ref_body_index == ref_body_index) ) {
      
      // cerr << "usign orbit cache!" << endl;
      
      orbit = orbit_cache_vector[evol_counter][j].orbit;
      
    } else {
      
      OrbitCache orbit_cache;
      
      orbit.Compute(evol_frame[j],evol_frame[ref_body_index]);
      
      orbit_cache.orbit          = orbit;
      orbit_cache.ref_body_index = ref_body_index;
      orbit_cache.set            = true;
      
      orbit_cache_vector[evol_counter][j] = orbit_cache;
    }
    
    if (orbit.e<1.0) {
      
      glPushMatrix();	      
      //
      glRotated(orbit.omega_node*(180.0/pi),0,0,1);
      glRotated(orbit.i*(180.0/pi),1,0,0);
      glRotated(orbit.omega_pericenter*(180.0/pi),0,0,1);
      //
      glTranslated(-orbit.a*orbit.e,0.0,0.0);
      //
      // glPushMatrix();
      glScaled(orbit.a,orbit.a*secure_sqrt(1-orbit.e*orbit.e),orbit.a);
      
      // test: a rotation to keep a point on the body: better accuracy when zooming?
      double E = orbit.M;
      double tmp_E;
      int count = 0;
      do {
        tmp_E = E;
	E = orbit.e * sin(E) + orbit.M;
        count++;
      } while ( (fabs((E-tmp_E)/E) > 1.0e-15) && (count < 100) );
      //
      glRotated(E*(180.0/pi),0,0,1);
      
      // select the appropriate circle
      //
      const int max_pixels_per_circle_element = 4; // should be tuned; the smaller this value, the better quality and more CPU
      // const int max_pixels_per_circle_element = 6; // should be tuned; the smaller this value, the better quality and more CPU
      // const int max_pixels_per_circle_element = 10; // should be tuned; the smaller this value, the better quality and more CPU
      //
      int appropriate_circle_index=0;
      switch(projection) {
      case OGL_ORTHO:
	// appropriate_circle_index = (int)rint(secure_log(orbit.a*(1.0+orbit.e)*pi/(max_pixels_per_circle_element*pixel_length))/secure_log(2.0)); // log base 2
	appropriate_circle_index = (int)rint(secure_log(2.0*pi*orbit.a*(1.0+orbit.e)/(max_pixels_per_circle_element*PixelLength()))/secure_log(2.0)); // log base 2
	break;
      case OGL_PERSPECTIVE:
	
	const double eye_to_orbit_center_distance = (evol_frame[ref_body_index].position()-eye_position).Length();
	//
	// naif method, doesn't take into account real orbit inclination and orientation...
	double minimum_eye_to_orbit_distance = 0.0;
	const double q = (orbit.a*(1.0-orbit.e));
	const double Q = (orbit.a*(1.0+orbit.e));
	if (eye_to_orbit_center_distance < q) {
	  minimum_eye_to_orbit_distance = q - eye_to_orbit_center_distance;
	} else if (eye_to_orbit_center_distance > Q) {
	  minimum_eye_to_orbit_distance = eye_to_orbit_center_distance - Q;
	}
	//
	// cerr << "minimum_eye_to_orbit_distance: " << minimum_eye_to_orbit_distance << endl;
	//
	// debug
	// if (evol_frame[j].name() == "Moon") {
	/* if ((int)j == center_body) {
	   cerr << "name: " << evol_frame[j].name() << endl;
	   cerr << "ref_body_index: " << ref_body_index << endl;
	   cerr << "ref_body name: " << evol_frame[ref_body_index].name() << endl;
	   //
	   {
	   // debug
	   const Vector ref_body_pos = evol_frame[ref_body_index].position();
	   cerr << "ref_body position: " << ref_body_pos.x << " " << ref_body_pos.y << " " << ref_body_pos.z << endl;
	   }
	   //
	   cerr << "distance: " << distance << endl;
	   cerr << "eye_to_orbit_center_distance: " << eye_to_orbit_center_distance << endl;
	   cerr << "minimum_eye_to_orbit_distance: " << minimum_eye_to_orbit_distance << endl;
	   }
	*/
	//
	if (minimum_eye_to_orbit_distance > 0.0) {
	  appropriate_circle_index = (int)rint(secure_log(2.0*pi*orbit.a*(1.0+orbit.e)/(minimum_eye_to_orbit_distance*max_pixels_per_circle_element*(pi/180.0)*PixelLength()))/secure_log(2.0)); // log base 2
	} else {
	  appropriate_circle_index=max_circle_index-1;
	}
	
	break;
      }
      
      if (appropriate_circle_index <  0) appropriate_circle_index=0;
      if (appropriate_circle_index >= max_circle_index) appropriate_circle_index=max_circle_index-1;
      // 
      // cerr << evol_frame[j].name() << " ellipse index: " << appropriate_circle_index << endl;
      
      if (appropriate_circle_index > 1) {
	
	const GLint appropriate_circle = circle+appropriate_circle_index;
	
	if (bright_positive_Z) {
	  glColor4d(0.2,0.2,0.2,1.0);
	  glEnable(GL_CLIP_PLANE0);
	  glCallList(appropriate_circle);
	  glDisable(GL_CLIP_PLANE0);
	  
	  glColor4d(0.1,0.1,0.1,1.0);
	  glEnable(GL_CLIP_PLANE1);
	  glCallList(appropriate_circle);
	  glDisable(GL_CLIP_PLANE1);
	} else {
	  glColor4d(0.2,0.2,0.2,1.0);
	  glCallList(appropriate_circle);
	}
	
      }
      
      glPopMatrix();
    }
    
    glPopMatrix();
  } 
  
  glPopMatrix();
  
  glDepthMask(GL_TRUE);   
  glDisable(GL_BLEND);
  glDisable(GL_LINE_SMOOTH);
}

void XOrsaOpenGLEvolutionWidget::internal_draw_MOID() {
#ifdef HAVE_GSL	
  
  glEnable(GL_LINE_SMOOTH);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA,GL_ONE);
  glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);
  glDepthMask(GL_FALSE);
  
  if (draw_orbits)
    if (evolution != 0)
      if (evolution->size() != 0)
     	if ((evol_frame.size() >= 3) && (MOID_body_1!=MOID_body_2)) {
	  
	  if (orbit_reference_body_index==-HEX_AUTO) {
	    if (orbit_reference_body_index_vector[evol_counter].size()==0) return;
	    if (MOID_body_1==orbit_reference_body_index_vector[evol_counter][MOID_body_1]) return;
	    if (MOID_body_2==orbit_reference_body_index_vector[evol_counter][MOID_body_2]) return;
	  } else {
	    if (MOID_body_1==orbit_reference_body_index) return;
	    if (MOID_body_2==orbit_reference_body_index) return;
	  }
	  
	  Orbit  o_1, o_2;
	  Vector r_1, r_2;
	  
	  if (orbit_reference_body_index==-HEX_AUTO) {
	    o_1.Compute(evol_frame[MOID_body_1],evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_1]]);
	    o_2.Compute(evol_frame[MOID_body_2],evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_2]]);
	  } else {
	    o_1.Compute(evol_frame[MOID_body_1],evol_frame[orbit_reference_body_index]);
	    o_2.Compute(evol_frame[MOID_body_2],evol_frame[orbit_reference_body_index]);
	  }
	  
	  // really needed the eccentricity check?	
	  // if ((o_1.e < 1.0) && (o_2.e < 1.0)) {
	  
	  double _MOID;
	  if ((orbit_reference_body_index==-HEX_AUTO) && (orbit_reference_body_index_vector[evol_counter][MOID_body_1]!=orbit_reference_body_index_vector[evol_counter][MOID_body_2])) { 
	    _MOID = MOID2RB(evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_1]].position(),
			    evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_2]].position(),
			    o_1,o_2,r_1,r_2);
	  } else {
	    _MOID = MOID(o_1,o_2,r_1,r_2);
	  }	
	  
	  // cerr << "MOID between bodies 1 and 2: " << _MOID << endl;
	  // cerr << "r1: " << r1.x << "  " << r1.y << "  " << r1.z << endl; 
	  // cerr << "r2: " << r2.x << "  " << r2.y << "  " << r2.z << endl; 
	  
	  Vector z1,z2;
	  if (orbit_reference_body_index==-HEX_AUTO) {
	    // z1 = evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_1]].position() - center;
	    // z2 = evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_2]].position() - center;
	    z1 = evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_1]].position();
	    z2 = evol_frame[orbit_reference_body_index_vector[evol_counter][MOID_body_2]].position();
	  } else {
	    // z1 = z2 = evol_frame[orbit_reference_body_index].position() - center;
	    z1 = z2 = evol_frame[orbit_reference_body_index].position();
	  }
	  
	  r_1 += z1;
	  r_2 += z2;
	  
	  // [x]
	  r_1 -= CenterBodyPosition();
	  r_2 -= CenterBodyPosition();
	  
	  glColor3d(0,1,1);
	  glLineWidth(1.0);
	  gl2psLineWidth(1.0);
	  // glLineStipple(1, 0x00FF);
	  // glEnable(GL_LINE_STIPPLE);
	  glBegin(GL_LINES);
	  glVertex3d(r_1.x,r_1.y,r_1.z);
	  glVertex3d(r_2.x,r_2.y,r_2.z);
	  glEnd();
	  // glDisable(GL_LINE_STIPPLE);
	  
	  char t[1024];
	  // sprintf(t," MOID: %.3g %s",_MOID,LengthLabel().c_str());
	  
	  length_unit lu = AutoLengthUnit(_MOID);
	  // sprintf(t," MOID: %.3g %s",_MOID,LengthLabel().c_str());
	  sprintf(t," MOID: %.3g %s",FromUnits(_MOID,lu,-1),units->label(lu).c_str());
	  
	  // string dpy_name = t;
	  Vector m = (r_1+r_2)/2;
	  // glRasterPos3d(m.x,m.y,m.z);
	  // glCallLists(dpy_name.size(),GL_BYTE,(GLubyte*) dpy_name.c_str());
	  // renderText(m.x,m.y,m.z,t,default_font); 
	  renderText(m,t); 
	  
	  // } 
	}
  
  glDepthMask(GL_TRUE);   
  glDisable(GL_BLEND);
  glDisable(GL_LINE_SMOOTH);
  
#endif // HAVE_GSL   
}

/* 
   void XOrsaOpenGLEvolutionWidget::generate_object_names() {
   // test (needed?)
   // renderText(0,0,"",default_font);
   //
   const Frame * b = bodies();
   const unsigned int size = b->size();
   string display_name;
   object_names_base_list = glGenLists(size);
   cerr << "object_names_base_list: " << object_names_base_list << endl;
   for (unsigned int k=0;k<size;++k) {
   display_name = " " + (*b)[k].name();
   glNewList(object_names_base_list+k, GL_COMPILE);
   glCallLists(display_name.size(),GL_BYTE,(GLubyte*) display_name.c_str());
   glEndList();
   cerr << "created list: " << display_name << endl;
   }
   }
*/


void XOrsaOpenGLEvolutionWidget::internal_draw_Lagrange_points() {
  
  if (Lagrange_body_1 == Lagrange_body_2) return;
  
  const double m_1 = evol_frame[Lagrange_body_1].mass();
  const double m_2 = evol_frame[Lagrange_body_2].mass();
  
  // assume m1 > m2
  double m1, m2;
  Vector tmp_x,tmp_y;
  Vector x1;
  if (m_1>m_2) {
    m1 = m_1;
    m2 = m_2;
    // x1 = - dir_pos_mod - evol_frame[Lagrange_body_1].position();   
    x1 = CenterBodyPosition() - evol_frame[Lagrange_body_1].position();   
    tmp_x  = evol_frame[Lagrange_body_2].position() - evol_frame[Lagrange_body_1].position();
    tmp_y  = evol_frame[Lagrange_body_2].velocity() - evol_frame[Lagrange_body_1].velocity();
    tmp_y /= tmp_y.Length();
    double tmp = tmp_y*tmp_x/tmp_x.Length();
    tmp_y -= tmp*tmp_x/tmp_x.Length();
    tmp_y /= tmp_y.Length();
    tmp_y *= tmp_x.Length();
    // cerr << " diff: " << tmp << " --> " << tmp_y*tmp_x << endl;
  } else {
    m1 = m_2;
    m2 = m_1;
    // x1 = - dir_pos_mod - evol_frame[Lagrange_body_2].position();
    x1 = CenterBodyPosition() - evol_frame[Lagrange_body_2].position();
    tmp_x  = evol_frame[Lagrange_body_1].position() - evol_frame[Lagrange_body_2].position();
    tmp_y  = evol_frame[Lagrange_body_1].velocity() - evol_frame[Lagrange_body_2].velocity();
    tmp_y /= tmp_y.Length();
    double tmp = tmp_y*tmp_x/tmp_x.Length();
    tmp_y -= tmp*tmp_x/tmp_x.Length();
    tmp_y /= tmp_y.Length();
    tmp_y *= tmp_x.Length();
    // cerr << " diff: " << tmp << " --> " << tmp_y*tmp_x << endl;
  }
  
  const Vector x = tmp_x;
  const Vector y = tmp_y;
  
  const double mu_1  = m1/(m1+m2); 
  const double mu_2  = m2/(m1+m2); 
  const double mu_21 = mu_2/mu_1;
  const double alpha = secure_pow(mu_2/(3*mu_1),1.0/3.0);
  
  double r1, r2;
  
  // L1
  r2 = alpha - (1.0/3.0)*secure_pow(alpha,2) - (1.0/9.0)*secure_pow(alpha,3) - (23.0/81.0)*secure_pow(alpha,4); 
  r1 = 1.0 - r2;
  const Vector L1 = r1 * x - x1;
  
  // L2 
  r2 = alpha + (1.0/3.0)*secure_pow(alpha,2) - (1.0/9.0)*secure_pow(alpha,3) - (31.0/81.0)*secure_pow(alpha,4); 
  r1 = 1.0 + r2;
  const Vector L2 = r1 * x - x1;
  
  // L3
  const double beta = - (7.0/12.0)*mu_21 + (7.0/12.0)*secure_pow(mu_21,2) - (13223.0/20736.0)*secure_pow(mu_21,3);
  r1 = 1.0 + beta;
  r2 = 2.0 + beta;
  const Vector L3 = - r1 * x - x1;
  
  // L4
  const Vector L4 = (0.5-mu_2)*x + secure_sqrt(3.0)/2.0*y - x1;
  
  // L5
  const Vector L5 = (0.5-mu_2)*x - secure_sqrt(3.0)/2.0*y - x1;
  
  // render Lagrangian points
  string L_name;
  
  // glColor3d(0,0.7,0);
  // glColor3d(0.6,0.4,0.0);
  glColor3d(1.0,0.6,0.0);
  glPointSize(2);
  gl2psPointSize(2);
  
  QString text;
  
  L_name = ("L1");
  glBegin(GL_POINTS);
  glVertex3d(L1.x, L1.y, L1.z);
  glEnd();
  // glRasterPos3d(L1.x,L1.y,L1.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  // renderText(L1.x, L1.y, L1.z,text,default_font); 
  renderText(L1,text); 
  
  L_name = ("L2");
  glBegin(GL_POINTS);
  glVertex3d(L2.x, L2.y, L2.z);
  glEnd();
  // glRasterPos3d(L2.x,L2.y,L2.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  // renderText(L2.x, L2.y, L2.z,text,default_font); 
  renderText(L2,text); 
  
  L_name = ("L3");
  glBegin(GL_POINTS);
  glVertex3d(L3.x, L3.y, L3.z);
  glEnd();
  // glRasterPos3d(L3.x,L3.y,L3.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  // renderText(L3.x, L3.y, L3.z,text,default_font); 
  renderText(L3,text); 
  
  L_name = ("L4");
  glBegin(GL_POINTS);
  glVertex3d(L4.x, L4.y, L4.z);
  glEnd();
  // glRasterPos3d(L4.x,L4.y,L4.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  // renderText(L4.x, L4.y, L4.z,text,default_font); 
  renderText(L4,text); 
  
  L_name = ("L5");
  glBegin(GL_POINTS);
  glVertex3d(L5.x, L5.y, L5.z);
  glEnd();
  // glRasterPos3d(L5.x,L5.y,L5.z);
  // glCallLists(L_name.size(),GL_BYTE,(GLubyte*) L_name.c_str());
  text.sprintf(" %s",L_name.c_str());
  // renderText(L5.x, L5.y, L5.z,text,default_font); 
  renderText(L5,text); 
}

void XOrsaOpenGLEvolutionWidget::internal_draw_OSD() {
  OSD osd(this);
  osd.draw();
}

// OSD

void OSD::draw() {
  
  GLint saved_depth_func;
  glGetIntegerv(GL_DEPTH_FUNC,&saved_depth_func);
  glDepthFunc(GL_ALWAYS);
  
  // OSD text color
  // glColor3d(0,1.0,0);  
  glColor3d(0.5,1.0,1.0);  
  
  // set writing zone
  zone = TOP_LEFT;
  
  // time
  {
    QString time_label;
    //
    switch (universe->GetUniverseType()) {
    case Real: {
      // FineDate(time_label,opengl->evol_frame.GetDate());
      FineDate_HMS(time_label,opengl->evol_frame.GetDate());
      time_label.prepend("time: ");
      time_label.append(" ");
      time_label.append(TimeScaleLabel(universe->GetTimeScale()).c_str());
      break;
    }
    case Simulated: 
      time_label.sprintf("time: %-6g %s",opengl->evol_frame.Time(),TimeLabel().c_str());
      break;
    }
    //
    write(time_label);
  }
  
  // center body
  {
    QString text;
    text.sprintf("central object: %s",opengl->CenterBodyName().c_str());
    write(text); 
    
    if (opengl->projection == OGL_PERSPECTIVE) {
      length_unit lu = AutoLengthUnit((double)opengl->distance);
      text.sprintf("distance from %s: %g %s",opengl->CenterBodyName().c_str(),FromUnits((double)opengl->distance,lu,-1),units->label(lu).c_str());
      write(text); 
    }
  }
  
  // eye
  {
    QString text;
    if (opengl->eye_on_body) {
      text.sprintf("observing from %s",opengl->EyeBodyName().c_str());
    } else {
      text.sprintf("observer position: free");
    }
    write(text); 
  }
  
  zone = TOP_RIGHT;
  QString text;
  text.sprintf("FPS: %.1f",opengl->get_fps());
  write(text); 
  
  glDepthFunc(saved_depth_func);
}
